/* hash.c - IRC network state database
 * Copyright 2000-2002 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: hash.h,v 1.52.2.14.2.17 2002/04/29 05:11:55 entrope Exp $
 */

#ifndef HASH_H
#define HASH_H

#include "config.h"

#include <time.h>
#include "common.h"
#include "dict.h"

#define MODE_CHANOP		0x0001
#define MODE_VOICE		0x0002
#define MODE_PRIVATE		0x0004
#define MODE_SECRET		0x0008
#define MODE_MODERATED		0x0010
#define MODE_TOPICLIMIT		0x0020
#define MODE_INVITEONLY		0x0040
#define MODE_NOPRIVMSGS		0x0080
#define MODE_KEY		0x0100
#define MODE_BAN		0x0200
#define MODE_LIMIT		0x0400
#define MODE_DELAYJOINS         0x0800

#define FLAGS_OPER		0x0001 /* Operator +O */
#define FLAGS_LOCOP		0x0002 /* Local operator +o */
#define FLAGS_INVISIBLE		0x0004 /* invisible +i */
#define FLAGS_WALLOP		0x0008 /* receives wallops +w */
#define FLAGS_SERVNOTICE	0x0010 /* receives server notices +s */
#define FLAGS_DEAF		0x0020 /* deaf +d */
#define FLAGS_SERVICE		0x0040 /* cannot be kicked, killed or deoped +k */
#define FLAGS_GLOBAL		0x0080 /* receives global messages +g */
#define FLAGS_HELPER		0x0100 /* (network?) helper +h */
#define FLAGS_PERSISTENT	0x0200 /* for reserved nicks, this isn't just one-shot */
#define FLAGS_GAGGED		0x0400 /* for gagged users */
#define FLAGS_AWAY		0x0800 /* for away users */

#define IsOper(x)               ((x)->modes & FLAGS_OPER)
#define IsService(x)            ((x)->modes & FLAGS_SERVICE)
#define IsDeaf(x)               ((x)->modes & FLAGS_DEAF)
#define IsInvisible(x)          ((x)->modes & FLAGS_INVISIBLE)
#define IsGlobal(x)             ((x)->modes & FLAGS_GLOBAL)
#define IsWallOp(x)             ((x)->modes & FLAGS_WALLOP)
#define IsServNotice(x)         ((x)->modes & FLAGS_SERVNOTICE)
#define IsHelperIrcu(x)         ((x)->modes & FLAGS_HELPER)
#define IsGagged(x)             ((x)->modes & FLAGS_GAGGED)
#define IsAway(x)               ((x)->modes & FLAGS_AWAY)
#define IsLocal(x)              ((x)->uplink == self)

#define NICKLEN         30
#define N2K_NICKLEN     30
#define USERLEN         10
#define HOSTLEN         63
#define REALLEN         50
#define KEYLEN          23
#define TOPICLEN        250
#define CHANNELLEN      200
extern unsigned int nicklen, force_n2k;
/* NICKLEN is the maximum we allow; nicklen is the maximum we think
 * the server allows.  force_n2k is set if we force using five-digit
 * numerics for users. */

#define MAXMODEPARAMS	6
#define MAXBANS		30

DECLARE_LIST(userList, struct userNode*);
DECLARE_LIST(modeList, struct modeNode*);
DECLARE_LIST(banList, struct banNode*);
DECLARE_LIST(channelList, struct chanNode*);
DECLARE_LIST(serverList, struct server*);

struct userNode {
    unsigned int num_local : 18;
    unsigned int dead : 1;
    unsigned long ip;
    long modes;                   /* user flags +isw etc... */

    char numeric[COMBO_NUMERIC_LEN+1];
    char nick[NICKLEN + 1];       /* Unique name of the client, nick or host */
    char ident[USERLEN + 1];      /* Per-host identification for user */
    char info[REALLEN + 1];       /* Free form additional client information */
    char hostname[HOSTLEN + 1];

    time_t timestamp;
    struct server *uplink;
    struct modeList channels;

    /* from nickserv */
    struct handle_info *handle_info;
    struct userNode *next_authed;
    struct policer *auth_policer;
    /* from policer */
    struct policer *command_policer;
};

typedef unsigned long chan_mode_t;

struct chanNode {
  char name[CHANNELLEN + 1];
  chan_mode_t modes;
  unsigned int limit;
  char key[KEYLEN + 1];
  time_t timestamp; /* creation time */
  
  char topic[TOPICLEN + 1];
  char topic_nick[NICKLEN + 1];
  time_t topic_time;

    struct modeList members;
    struct banList banlist;
    struct policer *join_policer;
    unsigned int join_flooded : 1;
    unsigned int bad_channel : 1;

    struct chanData *channel_info;
    struct channel_help *channel_help;
};

struct banNode {
    char ban[NICKLEN + USERLEN + HOSTLEN + 3]; /* 1 for '\0', 1 for ! and 1 for @ = 3 */
    char who[NICKLEN + 1]; /* who set ban */
    time_t set; /* time ban was set */
};

struct modeNode {
    struct chanNode *channel;
    struct userNode *user;
    long modes;	
    time_t idle_since;
};

#define SERVERNAMEMAX 64
#define SERVERDESCRIPTMAX 128

struct server {
    char name[SERVERNAMEMAX+1];
    time_t boot;
    time_t link;
    char numeric[COMBO_NUMERIC_LEN+1];
    char description[SERVERDESCRIPTMAX+1];
    unsigned int num_mask, hops, clients, max_clients;
    unsigned int burst : 1;
    struct server *uplink;
    struct userNode **users; /* flat indexed by numeric */
    struct serverList children;
};

extern struct server *self;
extern dict_t channels;
extern dict_t opserv_chan_warn;
extern dict_t clients;
extern dict_t servers;
extern struct server *servers_num[64*64];
extern unsigned int max_clients, invis_clients;
extern time_t max_clients_time;
extern struct userList curr_opers, curr_helpers;

struct server* GetServer(const char *numeric); /* using numeric */
struct server* GetServerH(const char *name); /* using full name */
struct userNode* GetUserN(const char *numeric); /* using numeric */
struct userNode* GetUserH(const char *nick);   /* using nick */
struct userNode* GetUser(const char *user); /* either :Nick or NUM */
struct chanNode* GetChannel(const char *name);

struct modeNode* GetUserMode(struct chanNode* channel, struct userNode* user);

typedef void (*server_link_func_t) (struct server *server);
void reg_server_link_func(server_link_func_t handler);

struct server* AddServer(struct server* uplink, const char *name, int hops, time_t boot, time_t link, const char *numeric, const char *description);
void DelServer(struct server* serv, int announce, const char *message);

#define IsNickName(x) ( \
        (isalpha(*(x)) || \
         ((*(x) >= 0x5B) && (*(x) <= 0x60)) || \
         ((*(x) >= 0x7B) && (*(x) <= 0x7D))) && \
        !strcspn((x)+1, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789[]\\`_^{|}"))
typedef int (*new_user_func_t) (struct userNode *user);
void reg_new_user_func(new_user_func_t handler);
struct userNode* AddUser(struct server* uplink, const char *nick, const char *ident, const char *hostname, const char *modes, const char *numeric, const char *userinfo, time_t timestamp, const char *realip);
typedef void (*del_user_func_t) (struct userNode *user);
void reg_del_user_func(del_user_func_t handler);
void unreg_del_user_func(del_user_func_t handler);
void DelUser(struct userNode* user, struct userNode *killer, int announce, const char *why);
void ReintroduceUser(struct userNode* user);
typedef void (*nick_change_func_t)(struct userNode *user, const char *new_nick);
void reg_nick_change_func(nick_change_func_t handler);
void NickChange(struct userNode* user, const char *new_nick, int no_announce);

typedef void (*new_channel_func_t) (struct chanNode *chan);
void reg_new_channel_func(new_channel_func_t handler);
typedef void (*join_func_t) (struct userNode *user, struct chanNode *chan);
void reg_join_func(join_func_t handler);
typedef void (*del_channel_func_t) (struct chanNode *chan);
void reg_del_channel_func(del_channel_func_t handler);

#define IsChannelName(x) ((*(x) == '#' || *(x) == '+') && !strpbrk((x)+1,"\a,"))
struct chanNode* AddChannel(const char *name, time_t time_, const char *modes, const char *key, int limit, char *members, char *banlist);
void DelChannel(struct chanNode* channel);

struct modeNode* AddChannelUser(struct userNode* user, struct chanNode* channel);

typedef void (*part_func_t) (struct userNode *user, struct chanNode *chan);
void reg_part_func(part_func_t handler);
void unreg_part_func(part_func_t handler);
void DelChannelUser(struct userNode* user, struct chanNode* channel, const char *reason, int deleting);
void KickChannelUser(struct userNode* target, struct chanNode* channel, struct userNode *kicker, const char *why);

typedef int (*kick_func_t) (struct userNode *kicker, struct userNode *user, struct chanNode *chan);
void reg_kick_func(kick_func_t handler);
void ChannelUserKicked(struct userNode* kicker, struct userNode* victim, struct chanNode* channel);


int ChannelBanExists(struct chanNode *channel, char *ban);
void AddChannelBan(unsigned int num, char **banset, struct chanNode* channel, struct userNode *who, time_t set, int announce);
void DelChannelBan(unsigned int num, char **banset, struct chanNode* channel, struct userNode *who, int announce);

void AddChannelOp(unsigned int num, struct userNode **newops, struct chanNode *channel, struct userNode *who, int announce);
void DelChannelOp(unsigned int num, struct userNode **exops, struct chanNode *channel, struct userNode *who, int announce);

void AddChannelVoice(unsigned int num, struct userNode **newv, struct chanNode *channel, struct userNode *who, int announce);
void DelChannelVoice(unsigned int num, struct userNode **exv, struct chanNode *channel, struct userNode *who, int announce);

typedef int (*topic_func_t)(struct userNode *who, struct chanNode *chan, const char *old_topic);
void reg_topic_func(topic_func_t handler);
void SetChannelTopic(struct chanNode *channel, struct userNode *user, const char *topic, int announce);

void init_structs(void);
void send_burst(void);

#endif
