/* ioset.h - srvx event loop
 * Copyright 2002 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: ioset.c,v 1.1.2.8 2002/05/03 01:54:25 entrope Exp $
 */

#include "config.h"
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#include "ioset.h"
#include "log.h"
#include "timeq.h"

#ifndef IOSET_DEBUG
#define IOSET_DEBUG 0
#endif

extern int uplink_connect(void);

static struct io_fd **fds;
static unsigned int fds_size;
static fd_set read_fds, write_fds;

void ioset_cleanup(void)
{
    free(fds);
}

struct io_fd *ioset_add(int fd)
{
    struct io_fd *res;
    int flags;

    if (fd < 0) {
        log(MAIN_LOG, LOG_ERROR, "Somebody called ioset_add(%d) on a negative fd!\n", fd);
        return 0;
    }
    res = calloc(1, sizeof(*res));
    if (!res) return 0;
    res->send_size = 1024;
    res->send = malloc(res->send_size);
    if (!res->send) {
        free(res->send);
        return 0;
    }
    res->fd = fd;
    if ((unsigned)fd >= fds_size) {
        unsigned int old_size = fds_size;
        fds_size = fd + 8;
        fds = realloc(fds, fds_size*sizeof(*fds));
        memset(fds+old_size, 0, (fds_size-old_size)*sizeof(*fds));
    }
    fds[fd] = res;
    flags = fcntl(fd, F_GETFL);
    fcntl(fd, F_SETFL, flags|O_NONBLOCK);
    return res;
}

static void
ioset_try_write(struct io_fd *fd)
{
    int res;
    unsigned int req = fd->send_used;
    if (req > (fd->send_size-fd->send_pos)) req = fd->send_size - fd->send_pos;
    if (IOSET_DEBUG) {
        log(DEBUG_LOG, LOG_INFO, "write()'ing to %d: send_pos=%d send_used=%d send_size=%d req=%d\n",
            fd->send_pos, fd->send_used, fd->send_size, req);
    }
    res = write(fd->fd, fd->send+fd->send_pos, req);
    if (IOSET_DEBUG) {
        log(DEBUG_LOG, LOG_INFO, "ioset write(%d, \"%.*s\") -> %d\n", fd->fd, req, fd->send+fd->send_pos, res);
    }
    if (res < 0) {
        switch (errno) {
        case EAGAIN: break;
        default:
            log(MAIN_LOG, LOG_ERROR, "write() on fd %d error %d: %s\n", fd->fd, errno, strerror(errno));
        }
    } else {
        fd->send_pos += res;
        if (fd->send_pos == fd->send_size) fd->send_pos = 0;
        fd->send_used -= res;
    }
}

void ioset_close(int fd, int os_close)
{
    struct io_fd *fdp;
    if (!(fdp = fds[fd])) return;
    fds[fd] = NULL;
    if (fdp->destroy_cb) fdp->destroy_cb(fdp);
    if (fdp->send_used > 0) {
        int flags = fcntl(fd, F_GETFL);
        fcntl(fd, F_SETFL, flags&~O_NONBLOCK);
        ioset_try_write(fdp);
        /* it may need to send the beginning of the buffer now.. */
        if (fdp->send_used > 0) ioset_try_write(fdp);
    }
    if (fdp->send) free(fdp->send);
    if (os_close) close(fd);
    free(fdp);
    FD_CLR(fd, &read_fds);
    FD_CLR(fd, &write_fds);
}

void ioset_run(void)
{
    extern int quit_services;
    extern int socket_fd;

    struct timeval select_timeout;
    unsigned int nn;
    int select_result, max_fd;
    time_t wakey;
    struct io_fd *fd;

    while (!quit_services) {
        if (socket_fd == -1) uplink_connect();

        /* How long to sleep? (fill in select_timeout) */
        wakey = timeq_next();
        if ((wakey - now) < 0) {
            select_timeout.tv_sec = 0;
        } else {
            select_timeout.tv_sec = wakey - now;
        }
        select_timeout.tv_usec = 0;

        /* Set up read_fds and write_fds fdsets. */
        FD_ZERO(&read_fds);
        FD_ZERO(&write_fds);
        max_fd = 0;
        for (nn=0; nn<fds_size; nn++) {
            if (!(fd = fds[nn])) continue;
            max_fd = nn;
            if (fd->wants_reads) FD_SET(nn, &read_fds);
            if ((fd->send_used > 0) || !fd->connected) FD_SET(nn, &write_fds);
        }

        /* Check for activity, update time. */
        select_result = select(max_fd + 1, &read_fds, &write_fds, NULL, &select_timeout);
        now = time(NULL);
        if (select_result < 0) {
            if (errno != EINTR) {
                log(MAIN_LOG, LOG_ERROR, "select() error %d: %s\n", errno, strerror(errno));
                close_socket();
            }
            continue;
        }

        /* Call back anybody that has connect or read activity and wants to know. */
        for (nn=0; nn<fds_size; nn++) {
            if (!(fd = fds[nn])) continue;
            if (FD_ISSET(nn, &read_fds)) {
                fd->readable_cb(fd);
            }
            if (FD_ISSET(nn, &write_fds) && !fd->connected) {
                fd->connected = 1;
                if (fd->connect_cb) fd->connect_cb(fd);
            }
            /* Note: check whether write FD is still set, since the
             * connect_cb() might close the FD, making us dereference
             * a free()'d pointer for the fd.
             */
            if (FD_ISSET(nn, &write_fds) && (fd->send_used > 0)) ioset_try_write(fd);
        }

        /* Call any timeq events we need to call. */
        timeq_run(now);
    }
}

void
ioset_write(struct io_fd *fd, const char *buf, unsigned int nbw)
{
    unsigned int avail, next_pos;
    next_pos = fd->send_pos + fd->send_used;
    if (next_pos >= fd->send_size) next_pos -= fd->send_size;
    avail = (next_pos < fd->send_pos) ? (fd->send_pos - next_pos) : (fd->send_size - next_pos);
    if (avail > nbw) avail = nbw;
    if (IOSET_DEBUG) {
        log(DEBUG_LOG, LOG_INFO, "ioset_write(%d, ..., %d): send_pos=%d send_used=%d send_size=%d next_pos=%d avail=%d\n",
            fd->fd, nbw, fd->send_pos, fd->send_used, fd->send_size, next_pos, avail);
    }
    memcpy(fd->send + next_pos, buf, avail);
    fd->send_used += avail;
    buf += avail;
    nbw -= avail;
    if (nbw > 0) {
        if (nbw + fd->send_used > fd->send_size) {
            /* need to grow buffer */
            unsigned int new_size;
            char *new_buf;
            new_size = fd->send_size << 1;
            if (nbw + fd->send_used > new_size) new_size = nbw + fd->send_used;
            new_buf = malloc(new_size);
            if (next_pos < fd->send_pos) {
                if (IOSET_DEBUG) {
                    log(DEBUG_LOG, LOG_INFO, "Two-part copy: (%d, %d)->(0,%d) and (0,%d)->(%d,%d)\n",
                        fd->send_pos, fd->send_size, fd->send_size-fd->send_pos, fd->send_used-(fd->send_size-fd->send_pos),
                        fd->send_size-fd->send_pos, fd->send_used);
                }
                memcpy(new_buf, fd->send+fd->send_pos, fd->send_size-fd->send_pos);
                memcpy(new_buf+fd->send_size-fd->send_pos, fd->send, fd->send_used-(fd->send_size-fd->send_pos));
            } else {
                if (IOSET_DEBUG) {
                    log(DEBUG_LOG, LOG_INFO, "One-part copy: (%d, %d)->(0,%d)\n",
                        fd->send_pos, fd->send_pos+fd->send_used, fd->send_used);
                }
                memcpy(new_buf, fd->send+fd->send_pos, fd->send_used);
            }
            memcpy(new_buf+fd->send_used, buf, nbw);
            free(fd->send);
            fd->send = new_buf;
            fd->send_size = new_size;
            fd->send_pos = 0;
            fd->send_used += nbw;
            if (IOSET_DEBUG) {
                log(DEBUG_LOG, LOG_INFO, "After growing buffer: send_used=%d send_size=%d\n",
                    fd->send_used, fd->send_size);
            }
        } else {
            /* just put at beginning of buffer */
            memcpy(fd->send, buf, nbw);
            fd->send_used += nbw;
        }
    }
}
