/* nickserv.c - Nick/authentiction service
 * Copyright 2000-2002 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: nickserv.c,v 1.138.2.65.2.51 2002/05/04 13:09:28 entrope Exp $
 */

#include "config.h"
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <stdlib.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_REGEX_H
#include <regex.h>
#endif

#include "alias.h"
#include "chanserv.h"
#include "common.h"
#include "conf.h"
#include "dict.h"
#include "hash.h"
#include "helpfile.h"
#include "log.h"
#include "messages.h"
#include "nickserv.h"
#include "opserv.h"
#include "parse.h"
#include "policer.h"
#include "recdb.h"
#include "sendmail.h"
#include "timeq.h"

#define NICKSERV_CONF_NAME "services/nickserv"
#define NICKSERV_DB_NAME "nickserv.db"
#define TEMP_DB_NAME "nickserv.db.new"
#define NICKSERV_HELPFILE_NAME "nickserv.help"

#define KEY_DISABLE_NICKS "disable_nicks"
#define KEY_REQUIRE_QUALIFIED "require_qualified"
#define KEY_DEFAULT_HOSTMASK "default_hostmask"
#define KEY_NICKS_PER_HANDLE "nicks_per_handle"
#define KEY_NICKS_PER_ACCOUNT "nicks_per_account"
#define KEY_PASSWORD_MIN_LENGTH "password_min_length"
#define KEY_PASSWORD_MIN_DIGITS "password_min_digits"
#define KEY_PASSWORD_MIN_UPPER "password_min_upper"
#define KEY_PASSWORD_MIN_LOWER "password_min_lower"
#define KEY_VALID_HANDLE_REGEX "valid_handle_regex"
#define KEY_VALID_ACCOUNT_REGEX "valid_account_regex"
#define KEY_VALID_NICK_REGEX "valid_nick_regex"
#define KEY_DB_BACKUP_FREQ "db_backup_freq"
#define KEY_MODOPER_LEVEL "modoper_level"
#define KEY_SET_EPITHET_LEVEL "set_epithet_level"
#define KEY_FLAG_LEVELS "flag_levels"
#define KEY_HANDLE_EXPIRE_FREQ	"handle_expire_freq"
#define KEY_ACCOUNT_EXPIRE_FREQ "account_expire_freq"
#define KEY_HANDLE_EXPIRE_DELAY	"handle_expire_delay"
#define KEY_ACCOUNT_EXPIRE_DELAY "account_expire_delay"
#define KEY_NOCHAN_HANDLE_EXPIRE_DELAY "nochan_handle_expire_delay"
#define KEY_NOCHAN_ACCOUNT_EXPIRE_DELAY "nochan_account_expire_delay"
#define KEY_DICT_FILE "dict_file"
#define KEY_NICK "nick"
#define KEY_AUTOGAG_ENABLED "autogag_enabled"
#define KEY_AUTOGAG_DURATION "autogag_duration"
#define KEY_AUTH_POLICER "auth_policer"
#define KEY_EMAIL_VISIBLE_LEVEL "email_visible_level"
#define KEY_EMAIL_ENABLED "email_enabled"
#define KEY_EMAIL_REQUIRED "email_required"
#define KEY_COOKIE_TIMEOUT "cookie_timeout"
#define KEY_WARN_NICK_OWNED "warn_nick_owned"
#define KEY_ACCOUNTS_PER_EMAIL "accounts_per_email"

#define KEY_PASSWD "passwd"
#define KEY_NICKS "nicks"
#define KEY_MASKS "masks"
#define KEY_OPSERV_LEVEL "opserv_level"
#define KEY_FLAGS "flags"
#define KEY_REGISTER_ON "register"
#define KEY_LAST_SEEN "lastseen"
#define KEY_INFO "info"
#define KEY_USERLIST_STYLE "user_style"
#define KEY_SCREEN_WIDTH "screen_width"
#define KEY_LAST_AUTHED_HOST "last_authed_host"
#define KEY_EMAIL_ADDR "email_addr"
#define KEY_COOKIE "cookie"
#define KEY_COOKIE_DATA "data"
#define KEY_COOKIE_TYPE "type"
#define KEY_COOKIE_EXPIRES "expires"
#define KEY_ACTIVATION "activation"
#define KEY_PASSWORD_CHANGE "password change"
#define KEY_EMAIL_CHANGE "email change"
#define KEY_ALLOWAUTH "allowauth"
#define KEY_EPITHET "epithet"

#define NSMSG_RIGHT_NOW "Right now!"
#define NSMSG_NONE "(none)"
#define NSMSG_UNKNOWN "<unknown>"
#define NSMSG_HANDLE_EXISTS	"Account $b%s$b already registered."
#define NSMSG_HANDLE_TOO_LONG	"Account name too long."
#define NSMSG_INSECURE_PASSWORD	"Your password must be at least %lu characters long, with at least %lu digit(s), %lu capital letter(s) and %lu lower-case letter(s).  It should not be your account name, the word \"password\", or anything else obvious."
#define NSMSG_PARTIAL_REGISTER  "Account has been registered to you; nick was already registered to someone else."
#define NSMSG_OREGISTER_H_SUCCESS "Account has been registered."
#define NSMSG_REGISTER_H_SUCCESS "Account has been registered to you."
#define NSMSG_REGISTER_HN_SUCCESS "Account and nick have been registered to you."
#define NSMSG_REQUIRE_OPER	 "You must be an $bIRC Operator$b to register the first account."
#define NSMSG_ROOT_HANDLE	 "Account %s has been granted $broot-level privileges$b."
#define NSMSG_USE_COOKIE_REGISTER "To activate your account, you must check your email for the \"cookie\" that has been mailed to it.  When you have it, use the $bcookie$b command to complete registration."
#define NSMSG_USE_COOKIE_RESETPASS "A cookie has been mailed to your account's email address.  You must check your email and use the $bcookie$b command to confirm the password change."
#define NSMSG_USE_COOKIE_EMAIL_1  "A cookie has been mailed to the new address you requested.  To finish setting your email address, please check your email for the cookie and use the $bcookie$b command to verify."
#define NSMSG_USE_COOKIE_EMAIL_2  "A cookie has been generated, and half mailed to each your old and new addresses.  To finish changing your email address, please check your email for the cookie and use the $bcookie$b command to verify."
#define NSMSG_USE_COOKIE_AUTH     "A cookie has been generated and sent to your email address.  Once you have checked your email and received the cookie, auth using the $bcookie$b command." 
#define NSMSG_COOKIE_LIVE         "Account $b%s$b already has a cookie active.  Please either finish using that cookie, wait for it to expire, or auth to the account and use the $bdelcookie$b command."
#define NSMSG_EMAIL_UNACTIVATED   "That email address already has an unused cookie outstanding.  Please use the cookie or wait for it to expire."
#define NSMSG_NO_COOKIE           "Your account does not have any cookie issued right now."
#define NSMSG_BAD_COOKIE          "That cookie is not the right one.  Please make sure you are copying it EXACTLY from the email; it is case-sensitive, so $bABC$b is different from $babc$b."
#define NSMSG_HANDLE_ACTIVATED    "Your account is now activated, with the password you entered when you registered.  You are currently authenticated to your account."
#define NSMSG_PASSWORD_CHANGED    "You have successfully changed your password to what you requested with the $bresetpass$b command."
#define NSMSG_EMAIL_PROHIBITED    "%s may not be used as an email address: %s"
#define NSMSG_EMAIL_OVERUSED      "There are already the maximum number of accounts associated with that email address."
#define NSMSG_EMAIL_SAME          "That is the email address already there; no need to change it."
#define NSMSG_EMAIL_CHANGED       "You have successfully changed your email address."
#define NSMSG_BAD_COOKIE_TYPE     "Your account had bad cookie type %d; sorry.  I'm confused.  Please report this bug."
#define NSMSG_MUST_TIME_OUT       "You must wait for cookies of that type to time out."
#define NSMSG_ATE_COOKIE          "I ate the cookie for your account.  You may now have another."
#define NSMSG_USE_RENAME          "You're already authenticated to account $b%s$b -- contact the support staff to rename your account."
#define NSMSG_REGISTER_BAD_NICKMASK "Could not recognize $b%s$b as either a current nick or a hostmask."
#define NSMSG_NICK_NOT_REGISTERED "Nick $b%s$b hasn't been registered to any account."
#define NSMSG_HANDLE_NOT_FOUND	"Couldn't find your account -- did you register yet?"
#define NSMSG_ALREADY_AUTHED    "You are already authed to account $b%s$b; you must reconnect to auth to a different one."
#define NSMSG_HOSTMASK_INVALID	"Your hostmask is not valid for account $b%s$b."
#define NSMSG_USER_IS_SERVICE   "$b%s$b is a network service; you can only use that command on real users."
#define NSMSG_HANDLEINFO_ON	  "Account information for $b%s$b:"
#define NSMSG_HANDLEINFO_REGGED   "  Registered on: %s"
#define NSMSG_HANDLEINFO_LASTSEEN "  Last seen: %s"
#define NSMSG_HANDLEINFO_VACATION "  On vacation."
#define NSMSG_HANDLEINFO_EMAIL_ADDR "  Email address: %s"
#define NSMSG_HANDLEINFO_INFOLINE "  Infoline: %s"
#define NSMSG_HANDLEINFO_FLAGS    "  Flags: %s"
#define NSMSG_HANDLEINFO_EPITHET  "  Epithet: %s"
#define NSMSG_HANDLEINFO_LAST_HOST "  Last authed hostmask: %s"
#define NSMSG_HANDLEINFO_NICKS    "  Nickname(s): %s"
#define NSMSG_HANDLEINFO_MASKS    "  Hostmask(s): %s"
#define NSMSG_HANDLEINFO_CHANNELS "  Channel(s): %s"
#define NSMSG_HANDLEINFO_CURRENT  "  Current nickname(s): %s"
#define NSMSG_USERINFO_AUTHED_AS  "$b%s$b is authenticated to account $b%s$b."
#define NSMSG_USERINFO_NOT_AUTHED "$b%s$b is not authenticated to any account."
#define NSMSG_NICKINFO_OWNER    "Nick $b%s$b is owned by account $b%s$b."
#define NSMSG_PASSWORD_INVALID	"Incorrect password; please try again."
#define NSMSG_PLEASE_SET_EMAIL  "We now require email addresses for users.  Please use the $bset email$b command to set your email address!"
#define NSMSG_WEAK_PASSWORD     "WARNING: You are using a password that is considered weak (easy to guess).  It is STRONGLY recommended you change it (now, if not sooner) by typing \"/msg $N PASS oldpass newpass\" (with your current password and a new password)."
#define NSMSG_HANDLE_SUSPENDED	"Your $N account has been suspended."
#define NSMSG_AUTH_SUCCESS	"I recognize you."
#define NSMSG_AUTH_ALLOWED	"User $b%s$b may now authenticate to account $b%s$b."
#define NSMSG_AUTH_ALLOWED_MSG	"You may now authenticate to account $b%s$b by typing $b/msg $n auth %s password$b (using your password)."
#define NSMSG_AUTH_NORMAL_ONLY	"User $b%s$b may now only authenticate to accounts with matching hostmasks."
#define NSMSG_AUTH_UNSPECIAL	"User $b%s$b did not have any special auth allowance."
#define NSMSG_MUST_QUALIFY      "You $bMUST$b \"/msg $n %s\" (not just $N) for that command."
#define NSMSG_MUST_AUTH		"You must be authenticated first."
#define NSMSG_TOO_MANY_NICKS	"You have already registered the maximum permitted number of nicks."
#define NSMSG_NICK_TOO_LONG	"Nick too long."
#define NSMSG_NICK_EXISTS	"Nick $b%s$b already registered."
#define NSMSG_NICK_IN_USE	"Nick $b%s$b is already in use."
#define NSMSG_REGNICK_SUCCESS	"Nick $b%s$b has been registered to you."
#define NSMSG_PASS_SUCCESS	"Password changed."
#define NSMSG_MASK_INVALID	"$b%s$b is an invalid hostmask."
#define NSMSG_ADDMASK_ALREADY	"$b%s$b is already a hostmask in your account."
#define NSMSG_ADDMASK_SUCCESS	"Hostmask %s added."
#define NSMSG_DELMASK_NOTLAST   "You may not delete your last hostmask."
#define NSMSG_DELMASK_SUCCESS	"Hostmask %s deleted."
#define NSMSG_DELMASK_NOT_FOUND	"Unable to find mask to be deleted."
#define NSMSG_OPSERV_LEVEL_BAD  "You may not promote another oper above your level."
#define NSMSG_USE_CMD_PASS	"Please use the PASS command to change your password."
#define NSMSG_KILL_MESSAGE	"Nick owner requested kill."
#define NSMSG_MULTI_KILL	"Issued %d kills."
#define NSMSG_UNKNOWN_NICK	"I know nothing about nick $b%s$b."
#define NSMSG_NOT_YOUR_NICK	"$b%s$b Is Nacho Nick."
#define NSMSG_NICK_INACTIVE	"Couldn't find anyone using nick $b%s$b."
#define NSMSG_NICK_USER_YOU	"I won't kill you."
#define NSMSG_SINGLE_KILL	"Issued kill for $b%s$b."
#define NSMSG_UNREGNICK_SUCCESS	"Nick $b%s$b has been unregistered."
#define NSMSG_UNREGISTER_SUCCESS "Account $b%s$b has been unregistered."
#define NSMSG_UNREGISTER_NICKS_SUCCESS "Account $b%s$b and all its nicks have been unregistered."
#define NSMSG_HANDLE_STATS	"There are %d nicks registered to your account."
#define NSMSG_HANDLE_NONE	"You are not authenticated against any account."
#define NSMSG_GLOBAL_STATS	"There are %d accounts and %d nicks registered globally."
#define NSMSG_GLOBAL_STATS_NONICK "There are %d accounts registered."
#define NSMSG_LAST_WRITE	"Last db write was %s ago."
#define NSMSG_NO_DB_WRITE	"There have been no db writes performed."
#define NSMSG_ON_VACATION       "You are now on vacation.  Your account will be preserved until you authenticate again."
#define NSMSG_NO_ACCESS		"Access denied."
#define NSMSG_INVALID_FLAGS	"$b%s$b is not a valid account flag string."
#define NSMSG_INVALID_FLAG	"$b%c$b is not a valid $N account flag."
#define NSMSG_SET_FLAG		"Applied flags $b%s$b to %s's $N account."
#define NSMSG_USER_FLAGS	"User $b%s$b has flags $b%s$b set."
#define NSMSG_WRITE_SUCCESS	"$N db write completed (in "FMT_TIME_T".%03lu seconds)."
#define NSMSG_READHELP_SUCCESS	"$N help file completed (in "FMT_TIME_T".%03lu seconds)."
#define NSMSG_DB_MERGED         "$N merged DB from %s (in "FMT_TIME_T".%03lu seconds)."
#define NSMSG_HANDLE_CHANGED	"$b%s$b's account name has been changed to $b%s$b."
#define NSMSG_BAD_HANDLE	"Account $b%s$b not registered because it is in use by a network service, or contains invalid characters."
#define NSMSG_BAD_NICK		"Nickname $b%s$b not registered because it is in use by a network service, or contains invalid characters."
#define NSMSG_BAD_EMAIL_ADDR    "Please use a well-formed email address."
#define NSMSG_FAIL_RENAME	"Account $b%s$b not renamed to $b%s$b because it is in use by a network services, or contains invalid characters."
#define NSMSG_NO_DIGIT		"An account or nickname cannot have a digit as the first character, please try again."
#define NSMSG_SEARCH_MATCH      "Match: %s"
#define NSMSG_INVALID_ACTION    "%s is an invalid search action."
#define NSMSG_HANDLES_MERGED    "Merged account $b%s$b into $b%s$b."
#define NSMSG_NICK_OWNED        "%s is a registered nick - you must auth to account %s or change your nick, or the owner may disconnect you."

#define NSMSG_SETTING_LIST	"$b$N account settings:$b"
#define NSMSG_INVALID_OPTION	"$b%s$b is an invalid account setting."

#define NSMSG_STR_SETTING	"$b%s$b%s"
#define NSMSG_NUM_SETTING	"$b%s$b%d"

#define NICKSERV_VALID_CHARS	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_"

#define NICKSERV_FUNC(NAME) int NAME(struct userNode *user, unsigned int argc, unsigned char *argv[])
typedef NICKSERV_FUNC(nickserv_func_t);

#define OPTION_FUNC(NAME) int NAME(struct userNode *user, struct handle_info *hi, unsigned int override, unsigned int argc, unsigned char *argv[])
typedef OPTION_FUNC(option_func_t);

static NICKSERV_FUNC(cmd_help);

DEFINE_LIST(handle_info_list, struct handle_info*);

#define NICKSERV_SYNTAX()	nickserv_help(user, argv[0], 0)
#define NICKSERV_MIN_PARMS(N) do { \
  (void)argv; \
  if (argc < N) { \
    nickserv_notice(user, MSG_MISSING_PARAMS, argv[0]); \
    NICKSERV_SYNTAX(); \
    return 0; } } while (0)

struct userNode *nickserv;
struct helpfile *nickserv_helpfile;
static dict_t nickserv_func_dict; /* contains nickserv_cmd* */
static dict_t nickserv_handle_dict; /* contains struct handle_info* */
static dict_t nickserv_nick_dict; /* contains struct nick_info* */
static dict_t nickserv_opt_dict; /* contains option_func_t* */
static dict_t nickserv_allow_auth_dict; /* contains struct handle_info* */
static dict_t nickserv_email_dict; /* contains struct handle_info_list*, indexed by email addr */
static const unsigned char *handle_flags = HANDLE_FLAGS;
static unsigned char handle_inverse_flags[256];
static unsigned int flag_access_levels[32];
static time_t lastwrite = 0;

struct userList curr_helpers;

static struct {
    unsigned int disable_nicks : 1;
    unsigned int require_qualified : 1;
    unsigned int valid_handle_regex_set : 1;
    unsigned int valid_nick_regex_set : 1;
    unsigned int autogag_enabled : 1;
    unsigned int email_enabled : 1;
    unsigned int email_required : 1;
    unsigned int warn_nick_owned : 1;
    unsigned long nicks_per_handle;
    unsigned long password_min_length;
    unsigned long password_min_digits;
    unsigned long password_min_upper;
    unsigned long password_min_lower;
    unsigned long db_backup_frequency;
    unsigned long handle_expire_frequency;
    unsigned long autogag_duration;
    unsigned long email_visible_level;
    unsigned long cookie_timeout;
    unsigned long handle_expire_delay;
    unsigned long nochan_handle_expire_delay;
    unsigned long modoper_level;
    unsigned long set_epithet_level;
    unsigned long handles_per_email;
    char *default_hostmask;
    regex_t valid_handle_regex;
    regex_t valid_nick_regex;
    dict_t weak_password_dict;
    struct policer_params *auth_policer_params;
} nickserv_conf;

struct nickserv_cmd
{
    int min_oper_level;
    int must_auth : 1;
    int must_be_qualified : 1;
    nickserv_func_t *func;
    double weight;
    unsigned int refs;
};

const char *NSMSG_VERSION_ID = "$Id: nickserv.c,v 1.138.2.65.2.51 2002/05/04 13:09:28 entrope Exp $";

#define nickserv_notice(target, format...) send_message(target , nickserv , ## format)

static char *
canonicalize_hostmask(char *mask)
{
    char *out = mask, *temp;
    if ((temp = strchr(mask, '!'))) {
	temp++;
	while (*temp) *out++ = *temp++;
	*out++ = 0;
    }
    return mask;
}

static struct handle_info *
register_handle(const char *handle, const char *passwd)
{
    struct handle_info *hi;
    hi = calloc(1, sizeof(*hi));
    hi->userlist_style = HI_DEFAULT_STYLE;
    hi->screen_width = DEFAULT_LINE_SIZE;
    safestrncpy(hi->handle, handle, sizeof(hi->handle));
    safestrncpy(hi->passwd, passwd, sizeof(hi->passwd));
    hi->infoline = NULL;
    dict_insert(nickserv_handle_dict, hi->handle, hi);
    return hi;
}

static void
register_nick(const char *nick, struct handle_info *owner)
{
    struct nick_info *ni;
    ni = malloc(sizeof(struct nick_info));
    safestrncpy(ni->nick, nick, sizeof(ni->nick));
    ni->owner = owner;
    ni->next = owner->nicks;
    owner->nicks = ni;
    dict_insert(nickserv_nick_dict, ni->nick, ni);
}

static void
free_nick_info(void *vni)
{
    struct nick_info *ni = vni;
    free(ni);
}

static void
delete_nick(struct nick_info *ni)
{
    struct nick_info *last, *next;
    /* Remove ni from the nick_info linked list. */
    if (ni == ni->owner->nicks) {
	ni->owner->nicks = ni->next;
    } else {
	last = ni->owner->nicks;
	next = last->next;
	while (next != ni) {
	    last = next;
	    next = last->next;
	}
	last->next = next->next;
    }
    dict_remove(nickserv_nick_dict, ni->nick);
}

static unreg_func_t *unreg_func_list;
static unsigned int unreg_func_size = 0, unreg_func_used = 0;

void
reg_unreg_func(unreg_func_t func)
{
    if (unreg_func_used == unreg_func_size) {
	if (unreg_func_size) {
	    unreg_func_size <<= 1;
	    unreg_func_list = realloc(unreg_func_list, unreg_func_size*sizeof(unreg_func_t));
	} else {
	    unreg_func_size = 8;
	    unreg_func_list = malloc(unreg_func_size*sizeof(unreg_func_t));
	}
    }
    unreg_func_list[unreg_func_used++] = func;
}

static void
free_handle_info(void *vhi)
{
    struct userNode *un, *next;
    struct handle_info *hi = vhi;

    free_string_list(hi->masks);
    for (un = hi->users; un; un = next) {
	next = un->next_authed;
        un->handle_info = NULL;
	un->next_authed = NULL;
    }
    while (hi->nicks) delete_nick(hi->nicks);
    if (hi->infoline) free(hi->infoline);
    if (hi->epithet) free(hi->epithet);
    free(hi);
}

static void
nickserv_unregister_handle(struct handle_info *hi, struct userNode *notify)
{
    unsigned int n;

    for (n=0; n<unreg_func_used; n++) unreg_func_list[n](notify, hi);
    if (notify) {
        if (nickserv_conf.disable_nicks) {
            nickserv_notice(notify, NSMSG_UNREGISTER_SUCCESS, hi->handle);
        } else {
            nickserv_notice(notify, NSMSG_UNREGISTER_NICKS_SUCCESS, hi->handle);
        }
    }
    dict_remove(nickserv_handle_dict, hi->handle);
}

struct handle_info*
get_handle_info(const char *handle)
{
    return dict_find(nickserv_handle_dict, handle, 0);
}

struct nick_info*
get_nick_info(const char *nick)
{
    return nickserv_conf.disable_nicks ? 0 : dict_find(nickserv_nick_dict, nick, 0);
}

struct modeNode *
find_handle_in_channel(struct chanNode *channel, struct handle_info *handle, struct userNode *except)
{
    unsigned int nn;
    struct modeNode *mn;

    for (nn=0; nn<channel->members.used; ++nn) {
        mn = channel->members.list[nn];
        if (mn->user == except) continue;
        if (mn->user->handle_info == handle) return mn;
    }
    return NULL;
}

static int
is_valid_oper(struct userNode *user, unsigned int min_level, unsigned int quiet)
{
    if ((!IsHelping(user) && !IsOper(user)) || (min_level && !IsOper(user)) || !user->handle_info) {
	if (!quiet) nickserv_notice(user, NSMSG_NO_ACCESS);
	return 0;
    }

    if (HANDLE_FLAGGED(user->handle_info, OPER_SUSPENDED)) {
	if (!quiet) nickserv_notice(user, MSG_OPER_SUSPENDED);
	return 0;
    }

    if (user->handle_info->opserv_level < min_level) {
	if (!quiet) nickserv_notice(user, NSMSG_NO_ACCESS);
	return 0;
    }

    return 1;
}

static int
is_valid_handle(const char *handle)
{
    struct userNode *user;
    /* cant register a juped nick/service nick as handle, to prevent confusion */
    user = GetUserH(handle);
    if (user && IsLocal(user)) return 0;
    /* test either regex or containing all valid chars */
    if (nickserv_conf.valid_handle_regex_set) {
        int err = regexec(&nickserv_conf.valid_handle_regex, handle, 0, 0, 0);
        if (err) {
            char buff[256];
            buff[regerror(err, &nickserv_conf.valid_handle_regex, buff, sizeof(buff))] = 0;
            log(NS_LOG, LOG_INFO, "regexec error: %s (%d)\n", buff, err);
        }
        return !err;
    } else {
        return !handle[strspn(handle, NICKSERV_VALID_CHARS)];
    }
}

static int
is_valid_nick(const char *nick)
{
    struct userNode *user;
    /* cant register a juped nick/service nick.. client could kill it. */
    user = GetUserH(nick);
    if (user && IsLocal(user)) return 0;
    /* test either regex or as valid handle */
    if (nickserv_conf.valid_nick_regex_set) {
        int err = regexec(&nickserv_conf.valid_nick_regex, nick, 0, 0, 0);
        if (err) {
            char buff[256];
            buff[regerror(err, &nickserv_conf.valid_nick_regex, buff, sizeof(buff))] = 0;
            log(NS_LOG, LOG_INFO, "regexec error: %s (%d)\n", buff, err);
        }
        return !err;
    } else {
        return is_valid_handle(nick);
    }
}

static int
is_valid_email_addr(const char *email)
{
    return strchr(email, '@') != NULL;
}

static const char *
visible_email_addr(struct userNode *user, struct handle_info *hi)
{
    if (hi->email_addr) {
        if (is_valid_oper(user, nickserv_conf.email_visible_level, 1)) {
            return hi->email_addr;
        } else {
            return "Set.";
        }
    } else {
        return "Not set.";
    }
}

struct handle_info *
smart_get_handle_info(struct userNode *service, struct userNode *user, const unsigned char *name)
{
    struct handle_info *hi;
    struct userNode *target;

    switch (*name) {
    case '*':
        if (!(hi = get_handle_info(++name))) {
            send_message(user, service, MSG_HANDLE_UNKNOWN, name);
            return 0;
        }
        return hi;
    default:
        if (!(target = GetUserH(name))) {
            send_message(user, service, MSG_NICK_UNKNOWN, name);
            return 0;
        }
        if (IsLocal(target)) {
            send_message(user, service, NSMSG_USER_IS_SERVICE, target->nick);
            return 0;
        }
        if (!(hi = target->handle_info)) {
            send_message(user, service, MSG_USER_AUTHENTICATE, target->nick);
            return 0;
        }
        return hi;
    }
}

#define nickserv_get_handle_info(USER, NAME) smart_get_handle_info(nickserv, USER, NAME)

static struct handle_info *
get_victim_oper(struct userNode *user, const char *target)
{
    struct handle_info *hi;
    int outranked;
    if (!(hi = nickserv_get_handle_info(user, target))) return 0;
    if (HANDLE_FLAGGED(user->handle_info, OPER_SUSPENDED)) {
	nickserv_notice(user, MSG_OPER_SUSPENDED);
	return 0;
    }
    outranked = 1;
    if (hi->opserv_level < user->handle_info->opserv_level) outranked = 0;
    if (hi->opserv_level == user->handle_info->opserv_level) {
        if (user->handle_info->opserv_level == 1000) outranked = 0;
        if (user->handle_info == hi) outranked = 0;
        if ((hi->opserv_level == 0)
            && !(HANDLE_FLAGGED(hi, SUPPORT_HELPER) || HANDLE_FLAGGED(hi, NETWORK_HELPER))
            && HANDLE_FLAGGED(user->handle_info, HELPING)) {
            outranked = 0;
        }
    }
    if (outranked) {
        nickserv_notice(user, MSG_USER_OUTRANKED, hi->handle);
	return 0;
    }
    return hi;
}

static int
valid_user_for(struct userNode *user, struct handle_info *hi)
{
    unsigned int i;
    int res = 0;
    if (hi->masks->used) {
        for (i=0; !res && i<hi->masks->used; i++) {
            /* Maybe it'd just be simpler to require the *! at the
             * start of the hostmask.  For future thought, anyway. */
            if (user_matches_glob(user, hi->masks->list[i], 0)) res = 1;
        }
    } else {
        /* if they have no masks, let them pass */
        res = 1;
    }
    if (!res && (dict_find(nickserv_allow_auth_dict, user->nick, NULL) == hi)) {
	res = 2;
    }
    return res;
}

static int
is_secure_password(const unsigned char *handle, const unsigned char *pass)
{
    unsigned int i, len;
    unsigned int cnt_digits = 0, cnt_upper = 0, cnt_lower = 0;
    len = strlen(pass);
    if (len < nickserv_conf.password_min_length) return 0;
    if (!strcasecmp(pass, handle)) return 0;
    dict_find(nickserv_conf.weak_password_dict, pass, &i);
    if (i) return 0;
    for (i=0; i<len; i++) {
	if (isdigit(pass[i])) cnt_digits++;
	if (isupper(pass[i])) cnt_upper++;
	if (islower(pass[i])) cnt_lower++;
    }
    return (cnt_lower >= nickserv_conf.password_min_lower)
	&& (cnt_upper >= nickserv_conf.password_min_upper)
	&& (cnt_digits >= nickserv_conf.password_min_digits);
}

static auth_func_t *auth_func_list;
static unsigned int auth_func_size = 0, auth_func_used = 0;

void
reg_auth_func(auth_func_t func)
{
    if (auth_func_used == auth_func_size) {
	if (auth_func_size) {
	    auth_func_size <<= 1;
	    auth_func_list = realloc(auth_func_list, auth_func_size*sizeof(auth_func_t));
	} else {
	    auth_func_size = 8;
	    auth_func_list = malloc(auth_func_size*sizeof(auth_func_t));
	}
    }
    auth_func_list[auth_func_used++] = func;
}

static handle_rename_func_t *rf_list;
static unsigned int rf_list_size, rf_list_used;

void
reg_handle_rename_func(handle_rename_func_t func)
{
    if (rf_list_used == rf_list_size) {
        if (rf_list_size) {
            rf_list_size <<= 1;
            rf_list = realloc(rf_list, rf_list_size*sizeof(rf_list[0]));
        } else {
            rf_list_size = 8;
            rf_list = malloc(rf_list_size*sizeof(rf_list[0]));
        }
    }
    rf_list[rf_list_used++] = func;
}

static void
set_user_handle_info(struct userNode *user, struct handle_info *hi)
{
    unsigned int n;
    struct handle_info *old_info;

    /* This can happen if somebody uses COOKIE while authed, or if
     * they re-auth to their current handle (which is silly, but users
     * are like that). */
    if (user->handle_info == hi) return;

    if (user->handle_info) {
	struct userNode *other;

	if (IsHelper(user)) userList_remove(&curr_helpers, user);

	/* remove from next_authed linked list */
	if (user->handle_info->users == user) {
	    user->handle_info->users = user->next_authed;
	} else {
	    for (other = user->handle_info->users;
		 other->next_authed != user;
		 other = other->next_authed) ;
	    other->next_authed = user->next_authed;
	}
	user->handle_info->lastseen = now;
    }
    old_info = user->handle_info;
    user->handle_info = hi;
    if (hi && (hi->opserv_level == 0)) HANDLE_CLEAR_FLAG(hi, HELPING);
    for (n=0; n<auth_func_used; n++) auth_func_list[n](user, old_info);
    if (hi) {
        HANDLE_CLEAR_FLAG(hi, FROZEN);
        snprintf(hi->last_authed_host, sizeof(hi->last_authed_host), "%s@%s", user->ident, user->hostname);
	user->next_authed = hi->users;
	hi->users = user;
	hi->lastseen = now;
	if (IsHelper(user)) userList_append(&curr_helpers, user);
    } else {
	user->next_authed = NULL;
    }
}

static struct handle_info*
nickserv_register(struct userNode *user, struct userNode *settee, const char *handle, const char *passwd)
{
    struct handle_info *hi;
    char crypted[MD5_CRYPT_LENGTH];

    if ((hi = dict_find(nickserv_handle_dict, handle, NULL))) {
	nickserv_notice(user, NSMSG_HANDLE_EXISTS, handle);
	return 0;
    }

    if (strlen(handle) > NICKSERV_HANDLE_LEN) {
	nickserv_notice(user, NSMSG_HANDLE_TOO_LONG);
	return 0;
    }

    if (!is_secure_password(handle, passwd)) {
	nickserv_notice(user, NSMSG_INSECURE_PASSWORD,
			nickserv_conf.password_min_length,
			nickserv_conf.password_min_digits,
			nickserv_conf.password_min_upper,
			nickserv_conf.password_min_lower);
	return 0;
    }

    cryptpass(passwd, crypted);
    hi = register_handle(handle, crypted);
    hi->masks = alloc_string_list(1);
    hi->users = NULL;
    hi->registered = now;
    hi->lastseen = now;
    hi->flags = HI_DEFAULT_FLAGS;
    if (settee) set_user_handle_info(settee, hi);

    if (nickserv_conf.disable_nicks || (user != settee)) {
        if (user == settee) {
            nickserv_notice(user, NSMSG_REGISTER_H_SUCCESS);
        } else {
            nickserv_notice(user, NSMSG_OREGISTER_H_SUCCESS);
        }
    } else {
	struct nick_info *ni = dict_find(nickserv_nick_dict, user->nick, NULL);
	if (ni) {
            nickserv_notice(user, NSMSG_PARTIAL_REGISTER);
	} else {
	    register_nick(user->nick, hi);
            nickserv_notice(user, NSMSG_REGISTER_HN_SUCCESS);
	}
    }
    return hi;
}

static void
nickserv_help(struct userNode *user, const char *topic, int show_aliases)
{
    char *aliases, *resolved = resolve_alias("NickServ", (char *)topic);

    send_help(user, nickserv, nickserv_helpfile, resolved);
    if (show_aliases && (aliases = command_aliases("NickServ", resolved))) {
	char *c = aliases;
	for (c = aliases; *c; c++) *c = toupper(*c);
	nickserv_notice(user, "Aliases: %s", aliases);
	free(aliases);
    }
}

static void
nickserv_free_cookie(void *data)
{
    struct handle_cookie *cookie = data;
    if (cookie->hi) cookie->hi->cookie = NULL;
    if (cookie->data) free(cookie->data);
    free(cookie);
}

static void
nickserv_bake_cookie(struct handle_cookie *cookie)
{
    cookie->hi->cookie = cookie;
    timeq_add(cookie->expires, nickserv_free_cookie, cookie);
}

static void
nickserv_make_cookie(struct userNode *user, struct handle_info *hi, enum cookie_type type, const char *cookie_data)
{
    struct handle_cookie *cookie;
    char subject[128], body[4096], *misc;
    const char *netname;
    int first_time = 0;

    if (hi->cookie) {
        nickserv_notice(user, NSMSG_COOKIE_LIVE, hi->handle);
        return;
    }

    cookie = calloc(1, sizeof(*cookie));
    cookie->hi = hi;
    cookie->type = type;
    cookie->data = cookie_data ? strdup(cookie_data) : NULL;
    cookie->expires = now + nickserv_conf.cookie_timeout;
    inttobase64(cookie->cookie, rand(), 5);
    inttobase64(cookie->cookie+5, rand(), 5);

    netname = conf_network_name();
    if (!netname) netname = "some IRC network";
    subject[0] = 0;

    switch (cookie->type) {
    case ACTIVATION:
        hi->passwd[0] = 0; /* invalidate password */
        nickserv_notice(user, NSMSG_USE_COOKIE_REGISTER);
        snprintf(subject, sizeof(subject), "Account verification for %s", netname);
        snprintf(body, sizeof(body), "This email has been sent to verify that this email address belongs to the person who tried to register an account on %s.  Your cookie is:\n    %s\nTo verify your email address and complete the account registration, log on to %s and type the following command:\n    /msg %s@%s COOKIE %s %s\nIf you did NOT request this account, you do not need to do anything.  Please contact the %s staff if you have questions.", netname, cookie->cookie, netname, nickserv->nick, self->name, hi->handle, cookie->cookie, netname);
        first_time = 1;
        break;
    case PASSWORD_CHANGE:
        nickserv_notice(user, NSMSG_USE_COOKIE_RESETPASS);
        snprintf(subject, sizeof(subject), "Password change verification on %s", netname);
        snprintf(body, sizeof(body), "This email has been sent to verify that you wish to change the password on your account %s.  Your cookie is:\n    %s\nTo complete the password change, log on to %s and type the following command:\n    /msg %s@%s COOKIE %s %s\nIf you did NOT request your password to be changed, you do not need to do anything.  Please contact the %s staff if you have questions.", hi->handle, cookie->cookie, netname, nickserv->nick, self->name, hi->handle, cookie->cookie, netname);
        break;
    case EMAIL_CHANGE:
        misc = hi->email_addr;
        hi->email_addr = cookie->data;
        if (misc) {
            nickserv_notice(user, NSMSG_USE_COOKIE_EMAIL_2);
            snprintf(subject, sizeof(subject), "Email address change verification for %s", netname);
            snprintf(body, sizeof(body), "This email has been sent to verify that your email address belongs to the same person as the account %s on %s.  The SECOND HALF of your cookie is:\n    %.*s\nTo verify your address as associated with this account, log on to %s and type the following command:\n    /msg %s@%s COOKIE %s ?????%.*s\n(Replace the ????? with the FIRST HALF of the cookie, as sent to your OLD email address.)\nIf you did NOT request this email address to be associated with this account, you do not need to do anything.  Please contact the %s staff if you have questions.", hi->handle, netname, COOKIELEN/2, cookie->cookie+COOKIELEN/2, netname, nickserv->nick, self->name, hi->handle, COOKIELEN/2, cookie->cookie+COOKIELEN/2, netname);
            sendmail(nickserv, hi, subject, body, 1);
            snprintf(subject, sizeof(subject), "Email address change verification for %s", netname);
            snprintf(body, sizeof(body), "This email has been sent to verify that you want to change your email for %s on %s from this address to %s.  The FIRST HALF of your cookie is:\n    %.*s\nTo verify your new address as associated with this account, log on to %s and type the following command:\n    /msg %s@%s COOKIE %s %.*s?????\n(Replace the ????? with the SECOND HALF of the cookie, as sent to your NEW email address.)\nIf you did NOT request this change of email address, you do not need to do anything.  Please contact the %s staff if you have questions.", hi->handle, netname, hi->email_addr, COOKIELEN/2, cookie->cookie, netname, nickserv->nick, self->name, hi->handle, COOKIELEN/2, cookie->cookie, netname);
        } else {
            nickserv_notice(user, NSMSG_USE_COOKIE_EMAIL_1);
            snprintf(subject, sizeof(subject), "Email address verification for %s", netname);
            snprintf(body, sizeof(body), "This email has been sent to verify that this address belongs to the same person as %s on %s.  Your cookie is:\n    %s\nTo verify your address as associated with this account, log on to %s and type the following command:\n    /msg %s@%s COOKIE %s %s\nIf you did NOT request this email address to be associated with this account, you do not need to do anything.  Please contact the %s staff if you have questions.", hi->handle, netname, cookie->cookie, netname, nickserv->nick, self->name, hi->handle, cookie->cookie, netname);
            sendmail(nickserv, hi, subject, body, 1);
            subject[0] = 0;
        }
        hi->email_addr = misc;
        break;
    case ALLOWAUTH:
        snprintf(subject, sizeof(subject), "Authentication allowed for %s", netname);
        snprintf(body, sizeof(body), "This email has been sent to let you authenticate (auth) to account %s on %s.  Your cookie is:\n    %s\nTo auth to that account, log on to %s and type the following command:\n    /msg %s@%s COOKIE %s %s\nIf you did NOT request this authorization, you do not need to do anything.  Please contact the %s staff if you have questions.", hi->handle, netname, cookie->cookie, netname, nickserv->nick, self->name, hi->handle, cookie->cookie, netname);
        break;
    default:
        log(NS_LOG, LOG_ERROR, "Bad cookie type %d in nickserv_make_cookie\n", cookie->type);
        break;
    }
    if (subject[0]) sendmail(nickserv, hi, subject, body, first_time);
    nickserv_bake_cookie(cookie);
}

static void
nickserv_eat_cookie(struct handle_cookie *cookie)
{
    cookie->hi->cookie = NULL;
    timeq_del(cookie->expires, nickserv_free_cookie, cookie, 0);
    nickserv_free_cookie(cookie);
}

static void
nickserv_free_email_addr(void *data)
{
    handle_info_list_clean(data);
    free(data);
}

static void
nickserv_set_email_addr(struct handle_info *hi, const char *new_email_addr)
{
    struct handle_info_list *hil;
    /* Remove from old handle_info_list ... */
    if (hi->email_addr && (hil = dict_find(nickserv_email_dict, hi->email_addr, 0))) {
        handle_info_list_remove(hil, hi);
        if (!hil->used) dict_remove(nickserv_email_dict, hil->tag);
        hi->email_addr = NULL;
    }
    /* Add to the new list.. */
    if (new_email_addr) {
        if (!(hil = dict_find(nickserv_email_dict, new_email_addr, 0))) {
            hil = calloc(1, sizeof(*hil));
            hil->tag = strdup(new_email_addr);
            handle_info_list_init(hil);
            dict_insert(nickserv_email_dict, hil->tag, hil);
        }
        handle_info_list_append(hil, hi);
        hi->email_addr = hil->tag;
    }
}

static NICKSERV_FUNC(cmd_register)
{
    struct handle_info *hi;
    const char *email_addr;
    struct userNode *settee;

    if (!IsOper(user) && !dict_size(nickserv_handle_dict)) {
	/* Require the first handle registered to belong to someone +o. */
	nickserv_notice(user, NSMSG_REQUIRE_OPER);
	return 0;
    }

    if (user->handle_info) {
        nickserv_notice(user, NSMSG_USE_RENAME, user->handle_info->handle);
        return 0;
    }

    NICKSERV_MIN_PARMS((unsigned)3 + nickserv_conf.email_required);

    if (!is_valid_handle(argv[1])) {
        nickserv_notice(user, NSMSG_BAD_HANDLE, argv[1]);
        return 0;
    }

    if ((argc >= 4) && nickserv_conf.email_enabled) {
        struct handle_info_list *hil;
        const char *str;

        /* Remember email address. */
        email_addr = argv[3];

        /* Check that the email address looks valid.. */
        if (!is_valid_email_addr(email_addr)) {
            nickserv_notice(user, NSMSG_BAD_EMAIL_ADDR);
            return 0;
        }

        /* .. and that we are allowed to send to it. */
        if ((str = sendmail_prohibited_address(email_addr))) {
            nickserv_notice(user, NSMSG_EMAIL_PROHIBITED, email_addr, str);
            return 0;
        }

        /* If we do email verify, make sure we don't spam the address. */
        if ((hil = dict_find(nickserv_email_dict, email_addr, NULL))) {
            unsigned int nn;
            for (nn=0; nn<hil->used; nn++) {
                if (hil->list[nn]->cookie) {
                    nickserv_notice(user, NSMSG_EMAIL_UNACTIVATED);
                    return 0;
                }
            }
            if (hil->used >= nickserv_conf.handles_per_email) {
                nickserv_notice(user, NSMSG_EMAIL_OVERUSED);
                return 0;
            }
        }

        /* Don't auth them against handle. */
        settee = NULL;
    } else {
        email_addr = NULL;
        settee = user;
    }

    if ((hi = nickserv_register(user, settee, argv[1], argv[2]))) {
	log(NS_LOG, LOG_INFO, "Registered account '%s' to "IDENT_FORMAT".\n", argv[1], IDENT_DATA(user));

        /* Add any masks they should get. */
        if (nickserv_conf.default_hostmask) {
            string_list_append(hi->masks, strdup(nickserv_conf.default_hostmask));
        } else {
            string_list_append(hi->masks, generate_hostmask(user, GENMASK_OMITNICK));
            if (user->ip != 0) {
                string_list_append(hi->masks, generate_hostmask(user, GENMASK_OMITNICK|GENMASK_BYIP));
            }
        }

        /* If they're the first to register, give them level 1000. */
	if (dict_size(nickserv_handle_dict) == 1) {
	    hi->opserv_level = 1000;
	    nickserv_notice(user, NSMSG_ROOT_HANDLE, argv[1]);
	}

        /* Set their email address. */
        if (email_addr) nickserv_set_email_addr(hi, email_addr);

        /* If they need to do email verification, tell them. */
        if (!settee) nickserv_make_cookie(user, hi, ACTIVATION, hi->passwd);

	return 1;
    } else {
	return 0;
    }
}

static NICKSERV_FUNC(cmd_oregister)
{
    char *mask;
    struct userNode *settee;
    struct handle_info *hi;

    NICKSERV_MIN_PARMS(4);

    if (!is_valid_handle(argv[1])) {
        nickserv_notice(user, NSMSG_BAD_HANDLE, argv[1]);
        return 0;
    }

    if (strchr(argv[3], '@')) {
	mask = canonicalize_hostmask(strdup(argv[3]));
	if (argc > 4) {
	    settee = GetUserH(argv[4]);
	    if (!settee) {
		nickserv_notice(user, MSG_NICK_UNKNOWN, argv[4]);
		return 0;
	    }
	} else {
	    settee = NULL;
	}
    } else if ((settee = GetUserH(argv[3]))) {
	mask = generate_hostmask(settee, GENMASK_OMITNICK);
    } else {
	nickserv_notice(user, NSMSG_REGISTER_BAD_NICKMASK, argv[3]);
	return 0;
    }

    if ((hi = nickserv_register(user, settee, argv[1], argv[2]))) {
        string_list_append(hi->masks, mask);
	if (settee) {
	    log(NS_LOG, LOG_INFO, "%s registered account '%s' to "IDENT_FORMAT".\n", user->nick, argv[1], IDENT_DATA(settee));
	} else {
	    log(NS_LOG, LOG_INFO, "%s registered account '%s' (no current user).\n", user->nick, argv[1]);
	}
	return 1;
    } else {
	return 0;
    }
}

static NICKSERV_FUNC(cmd_handleinfo)
{
    char buff[400];
    unsigned int i, pos=0, herelen;
    struct userNode *target, *next_un;
    struct handle_info *hi;

    if (argc < 2) {
        if (!(hi = user->handle_info)) {
            nickserv_notice(user, NSMSG_MUST_AUTH);
            return 0;
        }
    } else if (!(hi = nickserv_get_handle_info(user, argv[1]))) {
        return 0;
    }

    log(NS_LOG, LOG_OTHER, "Responded to inquiry about account '%s' from nick %s.\n", hi->handle, user->nick);
    nickserv_notice(user, NSMSG_HANDLEINFO_ON, hi->handle);
    nickserv_notice(user, NSMSG_HANDLEINFO_REGGED, ctime(&hi->registered));

    if (!hi->users) {
	intervalString(buff, now - hi->lastseen);
	nickserv_notice(user, NSMSG_HANDLEINFO_LASTSEEN, buff);
    } else {
	nickserv_notice(user, NSMSG_HANDLEINFO_LASTSEEN, NSMSG_RIGHT_NOW);
    }

    nickserv_notice(user, NSMSG_HANDLEINFO_INFOLINE, (hi->infoline ? hi->infoline : NSMSG_NONE));
    if (HANDLE_FLAGGED(hi, FROZEN)) nickserv_notice(user, NSMSG_HANDLEINFO_VACATION);

    if (hi != user->handle_info) {
        if (!is_valid_oper(user, 0, 1)) return 1;
        /* Normally, since we check for is_valid_oper, we'd want to log this
         * separately.  But we've already logged it above. */
    }

    if (nickserv_conf.email_enabled) {
        nickserv_notice(user, NSMSG_HANDLEINFO_EMAIL_ADDR, visible_email_addr(user, hi));
    }

    if (hi->flags) {
	unsigned long flen = 1;
	char *flags = alloca(strlen(handle_flags) + 2);
	flags[0] = '+';
	for (i=0, flen=1; handle_flags[i]; i++) {
	    if (hi->flags & 1 << i) flags[flen++] = handle_flags[i];
	}
	flags[flen] = 0;
	nickserv_notice(user, NSMSG_HANDLEINFO_FLAGS, flags);
    } else {
	nickserv_notice(user, NSMSG_HANDLEINFO_FLAGS, NSMSG_NONE);
    }

    if (HANDLE_FLAGGED(hi, SUPPORT_HELPER)
        || HANDLE_FLAGGED(hi, NETWORK_HELPER)
        || (hi->opserv_level > 0)) {
        nickserv_notice(user, NSMSG_HANDLEINFO_EPITHET, (hi->epithet ? hi->epithet : NSMSG_NONE));
    }

    if (hi->last_authed_host[0]) {
        nickserv_notice(user, NSMSG_HANDLEINFO_LAST_HOST, hi->last_authed_host);
    } else {
        nickserv_notice(user, NSMSG_HANDLEINFO_LAST_HOST, NSMSG_UNKNOWN);
    }

    if (nickserv_conf.disable_nicks) {
	/* nicks disabled; don't show anything about registered nicks */
    } else if (hi->nicks) {
	struct nick_info *ni, *next_ni;
	for (ni = hi->nicks; ni; ni = next_ni) {
	    herelen = strlen(ni->nick);
	    if (pos + herelen + 1 > ArrayLength(buff)) {
		next_ni = ni;
		goto print_nicks_buff;
	    } else {
		next_ni = ni->next;
	    }
	    memcpy(buff+pos, ni->nick, herelen);
	    pos += herelen; buff[pos++] = ' ';
	    if (!next_ni) {
	      print_nicks_buff:
		buff[pos-1] = 0;
		nickserv_notice(user, NSMSG_HANDLEINFO_NICKS, buff);
		pos = 0;
	    }
	}
    } else {
	nickserv_notice(user, NSMSG_HANDLEINFO_NICKS, NSMSG_NONE);
    }

    if (hi->masks->used) {
        for (i=0; i < hi->masks->used; i++) {
            herelen = strlen(hi->masks->list[i]);
            if (pos + herelen + 1 > ArrayLength(buff)) {
                i--; goto print_mask_buff;
            }
            memcpy(buff+pos, hi->masks->list[i], herelen);
            pos += herelen; buff[pos++] = ' ';
            if (i+1 == hi->masks->used) {
              print_mask_buff:
                buff[pos-1] = 0;
                nickserv_notice(user, NSMSG_HANDLEINFO_MASKS, buff);
                pos = 0;
            }
        }
    } else {
        nickserv_notice(user, NSMSG_HANDLEINFO_MASKS, "(none)");
    }

    if (hi->channels) {
	struct chanList *channel, *next;
	extern char accessChars[];
	char *name;

	for (channel = hi->channels; channel; channel = next) {
	    next = channel->next;
	    if(IsSuspended(channel->channel)) {
		name = channel->channel->suspended->name;
	    } else {
		name = channel->channel->channel->name;
	    }
	    herelen = strlen(name);
	    if (pos + herelen + 2 > ArrayLength(buff)) {
		next = channel; goto print_chans_buff;
	    }
	    buff[pos++] = accessChars[channel->user->access];
	    memcpy(buff+pos, name, herelen);
	    pos += herelen; buff[pos++] = ' ';
	    if (next == NULL) {
	      print_chans_buff:
		buff[pos-1] = 0;
		nickserv_notice(user, NSMSG_HANDLEINFO_CHANNELS, buff);
		pos = 0;
	    }
	}
    } else {
	nickserv_notice(user, NSMSG_HANDLEINFO_CHANNELS, NSMSG_NONE);
    }

    for (target = hi->users; target; target = next_un) {
	herelen = strlen(target->nick);
	if (pos + herelen + 1 > ArrayLength(buff)) {
	    next_un = target;
	    goto print_cnick_buff;
	} else {
	    next_un = target->next_authed;
	}
	memcpy(buff+pos, target->nick, herelen);
	pos += herelen; buff[pos++] = ' ';
	if (!next_un) {
	  print_cnick_buff:
	    buff[pos-1] = 0;
	    nickserv_notice(user, NSMSG_HANDLEINFO_CURRENT, buff);
	    pos = 0;
	}
    }
    return 1;
}

static NICKSERV_FUNC(cmd_userinfo)
{
    struct userNode *target;

    NICKSERV_MIN_PARMS(2);
    if (!(target = GetUserH(argv[1]))) {
	nickserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	return 0;
    }
    if (target->handle_info) {
	nickserv_notice(user, NSMSG_USERINFO_AUTHED_AS, target->nick, target->handle_info->handle);
    } else {
	nickserv_notice(user, NSMSG_USERINFO_NOT_AUTHED, target->nick);
    }
    return 1;
}

static NICKSERV_FUNC(cmd_nickinfo)
{
    struct nick_info *ni;

    NICKSERV_MIN_PARMS(2);
    if (!(ni = get_nick_info(argv[1]))) {
	nickserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	return 0;
    }
    nickserv_notice(user, NSMSG_NICKINFO_OWNER, ni->nick, ni->owner->handle);
    return 1;
}

static NICKSERV_FUNC(cmd_rename_handle)
{
    struct handle_info *hi;
    unsigned int nn;

    NICKSERV_MIN_PARMS(3);
    if (!(hi = get_victim_oper(user, argv[1]))) return 0;
    if (!is_valid_handle(argv[2])) {
        nickserv_notice(user, NSMSG_FAIL_RENAME, argv[1], argv[2]);
        return 0;
    }
    /* just to be consistant with IRC */
    if (isdigit(argv[2][0])) {
        nickserv_notice(user, NSMSG_NO_DIGIT);
        return 0;
    }
    if (get_handle_info(argv[2])) {
        nickserv_notice(user, NSMSG_HANDLE_EXISTS, argv[2]);
        return 0;
    }
    for (nn=0; nn<rf_list_used; nn++) {
        rf_list[nn](hi, argv[2]);
    }
    dict_remove2(nickserv_handle_dict, hi->handle, 1);
    safestrncpy(hi->handle, argv[2], sizeof(hi->handle));
    dict_insert(nickserv_handle_dict, hi->handle, hi);
    if (argv[1][0] == '*') {
	nickserv_notice(user, NSMSG_HANDLE_CHANGED, argv[1]+1, argv[2]);
    } else {
	nickserv_notice(user, NSMSG_HANDLE_CHANGED, argv[1], argv[2]);
    }
    return 1;
}

static NICKSERV_FUNC(cmd_auth)
{
    int res;
    struct handle_info *hi;
    const char *passwd;

    if (user->handle_info) {
        nickserv_notice(user, NSMSG_ALREADY_AUTHED, user->handle_info->handle);
        return 0;
    }
    if (argc == 3) {
	hi = dict_find(nickserv_handle_dict, argv[1], NULL);
	passwd = argv[2];
    } else if (argc == 2) {
        if (nickserv_conf.disable_nicks) {
            if (!(hi = get_handle_info(user->nick))) {
                nickserv_notice(user, NSMSG_HANDLE_NOT_FOUND);
                return 0;
            }
        } else {
            /* try to look up their handle from their nick */
            struct nick_info *ni;
            ni = get_nick_info(user->nick);
            if (!ni) {
                nickserv_notice(user, NSMSG_NICK_NOT_REGISTERED, user->nick);
                return 0;
            }
            hi = ni->owner;
        }
        passwd = argv[1];
    } else {
	nickserv_notice(user, MSG_MISSING_PARAMS, argv[0]);
	NICKSERV_SYNTAX();
	return 0;
    }
    if (!hi) {
	nickserv_notice(user, NSMSG_HANDLE_NOT_FOUND);
	return 0;
    }
    res = valid_user_for(user, hi);
    if (res == 0) {
	nickserv_notice(user, NSMSG_HOSTMASK_INVALID, hi->handle);
	log(NS_LOG, LOG_OTHER, IDENT_FORMAT" did not have a matching host for account '%s'.\n", IDENT_DATA(user), hi->handle);
	return 0;
    } else if (res == 2) {
	dict_remove(nickserv_allow_auth_dict, user->nick);
    }
    if (!checkpass(passwd, hi->passwd)) {
	nickserv_notice(user, NSMSG_PASSWORD_INVALID);
	log(NS_LOG, LOG_OTHER, IDENT_FORMAT" had the wrong password for account '%s'.\n", IDENT_DATA(user), hi->handle);
        if (nickserv_conf.autogag_enabled) {
            if (!user->auth_policer) {
                user->auth_policer = policer_new(nickserv_conf.auth_policer_params, now);
            }
            if (!policer_conforms(user->auth_policer, now, 1.0)) {
                char *hostmask;
                hostmask = generate_hostmask(user, GENMASK_STRICT_HOST|GENMASK_BYIP);
                log(NS_LOG, LOG_OTHER, "%s auto-gagged for repeated password guessing.\n", hostmask);
                gag_create(hostmask, nickserv->nick, "Repeated password guessing.", now+nickserv_conf.autogag_duration);
                free(hostmask);
            }
        }
	return 0;
    }
    if (HANDLE_FLAGGED(hi, SUSPENDED)) {
	nickserv_notice(user, NSMSG_HANDLE_SUSPENDED);
	log(NS_LOG, LOG_OTHER, IDENT_FORMAT" tried to auth to suspended account '%s'.\n", IDENT_DATA(user), hi->handle);
	return 0;
    }

    if (nickserv_conf.email_required && !hi->email_addr) {
        nickserv_notice(user, NSMSG_PLEASE_SET_EMAIL);
    }
    if (!is_secure_password(hi->handle, passwd)) {
        nickserv_notice(user, NSMSG_WEAK_PASSWORD);
    }
    if (hi->passwd[0] != '$') cryptpass(passwd, hi->passwd);

    set_user_handle_info(user, hi);
    nickserv_notice(user, NSMSG_AUTH_SUCCESS);

    log(NS_LOG, LOG_INFO, IDENT_FORMAT" was authenticated to account '%s'.\n", IDENT_DATA(user), hi->handle);

#ifdef USE_IRCO
    {
        char modecmd[512];
        int len;

        snprintf(modecmd, sizeof(modecmd)-2, "%s MODE %s +", self->numeric, user->numeric);
        len = strlen(modecmd);
        if (is_valid_oper(user, 900, 1)) modecmd[len++] = 'n';
        modecmd[len++] = 'r';
        modecmd[len] = 0;
        irc_raw(modecmd);
    }
#endif

    return 1;
}

static NICKSERV_FUNC(cmd_allowauth)
{
    struct userNode *target;
    struct handle_info *hi;

    NICKSERV_MIN_PARMS(2);
    if (!(target = GetUserH(argv[1]))) {
	nickserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	return 0;
    }
    if (argc == 2) hi = NULL;
    else if (!(hi = get_handle_info(argv[2]))) {
        nickserv_notice(user, MSG_HANDLE_UNKNOWN, argv[2]);
        return 0;
    }
    if (hi) {
	if (hi->opserv_level > user->handle_info->opserv_level) {
	    nickserv_notice(user, MSG_USER_OUTRANKED, hi->handle);
	    return 0;
	}
	dict_insert(nickserv_allow_auth_dict, strdup(target->nick), hi);
	nickserv_notice(user, NSMSG_AUTH_ALLOWED, target->nick, hi->handle);
	nickserv_notice(target, NSMSG_AUTH_ALLOWED_MSG, hi->handle, hi->handle);
	log(NS_LOG, LOG_INFO, "%s allowed "IDENT_FORMAT" to authenticate to account '%s'.\n", user->handle_info->handle, IDENT_DATA(target), hi->handle);
    } else {
	if (dict_remove(nickserv_allow_auth_dict, target->nick)) {
	    nickserv_notice(user, NSMSG_AUTH_NORMAL_ONLY, target->nick);
	    log(NS_LOG, LOG_INFO, "%s removed "IDENT_FORMAT"'s authorization to auth without hostmask.\n", user->handle_info->handle, IDENT_DATA(target));
	} else {
	    nickserv_notice(user, NSMSG_AUTH_UNSPECIAL, target->nick);
	}
    }
    return 1;
}

static NICKSERV_FUNC(cmd_authcookie)
{
    struct handle_info *hi;

    NICKSERV_MIN_PARMS(2);
    if (!(hi = get_handle_info(argv[1]))) {
        nickserv_notice(user, MSG_HANDLE_UNKNOWN, argv[1]);
        return 0;
    }
    if (!hi->email_addr) {
        nickserv_notice(user, MSG_SET_EMAIL_ADDR);
        return 0;
    }
    nickserv_make_cookie(user, hi, ALLOWAUTH, NULL);
    nickserv_notice(user, NSMSG_USE_COOKIE_AUTH);
    return 1;
}

static NICKSERV_FUNC(cmd_delcookie)
{
    struct handle_info *hi;

    (void)argc; (void)argv;
    hi = user->handle_info;
    if (hi->cookie) {
        switch (hi->cookie->type) {
        case ACTIVATION:
        case EMAIL_CHANGE:
            nickserv_notice(user, NSMSG_MUST_TIME_OUT);
            break;
        default:
            nickserv_eat_cookie(hi->cookie);
            nickserv_notice(user, NSMSG_ATE_COOKIE);
            break;
        }
        return 1;
    } else {
        nickserv_notice(user, NSMSG_NO_COOKIE);
        return 0;
    }
}

static NICKSERV_FUNC(cmd_resetpass)
{
    struct handle_info *hi;
    char crypted[MD5_CRYPT_LENGTH];

    NICKSERV_MIN_PARMS(3);
    if (!(hi = get_handle_info(argv[1]))) {
        nickserv_notice(user, MSG_HANDLE_UNKNOWN, argv[1]);
        return 0;
    }
    if (!hi->email_addr) {
        nickserv_notice(user, MSG_SET_EMAIL_ADDR);
        return 0;
    }
    cryptpass(argv[2], crypted);
    nickserv_make_cookie(user, hi, PASSWORD_CHANGE, crypted);
    return 1;
}

static NICKSERV_FUNC(cmd_cookie)
{
    struct handle_info *hi;
    const char *cookie;

    if ((argc == 2) && (hi = user->handle_info) && hi->cookie && (hi->cookie->type == EMAIL_CHANGE)) {
        hi = user->handle_info;
        cookie = argv[1];
    } else {
        NICKSERV_MIN_PARMS(3);
        if (!(hi = get_handle_info(argv[1]))) {
            nickserv_notice(user, MSG_HANDLE_UNKNOWN, argv[1]);
            return 0;
        }
        cookie = argv[2];
    }

    if (!hi->cookie) {
        nickserv_notice(user, NSMSG_NO_COOKIE);
        return 0;
    }

    if (strcmp(cookie, hi->cookie->cookie)) {
        nickserv_notice(user, NSMSG_BAD_COOKIE);
        return 0;
    }

    switch (hi->cookie->type) {
    case ACTIVATION:
        safestrncpy(hi->passwd, hi->cookie->data, sizeof(hi->passwd));
        set_user_handle_info(user, hi);
        nickserv_notice(user, NSMSG_HANDLE_ACTIVATED);
        break;
    case PASSWORD_CHANGE:
        set_user_handle_info(user, hi);
        safestrncpy(hi->passwd, hi->cookie->data, sizeof(hi->passwd));
        nickserv_notice(user, NSMSG_PASSWORD_CHANGED);
        break;
    case EMAIL_CHANGE:
        nickserv_set_email_addr(hi, hi->cookie->data);
        nickserv_notice(user, NSMSG_EMAIL_CHANGED);
        break;
    case ALLOWAUTH:
        set_user_handle_info(user, hi);
        nickserv_notice(user, NSMSG_AUTH_SUCCESS);
        break;
    default:
        nickserv_notice(user, NSMSG_BAD_COOKIE_TYPE, hi->cookie->type);
        log(NS_LOG, LOG_ERROR, "Bad cookie type %d for account %s.\n", hi->cookie->type, hi->handle);
        break;
    }

    nickserv_eat_cookie(hi->cookie);

    return 1;
}

static NICKSERV_FUNC(cmd_regnick)
{
    const unsigned char *nick;
    struct nick_info *ni;
    struct userNode *target;
    unsigned n;

    /* count their nicks, see if it's too many */
    for (n=0,ni=user->handle_info->nicks; ni; n++,ni=ni->next) ;
    if (n >= nickserv_conf.nicks_per_handle) {
	nickserv_notice(user, NSMSG_TOO_MANY_NICKS);
	log(NS_LOG, LOG_OTHER, "Account '%s' had too many nicks registered to it.\n", user->handle_info->handle);
	return 0;
    }
    if (is_valid_oper(user, 0, 0)) {
	nick = (argc < 2) ? user->nick : (char*)argv[1];
	if (!is_valid_nick(nick)) {
	    nickserv_notice(user, NSMSG_BAD_NICK, nick);
	    return 0;
	}
	if (strlen(nick) > NICKLEN) {
	    nickserv_notice(user, NSMSG_NICK_TOO_LONG);
	    return 0;
	}
	target = GetUserH(nick);
	if (target && target != user) {
	    nickserv_notice(user, NSMSG_NICK_IN_USE, nick);
	    log(NS_LOG, LOG_OTHER, "Account '%s' tried to register someone else's nick.\n", user->handle_info->handle);
	    return 0;
	}
    } else {
	nick = user->nick;
        if (!is_valid_nick(nick)) {
            nickserv_notice(user, NSMSG_BAD_NICK, nick);
            return 0;
        }
	/* just to be consistant with IRC */
	if (isdigit(nick[0])) {
            nickserv_notice(user, NSMSG_NO_DIGIT);
            return 0;
	}
	log(NS_LOG, LOG_OTHER, "Account '%s' tried to register someone else's nick.\n", user->handle_info->handle);
    }
    ni = dict_find(nickserv_nick_dict, nick, NULL);
    if (ni) {
	nickserv_notice(user, NSMSG_NICK_EXISTS, nick);
	return 0;
    }
    register_nick(nick, user->handle_info);
    nickserv_notice(user, NSMSG_REGNICK_SUCCESS, nick);
    log(NS_LOG, LOG_OTHER, "Account '%s' registered nick '%s'.\n", user->handle_info->handle, nick);
    return 1;
}

static NICKSERV_FUNC(cmd_pass)
{
    struct handle_info *hi;
    const char *old_pass, *new_pass;

    NICKSERV_MIN_PARMS(3);
    hi = user->handle_info;
    old_pass = argv[1];
    new_pass = argv[2];
    if (!is_secure_password(hi->handle, new_pass)) {
	nickserv_notice(user, NSMSG_INSECURE_PASSWORD,
			nickserv_conf.password_min_length,
			nickserv_conf.password_min_digits,
			nickserv_conf.password_min_upper,
			nickserv_conf.password_min_lower);
	return 0;
    }
    if (!checkpass(old_pass, hi->passwd)) {
	nickserv_notice(user, NSMSG_PASSWORD_INVALID);
	return 0;
    }
    cryptpass(new_pass, hi->passwd);
    nickserv_notice(user, NSMSG_PASS_SUCCESS);
    log(NS_LOG, LOG_OTHER, IDENT_FORMAT" changed passwords for their account '%s'.\n", IDENT_DATA(user), hi->handle);
    return 1;
}

static int
nickserv_addmask(struct userNode *user, struct handle_info *hi, const char *mask)
{
    unsigned int i;
    char *new_mask = canonicalize_hostmask(strdup(mask));
    for (i=0; i<hi->masks->used; i++) {
        if (!strcasecmp(new_mask, hi->masks->list[i])) {
            nickserv_notice(user, NSMSG_ADDMASK_ALREADY, new_mask);
            free(new_mask);
            return 0;
        }
    }
    string_list_append(hi->masks, new_mask);
    nickserv_notice(user, NSMSG_ADDMASK_SUCCESS, new_mask);
    log(NS_LOG, LOG_OTHER, "'%s' added new mask '%s' to account '%s'.\n", user->nick, new_mask, hi->handle);
    return 1;
}

static NICKSERV_FUNC(cmd_addmask)
{
    if (argc < 2) {
        char *mask = generate_hostmask(user, GENMASK_OMITNICK);
        int res = nickserv_addmask(user, user->handle_info, mask);
        free(mask);
        return res;
    } else {
        if (!is_gline(argv[1])) {
            nickserv_notice(user, NSMSG_MASK_INVALID, argv[1]);
            return 0;
        }
        return nickserv_addmask(user, user->handle_info, argv[1]);
    }
}

static NICKSERV_FUNC(cmd_oaddmask)
{
    struct handle_info *hi;

    NICKSERV_MIN_PARMS(3);
    if (!(hi = get_victim_oper(user, argv[1]))) return 0;
    return nickserv_addmask(user, hi, argv[2]);
}

static int
nickserv_delmask(struct userNode *user, struct handle_info *hi, const char *del_mask)
{
    unsigned int i;
    for (i=0; i<hi->masks->used; i++) {
	if (!strcmp(del_mask, hi->masks->list[i])) {
	    char *old_mask = hi->masks->list[i];
	    if (hi->masks->used == 1) {
		nickserv_notice(user, NSMSG_DELMASK_NOTLAST);
		return 0;
	    }
	    hi->masks->list[i] = hi->masks->list[--hi->masks->used];
	    nickserv_notice(user, NSMSG_DELMASK_SUCCESS, old_mask);
	    log(NS_LOG, LOG_OTHER, "'%s' deleted mask '%s' from account '%s'.\n", user->nick, old_mask, hi->handle);
	    free(old_mask);
	    return 1;
	}
    }
    nickserv_notice(user, NSMSG_DELMASK_NOT_FOUND);
    return 0;
}

static NICKSERV_FUNC(cmd_delmask)
{
    NICKSERV_MIN_PARMS(2);
    return nickserv_delmask(user, user->handle_info, argv[1]);
}

static NICKSERV_FUNC(cmd_odelmask)
{
    struct handle_info *hi;
    NICKSERV_MIN_PARMS(3);
    if (!(hi = get_victim_oper(user, argv[1]))) return 0;
    return nickserv_delmask(user, hi, argv[2]);
}

static int
nickserv_apply_flags(struct userNode *user, struct handle_info *hi, const unsigned char *flags)
{
    unsigned int n, add = 1, pos, before, after;
    struct userNode *uNode;

    before = hi->flags & (HI_FLAG_SUPPORT_HELPER|HI_FLAG_NETWORK_HELPER);
    for (n=0; flags[n]; n++) {
	switch (flags[n]) {
	case '+': add = 1; break;
	case '-': add = 0; break;
	default:
	    if (!(pos = handle_inverse_flags[flags[n]])) {
		nickserv_notice(user, NSMSG_INVALID_FLAG, flags[n]);
		return 0;
	    }
            if (user->handle_info->opserv_level < flag_access_levels[pos-1]) {
                /* cheesy avoidance of looking up the flag name.. */
                nickserv_notice(user, MSG_SETTING_PRIVILEGED, "That flag");
                return 0;
            }
	    if (add) hi->flags |= 1 << (pos - 1);
	    else hi->flags &= ~(1 << (pos - 1));
	    break;
	}
    }
    after = hi->flags & (HI_FLAG_SUPPORT_HELPER|HI_FLAG_NETWORK_HELPER);

    /* Strip helping flag if they're only a support helper and not
     * currently in #support. */
    if (HANDLE_FLAGGED(hi, HELPING) && (after == HI_FLAG_SUPPORT_HELPER)) {
        struct chanNode *schannel;
        schannel = chanserv_support_channel();
        for (uNode = hi->users; uNode; uNode = uNode->next_authed) {
            if (GetUserMode(schannel, uNode)) break;
        }
        if (!uNode) HANDLE_CLEAR_FLAG(hi, HELPING);
    }

    if (after && !before) {
        /* Add user to current helper list. */
        for (uNode = hi->users; uNode; uNode = uNode->next_authed) {
            userList_append(&curr_helpers, uNode);
        }
    } else if (!after) {
        /* Remove user from current helper list. */
        for (uNode = hi->users; uNode; uNode = uNode->next_authed) {
            userList_remove(&curr_helpers, uNode);
        }
    }

    return 1;
}

static void
set_list(struct userNode *user, struct handle_info *hi, int override)
{
    option_func_t *opt;
    unsigned int i = 0;
    char *set_display[] = {
    "INFO", "WIDTH", "COLOR", "PRIVMSG", "STYLE", "EMAIL"};

    nickserv_notice(user, NSMSG_SETTING_LIST);

    /* Do this so options are presented in a consistent order. */
    while(i < ArrayLength(set_display))
    {
	if((opt = dict_find(nickserv_opt_dict, set_display[i++], NULL)))
	{
	    opt(user, hi, override, 0, NULL);
	}
    }
}

static NICKSERV_FUNC(cmd_set)
{
    struct handle_info *hi;
    option_func_t *opt;

    hi = user->handle_info;
    if (argc < 2) {
	set_list(user, hi, 0);
	return 1;
    }
    if (!(opt = dict_find(nickserv_opt_dict, argv[1], NULL))) {
	nickserv_notice(user, NSMSG_INVALID_OPTION, argv[1]);
        return 0;
    }
    return opt(user, hi, 0, argc-1, argv+1);
}

static NICKSERV_FUNC(cmd_oset)
{
    struct handle_info *hi;
    option_func_t *opt;

    NICKSERV_MIN_PARMS(2);

    if (!(hi = get_victim_oper(user, argv[1]))) return 0;

    if (argc < 3) {
	set_list(user, hi, 0);
	return 1;
    }

    if (!(opt = dict_find(nickserv_opt_dict, argv[2], NULL))) {
	nickserv_notice(user, NSMSG_INVALID_OPTION, argv[2]);
        return 0;
    }

    return opt(user, hi, 1, argc-2, argv+2);
}

static OPTION_FUNC(opt_info)
{
    (void)override;

    if (argc > 1) {
	if ((argv[1][0] == '*') && (argv[1][1] == 0)) {
            if (hi->infoline) free(hi->infoline);
            hi->infoline = NULL;
	} else {
	    unsplit_string(argv+1, argc-1);
            hi->infoline = strdup(argv[1]);
	}
    }

    nickserv_notice(user, NSMSG_STR_SETTING, "INFO:     ", hi->infoline ? hi->infoline : "None.");
    return 1;
}

static OPTION_FUNC(opt_width)
{
    (void)override;

    if (argc > 1) {
	unsigned int new_width = strtoul(argv[1], NULL, 0);
	hi->screen_width = new_width;
    }

    if (hi->screen_width < MIN_LINE_SIZE) hi->screen_width = MIN_LINE_SIZE;
    else if (hi->screen_width > MAX_LINE_SIZE) hi->screen_width = MAX_LINE_SIZE;

    nickserv_notice(user, NSMSG_NUM_SETTING, "WIDTH:    ", hi->screen_width);
    return 1;
}

static OPTION_FUNC(opt_color)
{
    (void)override;

    if (argc > 1) {
	if (enabled_string(argv[1])) {
	    HANDLE_SET_FLAG(hi, MIRC_COLOR);
	} else if (disabled_string(argv[1])) {
	    HANDLE_CLEAR_FLAG(hi, MIRC_COLOR);
	} else {
	    nickserv_notice(user, MSG_INVALID_BINARY, argv[1]);
	    return 0;
	}
    }

    nickserv_notice(user, NSMSG_STR_SETTING, "COLOR:    ", HANDLE_FLAGGED(hi, MIRC_COLOR) ? "On." : "Off.");
    return 1;
}

static OPTION_FUNC(opt_privmsg)
{
    (void)override;

    if (argc > 1) {
	if (enabled_string(argv[1])) {
	    HANDLE_SET_FLAG(hi, USE_PRIVMSG);
	} else if (disabled_string(argv[1])) {
	    HANDLE_CLEAR_FLAG(hi, USE_PRIVMSG);
	} else {
	    nickserv_notice(user, MSG_INVALID_BINARY, argv[1]);
	    return 0;
	}
    }

    nickserv_notice(user, NSMSG_STR_SETTING, "PRIVMSG:  ", HANDLE_FLAGGED(hi, USE_PRIVMSG) ? "On." : "Off.");
    return 1;
}

static OPTION_FUNC(opt_style)
{
    char *style;
    (void)override;

    if (argc > 1) {
	if (!strcasecmp(argv[1], "Zoot")) {
	    hi->userlist_style = HI_STYLE_ZOOT;
	} else if (!strcasecmp(argv[1], "def")) {
	    hi->userlist_style = HI_STYLE_DEF;
	}
    }

    switch (hi->userlist_style) {
    case HI_STYLE_DEF:
	style = "def";
	break;
    case HI_STYLE_ZOOT:
    default:
	style = "Zoot";
    }

    nickserv_notice(user, NSMSG_STR_SETTING, "STYLE:    ", style);
    return 1;
}

static OPTION_FUNC(opt_password)
{
    if (!override) {
	nickserv_notice(user, NSMSG_USE_CMD_PASS);
	return 0;
    }

    if (argc > 1) {
	cryptpass(argv[1], hi->passwd);
    }

    nickserv_notice(user, NSMSG_STR_SETTING, "PASSWORD: ", "***");
    return 1;
}

static OPTION_FUNC(opt_flags)
{
    char flags[33];

    if (!override) {
	nickserv_notice(user, MSG_SETTING_PRIVILEGED, argv[0]);
	return 0;
    }

    if (argc > 1) {
	nickserv_apply_flags(user, hi, argv[1]);
    }

    if (hi->flags) {
	int i;
	unsigned long flen;

	for (i=flen=0; handle_flags[i]; i++) {
	    if (hi->flags & 1 << i) flags[flen++] = handle_flags[i];
	}
	flags[flen] = 0;
    } else {
	strcpy(flags, "None.");
    }

    nickserv_notice(user, NSMSG_STR_SETTING, "FLAGS:    ", flags);
    return 1;
}

static OPTION_FUNC(opt_email)
{
    if (argc > 1) {
        const char *str;
        if (!is_valid_email_addr(argv[1])) {
            nickserv_notice(user, NSMSG_BAD_EMAIL_ADDR);
            return 0;
        }
        if ((str = sendmail_prohibited_address(argv[1]))) {
            nickserv_notice(user, NSMSG_EMAIL_PROHIBITED, argv[1], str);
            return 0;
        }
        if (!hi->email_addr || strcasecmp(hi->email_addr, argv[1])) {
            if (override) {
                nickserv_set_email_addr(hi, argv[1]);
                nickserv_notice(user, NSMSG_STR_SETTING, "EMAIL:    ", visible_email_addr(user, hi));
            } else {
                nickserv_make_cookie(user, hi, EMAIL_CHANGE, argv[1]);
            }
        } else {
            nickserv_notice(user, NSMSG_EMAIL_SAME);
        }
    } else {
        nickserv_notice(user, NSMSG_STR_SETTING, "EMAIL:    ", visible_email_addr(user, hi));
    }
    return 1;
}

static OPTION_FUNC(opt_level)
{
    if (!override) {
	nickserv_notice(user, MSG_SETTING_PRIVILEGED, argv[0]);
	return 0;
    }

    if (argc > 1) {
	if (is_valid_oper(user, nickserv_conf.modoper_level, 0)) {
	    unsigned int new_level = strtoul(argv[1], NULL, 0);
	    if (user->handle_info->opserv_level < new_level) {
		nickserv_notice(user, NSMSG_OPSERV_LEVEL_BAD);
	    } else {
                log(NS_LOG, LOG_INFO, "Account %s setting oper level for account %s to %d (from %d)\n",
                    user->handle_info->handle, hi->handle, new_level, hi->opserv_level);
		hi->opserv_level = new_level;
	    }
	}
    }

    nickserv_notice(user, NSMSG_NUM_SETTING, "LEVEL:    ", hi->opserv_level);
    return 1;
}

static OPTION_FUNC(opt_epithet)
{
    if (!override) {
        nickserv_notice(user, MSG_SETTING_PRIVILEGED, argv[0]);
        return 0;
    }

    if ((argc > 1) && is_valid_oper(user, nickserv_conf.set_epithet_level, 0)) {
        unsplit_string(argv+1, argc-1);
        log(NS_LOG, LOG_INFO, "%s (%s) setting epithet for account %s to `%s'\n",
            user->nick, user->handle_info->handle, hi->handle, argv[1]);
        if (hi->epithet) free(hi->epithet);
        if ((argv[1][0] == '*') && !argv[1][1]) {
            hi->epithet = NULL;
        } else {
            hi->epithet = strdup(argv[1]);
        }
    }

    nickserv_notice(user, NSMSG_STR_SETTING, "EPITHET:  ", (hi->epithet ? hi->epithet : "None."));
    return 1;
}

static NICKSERV_FUNC(cmd_kill)
{
    struct handle_info *hi;
    struct nick_info *ni;
    struct userNode *victim;

    hi = user->handle_info;
    if (argc < 2) {
	unsigned int count = 0;
	/* if they didn't give any names, try killing indiscriminately */
	for (ni=hi->nicks; ni; ni=ni->next) {
	    victim = GetUserH(ni->nick);
	    /* don't kill nicks who don't exist, or who are the sender */
	    if (!victim || (victim == user)) continue;
	    DelUser(victim, nickserv, true, NSMSG_KILL_MESSAGE);
	    log(NS_LOG, LOG_OTHER, "Account '%s' requested a kill for %s.\n", hi->handle, victim->nick);
	    count++;
	}
	nickserv_notice(user, NSMSG_MULTI_KILL, count);
    } else {
	unsigned int i;
	for (i=1; i<argc; i++) {
	    ni = dict_find(nickserv_nick_dict, argv[i], 0);
	    if (!ni) {
		nickserv_notice(user, NSMSG_UNKNOWN_NICK, argv[i]);
		continue;
	    }
	    if (ni->owner != user->handle_info) {
		nickserv_notice(user, NSMSG_NOT_YOUR_NICK, argv[i]);
		continue;
	    }
	    victim = GetUserH(argv[i]);
	    if (!victim) {
		nickserv_notice(user, NSMSG_NICK_INACTIVE, argv[i]);
		continue;
	    }
	    if (victim == user) {
		nickserv_notice(user, NSMSG_NICK_USER_YOU);
		continue;
	    }
	    DelUser(victim, nickserv, true, NSMSG_KILL_MESSAGE);
	    nickserv_notice(user, NSMSG_SINGLE_KILL, argv[i]);
	    log(NS_LOG, LOG_OTHER, "Account '%s' requested a kill for %s.\n", hi->handle, victim->nick);
	}
    }
    return 1;
}

static NICKSERV_FUNC(cmd_unregnick)
{
    const char *nick;
    struct handle_info *hi;
    struct nick_info *ni;

    hi = user->handle_info;
    nick = (argc < 2) ? user->nick : (const char*)argv[1];
    ni = dict_find(nickserv_nick_dict, nick, NULL);
    if (!ni) {
	nickserv_notice(user, NSMSG_UNKNOWN_NICK, nick);
	return 0;
    }
    if (hi != ni->owner) {
	nickserv_notice(user, NSMSG_NOT_YOUR_NICK, nick);
	log(NS_LOG, LOG_OTHER, "User '%s' tried to unregister nick '%s' belonging to account '%s'.\n", user->nick, ni->nick, ni->owner->handle);
	return 0;
    }
    nickserv_notice(user, NSMSG_UNREGNICK_SUCCESS, ni->nick);
    log(NS_LOG, LOG_INFO, "User '%s' (account '%s') unregistered nick %s.\n", user->nick, hi->handle, ni->nick);
    delete_nick(ni);
    return 1;
}

static NICKSERV_FUNC(cmd_ounregnick)
{
    struct nick_info *ni;

    NICKSERV_MIN_PARMS(2);
    if (!(ni = get_nick_info(argv[1]))) {
	nickserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	return 0;
    }
    if (ni->owner->opserv_level >= user->handle_info->opserv_level) {
	nickserv_notice(user, MSG_USER_OUTRANKED, ni->nick);
	return 0;
    }
    nickserv_notice(user, NSMSG_UNREGNICK_SUCCESS, ni->nick);
    log(NS_LOG, LOG_INFO, "User '%s' unregistered nick '%s' belonging to account '%s'.\n", user->nick, ni->nick, ni->owner->handle);
    delete_nick(ni);
    return 1;
}

static NICKSERV_FUNC(cmd_unregister)
{
    struct handle_info *hi;

    NICKSERV_MIN_PARMS(2);
    hi = user->handle_info;
    if (checkpass(argv[1], hi->passwd)) {
        log(NS_LOG, LOG_INFO, "%s unregistered their account %s.\n", user->nick, hi->handle);
        nickserv_unregister_handle(hi, user);
    } else {
	log(NS_LOG, LOG_INFO, "Account '%s' tried to unregister with the wrong password.\n", hi->handle);
	nickserv_notice(user, NSMSG_PASSWORD_INVALID);
    }
    return 0;
}

static NICKSERV_FUNC(cmd_ounregister)
{
    struct handle_info *hi;

    NICKSERV_MIN_PARMS(2);
    if (!(hi = get_victim_oper(user, argv[1]))) return 0;
    nickserv_unregister_handle(hi, user);
    return 0;
}

static NICKSERV_FUNC(cmd_status)
{
    char interval[INTERVALLEN];
    (void)argc; (void)argv; /* ignore them!  ignore them all! */

    if (nickserv_conf.disable_nicks) {
        nickserv_notice(user, NSMSG_GLOBAL_STATS_NONICK,
                        dict_size(nickserv_handle_dict));
    } else {
        if (user->handle_info) {
            int cnt=0;
            struct nick_info *ni;
            for (ni=user->handle_info->nicks; ni; ni=ni->next) cnt++;
            nickserv_notice(user, NSMSG_HANDLE_STATS, cnt);
        } else {
            nickserv_notice(user, NSMSG_HANDLE_NONE);
        }
        nickserv_notice(user, NSMSG_GLOBAL_STATS,
                        dict_size(nickserv_handle_dict),
                        dict_size(nickserv_nick_dict));
    }
    if (IsOper(user)) {
        if (lastwrite) {
  	    intervalString(interval, now - lastwrite);
	    nickserv_notice(user, NSMSG_LAST_WRITE, interval);
        } else {
	    nickserv_notice(user, NSMSG_NO_DB_WRITE);
        }
    }
    return 1;
}

static NICKSERV_FUNC(cmd_vacation)
{
    (void)argc; (void)argv;
    HANDLE_SET_FLAG(user->handle_info, FROZEN);
    nickserv_notice(user, NSMSG_ON_VACATION);
    return 1;
}

static NICKSERV_FUNC(cmd_version)
{
    (void)argc; (void)argv;

    nickserv_notice(user, MSG_VERSION);
    nickserv_notice(user, NSMSG_VERSION_ID);

    return 1;
}

static int
nickserv_write_handle(const char *key, struct handle_info *hi, FILE *f)
{
    struct string_list *slist;
    struct nick_info *ni;
    struct record_data *rd;
    char num[12], flags[33];
    unsigned int i, flen;

    /* create the object */
    dict_t obj = alloc_object();
    /* set the password */
    dict_insert(obj, KEY_PASSWD, alloc_record_data_qstring(hi->passwd));
    /* set timestamps */
    snprintf(num, sizeof(num), FMT_TIME_T, hi->registered);
    dict_insert(obj, KEY_REGISTER_ON, alloc_record_data_qstring(num));
    snprintf(num, sizeof(num), FMT_TIME_T, hi->lastseen);
    dict_insert(obj, KEY_LAST_SEEN, alloc_record_data_qstring(num));
    /* record the hostmasks */
    if (hi->masks->used) {
        slist = string_list_copy(hi->masks);
        dict_insert(obj, KEY_MASKS, alloc_record_data_string_list(slist));
    }
    /* record the list of owned nicks */
    if (hi->nicks) {
        slist = alloc_string_list(nickserv_conf.nicks_per_handle);
        ni = hi->nicks;
        while (ni) {
            string_list_append(slist, strdup(ni->nick));
            ni = ni->next;
        }
        dict_insert(obj, KEY_NICKS, alloc_record_data_string_list(slist));
    }
    /* record the opserv level */
    if (hi->opserv_level) {
	snprintf(num, sizeof(num), "%d", hi->opserv_level);
	dict_insert(obj, KEY_OPSERV_LEVEL, alloc_record_data_qstring(num));
    }
    /* record the flags, if they have any set */
    if (hi->flags) {
	for (i=flen=0; handle_flags[i]; i++) {
	    if (hi->flags & 1 << i) flags[flen++] = handle_flags[i];
	}
	flags[flen] = 0;
	dict_insert(obj, KEY_FLAGS, alloc_record_data_qstring(flags));
    }
    /* record the user list style, cheat and reuse 'flags' */
    flags[0] = hi->userlist_style;
    flags[1] = 0;
    dict_insert(obj, KEY_USERLIST_STYLE, alloc_record_data_qstring(flags));
    /* record the infoline, if one is set */
    if (hi->infoline) {
	dict_insert(obj, KEY_INFO, alloc_record_data_qstring(hi->infoline));
    }
    /* record the screen width */
    snprintf(num, sizeof(num), "%d", hi->screen_width);
    dict_insert(obj, KEY_SCREEN_WIDTH, alloc_record_data_qstring(num));
    /* record the last authed host, email address, and epithet if they exist */
    if (hi->last_authed_host[0]) {
        dict_insert(obj, KEY_LAST_AUTHED_HOST, alloc_record_data_qstring(hi->last_authed_host));
    }
    if (hi->email_addr) {
        dict_insert(obj, KEY_EMAIL_ADDR, alloc_record_data_qstring(hi->email_addr));
    }
    if (hi->epithet) {
        dict_insert(obj, KEY_EPITHET, alloc_record_data_qstring(hi->epithet));
    }
    /* is there a cookie to write? */
    if (hi->cookie) {
        struct handle_cookie *cookie = hi->cookie;
        dict_t subdb = dict_new();
        char *type, expires[16];

        switch (cookie->type) {
        case ACTIVATION: type = KEY_ACTIVATION; break;
        case PASSWORD_CHANGE: type = KEY_PASSWORD_CHANGE; break;
        case EMAIL_CHANGE: type = KEY_EMAIL_CHANGE; break;
        case ALLOWAUTH: type = KEY_ALLOWAUTH; break;
        default: type = NULL;
        }
        if (type) dict_insert(subdb, KEY_COOKIE_TYPE, alloc_record_data_qstring(type));
        snprintf(expires, sizeof(expires), "%lu", cookie->expires);
        dict_insert(subdb, KEY_COOKIE_EXPIRES, alloc_record_data_qstring(expires));
        if (cookie->data) dict_insert(subdb, KEY_COOKIE_DATA, alloc_record_data_qstring(cookie->data));
        dict_insert(subdb, KEY_COOKIE, alloc_record_data_qstring(cookie->cookie));
        dict_insert(obj, KEY_COOKIE, alloc_record_data_object(subdb));
    }
    /* create the master record */
    rd = alloc_record_data_object(obj);
    /* write it */
    write_record(f, key, rd);
    /* now clean up (it works recursively) */
    free_record_data(rd);
    return 0;
}

static void
nickserv_db_write(void)
{
    dict_iterator_t it;

    FILE *f = fopen(TEMP_DB_NAME, "w");
    for (it=dict_first(nickserv_handle_dict); it; it=iter_next(it)) {
        nickserv_write_handle(iter_key(it), iter_data(it), f);
    }
    fclose(f);
    if (rename(TEMP_DB_NAME, NICKSERV_DB_NAME) == -1) {
	log(NS_LOG, LOG_ERROR, "Unable to rename new db to old: %s\n", strerror(errno));
    }
    lastwrite = now;
}

void
nickserv_timed_write(void *data)
{
    (void)data;

    nickserv_db_write();
    log(NS_LOG, LOG_INFO, "Successfully made scheduled backup of NickServ database.\n");
    timeq_add(now + nickserv_conf.db_backup_frequency, nickserv_timed_write, NULL);
}

static NICKSERV_FUNC(cmd_write)
{
    struct timeval start, stop;
    (void)argc; (void)argv; /* ignore them!  ignore them all! */

    gettimeofday(&start, NULL);
    nickserv_db_write();
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
	stop.tv_sec -= 1;
	stop.tv_usec += 1000000;
    }
    nickserv_notice(user, NSMSG_WRITE_SUCCESS, stop.tv_sec, stop.tv_usec/1000);
    log(NS_LOG, LOG_INFO, "%s ordered a NickServ database write.\n", user->nick);
    return 1;
}

static struct helpfile_expansion
nickserv_help_expand(char *variable)
{
    struct helpfile_expansion exp;
    dict_iterator_t it;
    int row;
    const char *cmdname;

    if (!strcasecmp(variable, "index")) {
        exp.type = HF_TABLE;
        exp.value.table.length = 1;
        exp.value.table.width = 1;
        exp.value.table.flags = TABLE_REPEAT_ROWS;
        exp.value.table.contents = calloc(dict_size(nickserv_func_dict)+1, sizeof(char**));
        exp.value.table.contents[0] = calloc(1, sizeof(char*));
        exp.value.table.contents[0][0] = "Commands:";
        for (it=dict_first(nickserv_func_dict); it; it=iter_next(it)) {
            cmdname = iter_key(it);
            if (strchr(cmdname, ' ')) continue;
            row = exp.value.table.length++;
            exp.value.table.contents[row] = calloc(1, sizeof(char*));
            exp.value.table.contents[row][0] = cmdname;
        }
    } else {
        exp.type = HF_STRING;
        exp.value.str = NULL;
    }
    return exp;
}

static void
nickserv_helpfile_read(void)
{
    nickserv_helpfile = open_helpfile(NICKSERV_HELPFILE_NAME, nickserv_help_expand);
}

static NICKSERV_FUNC(cmd_help)
{
    const char *topic;
    if (argc < 2) {
	topic = NULL;
    } else {
	unsplit_string(argv+1, argc-1);
	topic = argv[1];
    }
    nickserv_help(user, topic, 1);
    return 1;
}

static NICKSERV_FUNC(cmd_readhelp)
{
    struct timeval start, stop;
    struct helpfile *old_helpfile = nickserv_helpfile;
    (void)argc; (void)argv; /* ignore them!  ignore them all! */

    gettimeofday(&start, NULL);
    nickserv_helpfile_read();
    if (nickserv_helpfile) {
	close_helpfile(old_helpfile);
    } else {
	nickserv_helpfile = old_helpfile;
    }
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
	stop.tv_sec -= 1;
	stop.tv_usec += 1000000;
    }
    nickserv_notice(user, NSMSG_READHELP_SUCCESS, stop.tv_sec, stop.tv_usec/1000);
    log(NS_LOG, LOG_INFO, "%s ordered NickServ to read the help database.\n", user->nick);
    return 1;
}

static NICKSERV_FUNC(cmd_merge)
{
    struct handle_info *hi_from, *hi_to;
    struct userNode *last_user;
    struct chanList *cList, *cListNext, *cListEnd;
    unsigned int ii, jj;

    NICKSERV_MIN_PARMS(3);

    if (!(hi_from = get_victim_oper(user, argv[1]))) return 0;
    if (!(hi_to = get_victim_oper(user, argv[2]))) return 0;

    /* Append "from" handle's nicks to "to" handle's nick list. */
    if (hi_to->nicks) {
        struct nick_info *last_ni;
        for (last_ni=hi_to->nicks; last_ni->next; last_ni=last_ni->next) ;
        last_ni->next = hi_from->nicks;
    } else {
        hi_to->nicks = hi_from->nicks;
    }
    hi_from->nicks = NULL;

    /* Merge the hostmasks. */
    for (ii=0; ii<hi_from->masks->used; ii++) {
        char *mask = hi_from->masks->list[ii];
        for (jj=0; jj<hi_to->masks->used; jj++) {
            if (match_ircglobs(hi_to->masks->list[jj], mask)) break;
        }
        if (jj==hi_to->masks->used) {
            /* Nothing from the "to" handle covered this mask, so add it. */
            string_list_append(hi_to->masks, strdup(mask));
        }
    }

    /* Merge the users lists. */
    if (hi_to->users) {
        for (last_user=hi_to->users; last_user->next_authed; last_user=last_user->next_authed) ;
        last_user->next_authed = hi_from->users;
    } else {
        hi_to->users = hi_from->users;
    }
    /* Repoint the old "from" handle's users. */
    for (last_user=hi_from->users; last_user; last_user=last_user->next_authed) {
        last_user->handle_info = hi_to;
    }
    hi_from->users = NULL;

    /* Merge channel userlists. */
    if (hi_to->channels) {
        for (cListEnd=hi_to->channels; cListEnd->next; cListEnd=cListEnd->next) ;
    } else {
        cListEnd = NULL;
    }
    for (cList=hi_from->channels; cList; cList=cListNext) {
        struct chanList *cList2;
        cListNext = cList->next;
        for (cList2=hi_to->channels; cList2; cList2=cList2->next) {
            if (cList->channel == cList2->channel) break;
        }
        if (cList2 && (cList2->user->access >= cList->user->access)) {
            /* keep cList2 in hi_to; remove cList from hi_from */
            del_channel_user(cList->channel, cList->user);
        } else {
            if (cList2) {
                /* hi_from handle has higher access; remove cList2 from hi_to */
                del_channel_user(cList2->channel, cList2->user);
            }
            /* cList needs to be moved from hi_from to hi_to */
            cList->user->handle = hi_to;
            /* Remove from linked list for hi_from */
            if (cList->prev) {
                cList->prev->next = cList->next;
            } else {
                hi_from->channels = cList->next;
            }
            if (cList->next) {
                cList->next->prev = cList->prev;
            }
            /* Add to linked list for hi_to */
            if (cListEnd->next) {
                cListEnd->next->prev = cList;
            }
            cList->next = cListEnd->next;
            cList->prev = cListEnd;
            cListEnd->next = cList;
        }
    }

    /* Do they get an OpServ level promotion? */
    if (hi_from->opserv_level > hi_to->opserv_level) {
        hi_to->opserv_level = hi_from->opserv_level;
    }

    /* What about last seen time? */
    if (hi_from->lastseen > hi_to->lastseen) {
        hi_to->lastseen = hi_from->lastseen;
    }

    /* Notify of success. */
    nickserv_notice(user, NSMSG_HANDLES_MERGED, hi_from->handle, hi_to->handle);

    /* Unregister the "from" handle. */
    nickserv_unregister_handle(hi_from, NULL);

    return 1;
}

struct nickserv_discrim {
    unsigned int limit, min_level, max_level;
    unsigned long flags_on, flags_off;
    time_t registered, lastseen;
    enum { SUBSET, EXACT, SUPERSET } hostmask_type;
    const char *nickmask;
    const char *hostmask;
    const char *handlemask;
};

typedef void (*discrim_search_func)(struct userNode *source, struct handle_info *hi);

struct discrim_apply_info {
    struct nickserv_discrim *discrim;
    discrim_search_func func;
    struct userNode *source;
    unsigned int matched;
};

static struct nickserv_discrim *
nickserv_discrim_create(struct userNode *user, unsigned int argc, unsigned char *argv[])
{
    unsigned int i, j;
    struct nickserv_discrim *discrim;

    discrim = malloc(sizeof(*discrim));
    memset(discrim, 0, sizeof(*discrim));
    discrim->min_level = 0;
    discrim->max_level = ~0;
    discrim->limit = 50;
    discrim->registered = now;
    discrim->lastseen = now;

    for (i=0; i<argc; i++) {
        if (i == argc - 1) {
            nickserv_notice(user, MSG_MISSING_PARAMS, argv[i]);
            goto fail;
        }
        if (!strcasecmp(argv[i], "limit")) {
            discrim->limit = atoi(argv[++i]);
        } else if (!strcasecmp(argv[i], "flags")) {
            int is_on = 1, pos;
            i++;
            for (j=0; argv[i][j]; j++) {
                switch (argv[i][j]) {
                case '+': is_on = 1; break;
                case '-': is_on = 0; break;
                default:
                    pos = handle_inverse_flags[argv[i][j]];
                    if (!pos) break;
                    if (is_on) {
                        discrim->flags_on |= 1 << (pos-1);
                    } else {
                        discrim->flags_off |= 1 << (pos-1);
                    }
                }
            }
        } else if (!strcasecmp(argv[i], "registered")) {
            discrim->registered = now - ParseInterval(argv[++i]);
        } else if (!strcasecmp(argv[i], "seen")) {
            discrim->lastseen = now - ParseInterval(argv[++i]);
        } else if (!nickserv_conf.disable_nicks && !strcasecmp(argv[i], "nickmask")) {
            discrim->nickmask = argv[++i];
        } else if (!strcasecmp(argv[i], "hostmask")) {
            i++;
            if (!strcasecmp(argv[i], "exact")) {
                if (i == argc - 1) {
                    nickserv_notice(user, MSG_MISSING_PARAMS, argv[i]);
                    goto fail;
                }
                discrim->hostmask_type = EXACT;
            } else if (!strcasecmp(argv[i], "subset")) {
                if (i == argc - 1) {
                    nickserv_notice(user, MSG_MISSING_PARAMS, argv[i]);
                    goto fail;
                }
                discrim->hostmask_type = SUBSET;
            } else if (!strcasecmp(argv[i], "superset")) {
                if (i == argc - 1) {
                    nickserv_notice(user, MSG_MISSING_PARAMS, argv[i]);
                    goto fail;
                }
                discrim->hostmask_type = SUPERSET;
            } else {
                i--;
                discrim->hostmask_type = SUPERSET;
            }
            discrim->hostmask = argv[++i];
        } else if (!strcasecmp(argv[i], "handlemask")) {
            discrim->handlemask = argv[++i];
        } else if (!strcasecmp(argv[i], "access")) {
            const unsigned char *cmp = argv[++i];
            if (cmp[0] == '<') {
                if (discrim->min_level == 0) discrim->min_level = 1;
                if (cmp[1] == '=') {
                    discrim->max_level = strtoul(cmp+2, NULL, 0);
                } else {
                    discrim->max_level = strtoul(cmp+1, NULL, 0) - 1;
                }
            } else if (cmp[0] == '=') {
                discrim->min_level = discrim->max_level = strtoul(cmp+1, NULL, 0);
            } else if (cmp[0] == '>') {
                if (cmp[1] == '=') {
                    discrim->min_level = strtoul(cmp+2, NULL, 0);
                } else {
                    discrim->min_level = strtoul(cmp+1, NULL, 0) + 1;
                }
            } else {
                nickserv_notice(user, MSG_INVALID_CRITERIA, argv[i]);
            }
        } else {
            nickserv_notice(user, MSG_INVALID_CRITERIA, argv[i]);
            goto fail;
        }
    }
    return discrim;
  fail:
    free(discrim);
    return NULL;
}

static int
nickserv_discrim_match(struct nickserv_discrim *discrim, struct handle_info *hi)
{
    if (((discrim->flags_on & hi->flags) != discrim->flags_on)
        || (discrim->flags_off & hi->flags)
        || (discrim->registered < hi->registered)
        || (discrim->lastseen < (hi->users?now:hi->lastseen))
        || (discrim->handlemask && !match_ircglob(hi->handle, discrim->handlemask))
        || (discrim->min_level > hi->opserv_level)
        || (discrim->max_level < hi->opserv_level)) {
        return 0;
    }
    if (discrim->hostmask) {
        unsigned int i;
        for (i=0; i<hi->masks->used; i++) {
            const char *mask = hi->masks->list[i];
            if ((discrim->hostmask_type == SUBSET)
                && (match_ircglobs(discrim->hostmask, mask))) break;
            else if ((discrim->hostmask_type == EXACT)
                     && !strcasecmp(discrim->hostmask, mask)) break;
            else if (match_ircglobs(mask, discrim->hostmask)) break;
        }
        if (i==hi->masks->used) return 0;
    }
    if (discrim->nickmask) {
        struct nick_info *nick = hi->nicks;
        while (nick) {
            if (match_ircglob(nick->nick, discrim->nickmask)) break;
            nick = nick->next;
        }
        if (!nick) return 0;
    }
    return 1;
}

static unsigned int
nickserv_discrim_search(struct nickserv_discrim *discrim, discrim_search_func dsf, struct userNode *source)
{
    dict_iterator_t it, next;
    unsigned int matched = 0;

    for (it=dict_first(nickserv_handle_dict); it; it=next) {
        next = iter_next(it);
        if (nickserv_discrim_match(discrim, iter_data(it))) {
            dsf(source, iter_data(it));
            matched++;
        }
    }
    return matched;
}

static void
search_print_func(struct userNode *source, struct handle_info *match)
{
    nickserv_notice(source, NSMSG_SEARCH_MATCH, match->handle);
}

static void
search_count_func(struct userNode *source, struct handle_info *match)
{
    (void)source; (void)match;
}

static void
search_unregister_func (struct userNode *source, struct handle_info *match)
{
    if (is_valid_oper(source, match->opserv_level, 0)) {
        nickserv_unregister_handle(match, source);
    }
}

static NICKSERV_FUNC(cmd_search)
{
    struct nickserv_discrim *discrim;
    discrim_search_func action;
    struct nickserv_cmd *subcmd;
    unsigned int matches;

    NICKSERV_MIN_PARMS(3);
    if (!strcasecmp(argv[1], "print")) {
        action = search_print_func;
        subcmd = dict_find(nickserv_func_dict, "search print", NULL);
    } else if (!strcasecmp(argv[1], "count")) {
        action = search_count_func;
        subcmd = dict_find(nickserv_func_dict, "search count", NULL);
    } else if (!strcasecmp(argv[1], "unregister")) {
        action = search_unregister_func;
        subcmd = dict_find(nickserv_func_dict, "search unregister", NULL);
    } else {
        nickserv_notice(user, NSMSG_INVALID_ACTION, argv[1]);
        return 0;
    }

    if (subcmd
        && (subcmd->min_oper_level >= 0)
        && (user->handle_info->opserv_level < (unsigned)subcmd->min_oper_level)) {
        nickserv_notice(user, NSMSG_NO_ACCESS);
        return 0;
    }

    discrim = nickserv_discrim_create(user, argc-2, argv+2);
    if (!discrim) return 0;

    if (action == search_print_func) {
        nickserv_notice(user, MSG_SEARCH_RESULTS, "handles");
    } else if (action == search_count_func) {
        discrim->limit = INT_MAX;
    }

    matches = nickserv_discrim_search(discrim, action, user);

    if (matches) {
        nickserv_notice(user, MSG_MATCH_COUNT, matches);
    } else {
        nickserv_notice(user, MSG_NO_MATCHES);
    }

    free(discrim);
    return 0;
}

static void
nickserv_db_read_handle(const char *handle, dict_t obj)
{
    const unsigned char *str;
    struct string_list *masks, *slist;
    struct handle_info *hi;
    struct userNode *authed_users;
    unsigned int ii;
    dict_t subdb;

    str = database_get_data(obj, KEY_PASSWD, RECDB_QSTRING);
    if (!str) {
        log(NS_LOG, LOG_WARNING, "did not find a password for %s -- skipping user\n", handle);
        return;
    }
    if ((hi = get_handle_info(handle))) {
        authed_users = hi->users;
        hi->users = NULL;
        dict_remove(nickserv_handle_dict, hi->handle);
    } else {
        authed_users = NULL;
    }
    hi = register_handle(handle, str);
    if (authed_users) {
        hi->users = authed_users;
        while (authed_users) {
            authed_users->handle_info = hi;
            authed_users = authed_users->next_authed;
        }
    }
    masks = database_get_data(obj, KEY_MASKS, RECDB_STRING_LIST);
    hi->masks = masks ? string_list_copy(masks) : alloc_string_list(1);
    str = database_get_data(obj, KEY_OPSERV_LEVEL, RECDB_QSTRING);
    hi->opserv_level = str ? strtoul(str, NULL, 0) : 0;
    str = database_get_data(obj, KEY_INFO, RECDB_QSTRING);
    if (str) hi->infoline = strdup(str);
    str = database_get_data(obj, KEY_REGISTER_ON, RECDB_QSTRING);
    hi->registered = str ? (signed)strtoul(str, NULL, 0) : now;
    str = database_get_data(obj, KEY_LAST_SEEN, RECDB_QSTRING);
    hi->lastseen = str ? (signed)strtoul(str, NULL, 0) : hi->registered;
    /* We want to read the nicks even if disable_nicks is set.  This is so
     * that we don't lose the nick data entirely. */
    slist = database_get_data(obj, KEY_NICKS, RECDB_STRING_LIST);
    if (slist) {
        for (ii=0; ii<slist->used; ii++) {
            register_nick(slist->list[ii], hi);
        }
    }
    str = database_get_data(obj, KEY_FLAGS, RECDB_QSTRING);
    if (str) {
        for (ii=0; str[ii]; ii++) {
            hi->flags |= 1 << (handle_inverse_flags[str[ii]] - 1);
        }
    }
    str = database_get_data(obj, KEY_USERLIST_STYLE, RECDB_QSTRING);
    hi->userlist_style = str ? str[0] : HI_STYLE_ZOOT;
    str = database_get_data(obj, KEY_SCREEN_WIDTH, RECDB_QSTRING);
    hi->screen_width = str ? strtoul(str, NULL, 0) : DEFAULT_LINE_SIZE;
    if (hi->screen_width > MAX_LINE_SIZE) hi->screen_width = MAX_LINE_SIZE;
    str = database_get_data(obj, KEY_LAST_AUTHED_HOST, RECDB_QSTRING);
    if (str) safestrncpy(hi->last_authed_host, str, sizeof(hi->last_authed_host));
    str = database_get_data(obj, KEY_EMAIL_ADDR, RECDB_QSTRING);
    if (str) nickserv_set_email_addr(hi, str);
    str = database_get_data(obj, KEY_EPITHET, RECDB_QSTRING);
    if (str) hi->epithet = strdup(str);
    subdb = database_get_data(obj, KEY_COOKIE, RECDB_OBJECT);
    if (subdb) {
        const char *data, *type, *expires, *cookie_str;
        struct handle_cookie *cookie;

        cookie = calloc(1, sizeof(*cookie));
        type = database_get_data(subdb, KEY_COOKIE_TYPE, RECDB_QSTRING);
        data = database_get_data(subdb, KEY_COOKIE_DATA, RECDB_QSTRING);
        expires = database_get_data(subdb, KEY_COOKIE_EXPIRES, RECDB_QSTRING);
        cookie_str = database_get_data(subdb, KEY_COOKIE, RECDB_QSTRING);
        if (!type || !data || !expires || !cookie_str) {
            log(NS_LOG, LOG_ERROR, "Missing field(s) from cookie for account %s; dropping cookie.\n", hi->handle);
            goto cookie_out;
        }
        if (!strcasecmp(type, KEY_ACTIVATION)) cookie->type = ACTIVATION;
        else if (!strcasecmp(type, KEY_PASSWORD_CHANGE)) cookie->type = PASSWORD_CHANGE;
        else if (!strcasecmp(type, KEY_EMAIL_CHANGE)) cookie->type = EMAIL_CHANGE;
        else if (!strcasecmp(type, KEY_ALLOWAUTH)) cookie->type = ALLOWAUTH;
        else {
            log(NS_LOG, LOG_ERROR, "Invalid cookie type %s for account %s; dropping cookie.\n", type, cookie);
            goto cookie_out;
        }
        cookie->expires = strtoul(expires, NULL, 0);
        if (cookie->expires < now) goto cookie_out;
        cookie->data = strdup(data);
        safestrncpy(cookie->cookie, cookie_str, sizeof(cookie->cookie));
        cookie->hi = hi;
      cookie_out:
        if (cookie->hi) {
            nickserv_bake_cookie(cookie);
        } else {
            nickserv_free_cookie(cookie);
        }
    }
}

static void
nickserv_db_read(const char *fname)
{
    struct record_data *hir;
    dict_t db;
    dict_iterator_t it;

    db = parse_database(fname);
    if (!db) return;
    for (it=dict_first(db); it; it=iter_next(it)) {
        hir = iter_data(it);
        nickserv_db_read_handle(iter_key(it), hir->d.object);
    }
    free_database(db);
}

static NICKSERV_FUNC(cmd_mergedb)
{
    struct timeval start, stop;

    NICKSERV_MIN_PARMS(2);
    gettimeofday(&start, NULL);
    nickserv_db_read(argv[1]);
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
	stop.tv_sec -= 1;
	stop.tv_usec += 1000000;
    }
    nickserv_notice(user, NSMSG_DB_MERGED, argv[1], stop.tv_sec, stop.tv_usec/1000);
    log(NS_LOG, LOG_INFO, "%s merged %s into the database.\n", user->nick, argv[1]);
    return 0;
}

static void
nickserv_msg(struct userNode *user, struct userNode *target, char *command, int server_qualified)
{
    struct nickserv_cmd *cmd;
    unsigned char *argv[MAXNUMPARAMS];
    int argc;
    char *dupcmd;

    (void)target;

    dupcmd = alloca(strlen(command)+1);
    strcpy(dupcmd, command);

    argc = split_line(command, false, ArrayLength(argv), argv);
    if (!argc) return;

    cmd = dict_find(nickserv_func_dict, argv[0], NULL);

    if (!cmd) {
	nickserv_notice(user, MSG_COMMAND_UNKNOWN, argv[0]);
	return;
    }

    if (!policer_conforms(user->command_policer, now, cmd->weight)) {
	nickserv_notice(user, MSG_COMMAND_FLOOD);
	return;
    }

    if (nickserv_conf.require_qualified && cmd->must_be_qualified && !server_qualified) {
        unsplit_string(argv, argc);
        nickserv_notice(user, NSMSG_MUST_QUALIFY, argv[0]);
        return;
    }

    if (cmd->must_auth && !user->handle_info) {
        nickserv_notice(user, NSMSG_MUST_AUTH);
        return;
    }

    if (cmd->min_oper_level >= 0) {
        if (!is_valid_oper(user, cmd->min_oper_level, 0)) return;
        /* is_valid_oper constrains user to be authed */
        log(NS_LOG, LOG_INFO, "%s invoked: %s\n", user->handle_info->handle, dupcmd);
    }

    cmd->func(user, argc, argv);
}

static void
expire_handles(void *data)
{
    dict_iterator_t it, next;
    time_t expiry;
    struct handle_info *hi;

    (void)data;
    for (it=dict_first(nickserv_handle_dict); it; it=next) {
        next = iter_next(it);
        hi = iter_data(it);
        if ((hi->opserv_level > 0)
            || hi->users
            || HANDLE_FLAGGED(hi, FROZEN)
            || HANDLE_FLAGGED(hi, NODELETE)) {
            continue;
        }
        expiry = hi->channels ? nickserv_conf.handle_expire_delay : nickserv_conf.nochan_handle_expire_delay;
        if ((now - hi->lastseen) > expiry) {
            log(NS_LOG, LOG_INFO, "Expiring account %s for inactivity.\n", hi->handle);
            nickserv_unregister_handle(hi, NULL);
        }
    }

    if (nickserv_conf.handle_expire_frequency) {
        timeq_add(now + nickserv_conf.handle_expire_frequency, expire_handles, NULL);
    }
}

static void
nickserv_load_dict(const char *fname)
{
    FILE *file;
    char line[128];
    if (!(file = fopen(fname, "r"))) {
        log(NS_LOG, LOG_ERROR, "Unable to open dictionary file %s: %s\n", fname, strerror(errno));
        return;
    }
    while (!feof(file)) {
        fgets(line, sizeof(line), file);
        if (!line[0]) continue;
        if (line[strlen(line)-1] == '\n') line[strlen(line)-1] = 0;
        dict_insert(nickserv_conf.weak_password_dict, strdup(line), NULL);
    }
    fclose(file);
    log(NS_LOG, LOG_INFO, "Loaded %d words into weak password dictionary\n", dict_size(nickserv_conf.weak_password_dict));
}

static void
nickserv_conf_read(void)
{
    dict_t conf_node, child;
    const char *str;
    dict_iterator_t it;

    if (!(conf_node = conf_get_data(NICKSERV_CONF_NAME, RECDB_OBJECT))) {
	log(NS_LOG, LOG_ERROR, "config node `%s' is missing or has wrong type\n",
	    NICKSERV_CONF_NAME);
	return;
    }
    str = database_get_data(conf_node, KEY_VALID_HANDLE_REGEX, RECDB_QSTRING);
    if (!str) str = database_get_data(conf_node, KEY_VALID_ACCOUNT_REGEX, RECDB_QSTRING);
    if (nickserv_conf.valid_handle_regex_set) regfree(&nickserv_conf.valid_handle_regex);
    if (str) {
        int err = regcomp(&nickserv_conf.valid_handle_regex, str, REG_EXTENDED|REG_ICASE|REG_NOSUB);
        nickserv_conf.valid_handle_regex_set = !err;
        if (err) {
            log(NS_LOG, LOG_ERROR, "Bad valid_account_regex (error %d)\n", err);
        }
    } else {
        nickserv_conf.valid_handle_regex_set = 0;
    }
    str = database_get_data(conf_node, KEY_VALID_NICK_REGEX, RECDB_QSTRING);
    if (nickserv_conf.valid_nick_regex_set) regfree(&nickserv_conf.valid_nick_regex);
    if (str) {
        int err = regcomp(&nickserv_conf.valid_nick_regex, str, REG_EXTENDED|REG_ICASE|REG_NOSUB);
        nickserv_conf.valid_nick_regex_set = !err;
        if (err) {
            log(NS_LOG, LOG_ERROR, "Bad valid_nick_regex (error %d)\n", err);
        }
    } else {
        nickserv_conf.valid_nick_regex_set = 0;
    }
    str = database_get_data(conf_node, KEY_NICKS_PER_HANDLE, RECDB_QSTRING);
    if (!str) str = database_get_data(conf_node, KEY_NICKS_PER_ACCOUNT, RECDB_QSTRING);
    nickserv_conf.nicks_per_handle = str ? strtoul(str, NULL, 0) : 4;
    str = database_get_data(conf_node, KEY_DISABLE_NICKS, RECDB_QSTRING);
    nickserv_conf.disable_nicks = str ? strtoul(str, NULL, 0) : 0;
    str = database_get_data(conf_node, KEY_DEFAULT_HOSTMASK, RECDB_QSTRING);
    nickserv_conf.default_hostmask = str ? strdup(str) : 0;
    str = database_get_data(conf_node, KEY_PASSWORD_MIN_LENGTH, RECDB_QSTRING);
    nickserv_conf.password_min_length = str ? strtoul(str, NULL, 0) : 0;
    str = database_get_data(conf_node, KEY_PASSWORD_MIN_DIGITS, RECDB_QSTRING);
    nickserv_conf.password_min_digits = str ? strtoul(str, NULL, 0) : 0;
    str = database_get_data(conf_node, KEY_PASSWORD_MIN_UPPER, RECDB_QSTRING);
    nickserv_conf.password_min_upper = str ? strtoul(str, NULL, 0) : 0;
    str = database_get_data(conf_node, KEY_PASSWORD_MIN_LOWER, RECDB_QSTRING);
    nickserv_conf.password_min_lower = str ? strtoul(str, NULL, 0) : 0;
    str = database_get_data(conf_node, KEY_DB_BACKUP_FREQ, RECDB_QSTRING);
    nickserv_conf.db_backup_frequency = str ? ParseInterval(str) : 7200;
    str = database_get_data(conf_node, KEY_MODOPER_LEVEL, RECDB_QSTRING);
    nickserv_conf.modoper_level = str ? strtoul(str, NULL, 0) : 900;
    str = database_get_data(conf_node, KEY_SET_EPITHET_LEVEL, RECDB_QSTRING);
    nickserv_conf.set_epithet_level = str ? strtoul(str, NULL, 0) : 1;
    str = database_get_data(conf_node, KEY_HANDLE_EXPIRE_FREQ, RECDB_QSTRING);
    if (!str) str = database_get_data(conf_node, KEY_ACCOUNT_EXPIRE_FREQ, RECDB_QSTRING);
    nickserv_conf.handle_expire_frequency = str ? ParseInterval(str) : 86400;
    str = database_get_data(conf_node, KEY_HANDLE_EXPIRE_DELAY, RECDB_QSTRING);
    if (!str) str = database_get_data(conf_node, KEY_ACCOUNT_EXPIRE_DELAY, RECDB_QSTRING);
    nickserv_conf.handle_expire_delay = str ? ParseInterval(str) : 86400*30;
    str = database_get_data(conf_node, KEY_NOCHAN_HANDLE_EXPIRE_DELAY, RECDB_QSTRING);
    if (!str) str = database_get_data(conf_node, KEY_NOCHAN_ACCOUNT_EXPIRE_DELAY, RECDB_QSTRING);
    nickserv_conf.nochan_handle_expire_delay = str ? ParseInterval(str) : 86400*15;
    str = database_get_data(conf_node, KEY_WARN_NICK_OWNED, RECDB_QSTRING);
    nickserv_conf.warn_nick_owned = str ? enabled_string(str) : 0;
    child = database_get_data(conf_node, KEY_FLAG_LEVELS, RECDB_OBJECT);
    for (it=dict_first(child); it; it=iter_next(it)) {
        const char *key = iter_key(it), *value;
        unsigned char flag;
        int pos;

        if (!strncasecmp(key, "uc_", 3)) {
            flag = toupper(key[3]);
        } else if (!strncasecmp(key, "lc_", 3)) {
            flag = tolower(key[3]);
        } else {
            flag = key[0];
        }

        if ((pos = handle_inverse_flags[flag])) {
            value = GET_RECORD_QSTRING((struct record_data*)iter_data(it));
            flag_access_levels[pos - 1] = strtoul(value, NULL, 0);
        }
    }
    if (nickserv_conf.weak_password_dict) dict_delete(nickserv_conf.weak_password_dict);
    nickserv_conf.weak_password_dict = dict_new();
    dict_set_free_keys(nickserv_conf.weak_password_dict, free);
    dict_insert(nickserv_conf.weak_password_dict, strdup("password"), NULL);
    dict_insert(nickserv_conf.weak_password_dict, strdup("<password>"), NULL);
    str = database_get_data(conf_node, KEY_DICT_FILE, RECDB_QSTRING);
    if (str) nickserv_load_dict(str);
    str = database_get_data(conf_node, KEY_NICK, RECDB_QSTRING);
    if (str) NickChange(nickserv, str, 0);
    str = database_get_data(conf_node, KEY_REQUIRE_QUALIFIED, RECDB_QSTRING);
    nickserv_conf.require_qualified = str ? strtoul(str, NULL, 0) : 1;
    str = database_get_data(conf_node, KEY_AUTOGAG_ENABLED, RECDB_QSTRING);
    nickserv_conf.autogag_enabled = str ? strtoul(str, NULL, 0) : 1;
    str = database_get_data(conf_node, KEY_AUTOGAG_DURATION, RECDB_QSTRING);
    nickserv_conf.autogag_duration = str ? ParseInterval(str) : 1800;
    str = database_get_data(conf_node, KEY_EMAIL_VISIBLE_LEVEL, RECDB_QSTRING);
    nickserv_conf.email_visible_level = str ? strtoul(str, NULL, 0) : 800;
    str = database_get_data(conf_node, KEY_EMAIL_ENABLED, RECDB_QSTRING);
    nickserv_conf.email_enabled = str ? enabled_string(str) : 0;
    str = database_get_data(conf_node, KEY_COOKIE_TIMEOUT, RECDB_QSTRING);
    nickserv_conf.cookie_timeout = str ? ParseInterval(str) : 24*3600;
    str = database_get_data(conf_node, KEY_EMAIL_REQUIRED, RECDB_QSTRING);
    nickserv_conf.email_required = str ? enabled_string(str) : 0;
    str = database_get_data(conf_node, KEY_ACCOUNTS_PER_EMAIL, RECDB_QSTRING);
    nickserv_conf.handles_per_email = str ? strtoul(str, NULL, 0) : 1;
    if (!nickserv_conf.auth_policer_params) {
        nickserv_conf.auth_policer_params = policer_params_new();
        policer_params_set(nickserv_conf.auth_policer_params, "size", "5");
        policer_params_set(nickserv_conf.auth_policer_params, "drain-rate", "0.05");
    }
    child = database_get_data(conf_node, KEY_AUTH_POLICER, RECDB_OBJECT);
    for (it=dict_first(child); it; it=iter_next(it)) {
        set_policer_param(iter_key(it), iter_data(it), nickserv_conf.auth_policer_params);
    }
}

int
handle_new_user(struct userNode *user)
{
    struct nick_info *ni;
    if (nickserv_conf.warn_nick_owned
        && (ni = get_nick_info(user->nick))
        && (user->handle_info != ni->owner)) {
        nickserv_notice(user, NSMSG_NICK_OWNED, ni->nick, ni->owner->handle);
    }
    return 0;
}

void
handle_nick_change(struct userNode *user, const char *new_nick)
{
    struct nick_info *ni;
    struct handle_info *hi;
    if ((hi = dict_find(nickserv_allow_auth_dict, user->nick, 0))) {
        /* Need to update allow_auth_dict .. */
        dict_remove(nickserv_allow_auth_dict, user->nick);
        dict_insert(nickserv_allow_auth_dict, new_nick, hi);
    }
    if (nickserv_conf.warn_nick_owned
        && (ni = get_nick_info(new_nick))) {
        /* Send the user a warning that it's already registered. */
        nickserv_notice(user, NSMSG_NICK_OWNED, ni->nick, ni->owner->handle);
    }
}

void
nickserv_remove_user(struct userNode *user)
{
    dict_remove(nickserv_allow_auth_dict, user->nick);
    set_user_handle_info(user, NULL);
}

int nickserv_command_search(const char *command)
{
    return dict_find(nickserv_func_dict, command, NULL) ? 1 : 0;
}

int nickserv_add_alias(const char *alias, const char *command)
{
    struct nickserv_cmd *cmd = dict_find(nickserv_func_dict, command, NULL);
    if(!cmd) return 0;
    cmd->refs++;
    dict_insert(nickserv_func_dict, alias, cmd);
    return 1;
}

int nickserv_del_alias(const char *alias)
{
    return dict_remove(nickserv_func_dict, alias);
}

static struct nickserv_cmd *
nickserv_define_func(const char *name, nickserv_func_t *func, int min_level, int must_auth, int must_be_qualified)
{
    struct nickserv_cmd *cmd = malloc(sizeof(struct nickserv_cmd));
    cmd->min_oper_level = min_level;
    cmd->weight = 1.0;
    cmd->func = func;
    cmd->must_auth = must_auth;
    cmd->must_be_qualified = must_be_qualified;
    cmd->refs = 1;
    dict_insert(nickserv_func_dict, name, cmd);
    return cmd;
}

static void
nickserv_unref_cmd(void *data)
{
    struct nickserv_cmd *cmd = data;
    if (!--cmd->refs) free(cmd);
}

static void
nickserv_db_cleanup(void)
{
    close_helpfile(nickserv_helpfile);
    dict_delete(nickserv_func_dict);
    unreg_del_user_func(nickserv_remove_user);
    userList_clean(&curr_helpers);
    policer_params_delete(nickserv_conf.auth_policer_params);
    dict_delete(nickserv_handle_dict);
    dict_delete(nickserv_nick_dict);
    dict_delete(nickserv_opt_dict);
    dict_delete(nickserv_allow_auth_dict);
    dict_delete(nickserv_email_dict);
    dict_delete(nickserv_conf.weak_password_dict);
    if (auth_func_list) free(auth_func_list);
    if (unreg_func_list) free(unreg_func_list);
    if (rf_list) free(rf_list);
    if (nickserv_conf.valid_handle_regex_set) regfree(&nickserv_conf.valid_handle_regex);
    if (nickserv_conf.valid_nick_regex_set) regfree(&nickserv_conf.valid_nick_regex);
}

void
init_nickserv(const char *nick)
{
    unsigned int i;
    nickserv = AddService(nick, "Nick Services");
    reg_privmsg_func(nickserv->num_local, nickserv_msg);
    reg_new_user_func(handle_new_user);
    reg_nick_change_func(handle_nick_change);
    reg_del_user_func(nickserv_remove_user);

    /* set up handle_inverse_flags */
    memset(handle_inverse_flags, 0, sizeof(handle_inverse_flags));
    for (i=0; handle_flags[i]; i++) {
        handle_inverse_flags[handle_flags[i]] = i + 1;
        flag_access_levels[i] = 0;
    }

    nickserv_conf_read();
    nickserv_func_dict = dict_new();
    dict_set_free_data(nickserv_func_dict, nickserv_unref_cmd);
    nickserv_opt_dict = dict_new();
    nickserv_email_dict = dict_new();
    dict_set_free_keys(nickserv_email_dict, free);
    dict_set_free_data(nickserv_email_dict, nickserv_free_email_addr);
    /* handle management commands */
    nickserv_define_func("AUTH", cmd_auth, -1, 0, 1);
    nickserv_define_func("ALLOWAUTH", cmd_allowauth, 0, 1, 0);
    nickserv_define_func("REGISTER", cmd_register, -1, 0, 1);
    nickserv_define_func("OREGISTER", cmd_oregister, 0, 1, 0);
    nickserv_define_func("UNREGISTER", cmd_unregister, -1, 1, 1);
    nickserv_define_func("OUNREGISTER", cmd_ounregister, 0, 1, 0);
    nickserv_define_func("ADDMASK", cmd_addmask, -1, 1, 0);
    nickserv_define_func("OADDMASK", cmd_oaddmask, 0, 1, 0);
    nickserv_define_func("DELMASK", cmd_delmask, -1, 1, 0);
    nickserv_define_func("ODELMASK", cmd_odelmask, 0, 1, 0);
    nickserv_define_func("PASS", cmd_pass, -1, 1, 1);
    nickserv_define_func("SET", cmd_set, -1, 1, 0);
#ifdef DISTRUST_OPERS
    nickserv_define_func("OSET", cmd_oset, 500, 1, 0);
#else
    nickserv_define_func("OSET", cmd_oset, 0, 1, 0);
#endif
    nickserv_define_func("ACCOUNTINFO", cmd_handleinfo, -1, 0, 0);
    nickserv_define_func("USERINFO", cmd_userinfo, -1, 1, 0);
    nickserv_define_func("RENAME", cmd_rename_handle, 0, 1, 0);
    nickserv_define_func("VACATION", cmd_vacation, -1, 1, 0);
    nickserv_define_func("MERGE", cmd_merge, 0, 1, 0);
    if (!nickserv_conf.disable_nicks) {
	/* nick management commands */
	nickserv_define_func("REGNICK", cmd_regnick, -1, 1, 0);
	nickserv_define_func("OREGNICK", cmd_regnick, 0, 1, 0);
	nickserv_define_func("UNREGNICK", cmd_unregnick, -1, 1, 0);
	nickserv_define_func("OUNREGNICK", cmd_ounregnick, 0, 1, 0);
	nickserv_define_func("NICKINFO", cmd_nickinfo, -1, 1, 0);
	nickserv_define_func("KILL", cmd_kill, -1, 1, 0);
    }
    if (nickserv_conf.email_enabled) {
        nickserv_define_func("AUTHCOOKIE", cmd_authcookie, -1, 0, 0);
        nickserv_define_func("DELCOOKIE", cmd_delcookie, -1, 1, 0);
        nickserv_define_func("RESETPASS", cmd_resetpass, -1, 0, 1);
        nickserv_define_func("COOKIE", cmd_cookie, -1, 0, 1);
        dict_insert(nickserv_opt_dict, "EMAIL", opt_email);
    }
    /* miscellaneous commands */
    nickserv_define_func("STATUS", cmd_status, -1, 0, 0);
    nickserv_define_func("VERSION", cmd_version, -1, 0, 0);
    nickserv_define_func("HELP", cmd_help, -1, 0, 0);
    nickserv_define_func("WRITE", cmd_write, 800, 1, 0);
    nickserv_define_func("READHELP", cmd_readhelp, 900, 1, 0);
    nickserv_define_func("SEARCH", cmd_search, 100, 1, 0);
    nickserv_define_func("SEARCH UNREGISTER", NULL, 800, 1, 0);
    nickserv_define_func("MERGEDB", cmd_mergedb, 999, 1, 0);
    /* other options */
    dict_insert(nickserv_opt_dict, "INFO", opt_info);
    dict_insert(nickserv_opt_dict, "WIDTH", opt_width);
    dict_insert(nickserv_opt_dict, "COLOR", opt_color);
    dict_insert(nickserv_opt_dict, "PRIVMSG", opt_privmsg);
    dict_insert(nickserv_opt_dict, "STYLE", opt_style);
    dict_insert(nickserv_opt_dict, "PASS", opt_password);
    dict_insert(nickserv_opt_dict, "PASSWORD", opt_password);
    dict_insert(nickserv_opt_dict, "FLAGS", opt_flags);
    dict_insert(nickserv_opt_dict, "ACCESS", opt_level);
    dict_insert(nickserv_opt_dict, "LEVEL", opt_level);
    dict_insert(nickserv_opt_dict, "EPITHET", opt_epithet);

    nickserv_handle_dict = dict_new();
    dict_set_free_data(nickserv_handle_dict, free_handle_info);

    nickserv_nick_dict = dict_new();
    dict_set_free_data(nickserv_nick_dict, free_nick_info);

    nickserv_allow_auth_dict = dict_new();
    dict_set_free_keys(nickserv_allow_auth_dict, free);

    alias_register_service("NickServ", nickserv_command_search, nickserv_add_alias, nickserv_del_alias);

    userList_init(&curr_helpers);

    conf_register_reload(nickserv_conf_read);
    nickserv_helpfile_read();
    nickserv_db_read(NICKSERV_DB_NAME);

    if(nickserv_conf.db_backup_frequency) timeq_add(now + nickserv_conf.db_backup_frequency, nickserv_timed_write, NULL);

    if(nickserv_conf.handle_expire_frequency) timeq_add(now + nickserv_conf.handle_expire_frequency, expire_handles, NULL);

    reg_exit_func(nickserv_db_cleanup);
    reg_db_write_func(nickserv_db_write);
}
