/* chanserv.c - Channel service bot
 * Copyright 2000-2002 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: chanserv.c,v 1.161.2.112.2.87 2002/05/04 01:35:15 entrope Exp $
 */

#include "config.h"
#if !defined(TRANSITION)
#define TRANSITION 0
#endif

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if defined(HAVE_REGEX_H)
#include <regex.h>
#endif

#include "common.h"
#include "dict.h"
#include "conf.h"
#include "hash.h"
#include "helpfile.h"
#include "log.h"
#include "parse.h"
#include "proto.h"
#include "global.h"
#include "nickserv.h"
#include "chanserv.h"
#include "opserv.h"
#include "policer.h"
#include "recdb.h"
#include "timeq.h"
#include "alias.h"

#define CHANSERV_CONF_NAME	"services/chanserv"

#define CHANSERV_DB_NAME	"chanserv.db"
#define TEMP_DB_NAME		"chanserv.db.new"
#define CHANSERV_HELPFILE_NAME	"chanserv.help"

/* ChanServ options */
#define KEY_DEBUG_CHANNEL	"debug_channel"
#define KEY_SUPPORT_CHANNEL     "support_channel"
#define KEY_DB_BACKUP_FREQ	"db_backup_freq"
#define KEY_INFO_DELAY		"info_delay"
#define KEY_MAX_GREETLEN       	"max_greetlen"
#define KEY_ADJUST_THRESHOLD   	"adjust_threshold"
#define KEY_JOINFLOOD_THRESHOLD "joinflood_threshold"
#define KEY_ADJUST_DELAY   	"adjust_delay"
#define KEY_CHAN_EXPIRE_FREQ	"chan_expire_freq"
#define KEY_CHAN_EXPIRE_DELAY	"chan_expire_delay"
#define KEY_MAX_CHAN_USERS     	"max_chan_users"
#define KEY_MAX_CHAN_BANS	"max_chan_bans"
#define KEY_NICK		"nick"
#define KEY_TRIGGER             "trigger"
#define KEY_OLD_CHANSERV_NAME	"old_chanserv_name"
#define KEY_MAX_SWITCH_LOAD	"max_switch_load"
#define KEY_SWITCH_TIMEOUT	"switch_timeout"
#define KEY_ACCESS_REGEX	"access_regex"
#define KEY_AUTH_REGEX		"auth_regex"
#define KEY_LAME_TRICKS         "lame_tricks"
#define KEY_8BALL_RESPONSES     "8ball"
#define KEY_REFRESH_PERIOD      "refresh_period"
#define KEY_CTCP_SHORT_BAN_DURATION "ctcp_short_ban_duration"
#define KEY_CTCP_LONG_BAN_DURATION "ctcp_long_ban_duration"
#define KEY_MAX_OWNED           "max_owned"

/* ChanServ database */
#define KEY_COMMANDS		"commands"
#define KEY_CHANNELS		"channels"
#define KEY_NOTE_TYPES          "note_types"

/* Note type parameters */
#define KEY_NOTE_OPSERV_ACCESS  "opserv_access"
#define KEY_NOTE_CHANNEL_ACCESS "channel_access"
#define KEY_NOTE_SETTER_ACCESS  "setter_access"
#define KEY_NOTE_VISIBILITY     "visibility"
#define KEY_NOTE_VIS_PRIVILEGED "privileged"
#define KEY_NOTE_VIS_CHANNEL_USERS "channel_users"
#define KEY_NOTE_VIS_ALL        "all"
#define KEY_NOTE_MAX_LENGTH     "max_length"
#define KEY_NOTE_SETTER         "setter"
#define KEY_NOTE_NOTE           "note"

/* Do-not-register channels */
#define KEY_DNR                 "dnr"
#define KEY_DNR_SETTER          "setter"
#define KEY_DNR_REASON          "reason"

/* Command data */
#define KEY_WEIGHT		"weight"
#define KEY_ACCESS		"access"
#define KEY_FLAGS		"flags"

/* Channel data */
#define KEY_REGISTERED		"registered"
#define KEY_REGISTRAR		"registrar"
#define KEY_SUSPENDER		"suspender"
#define KEY_SUSPEND_EXPIRES     "suspend_expires"
#define KEY_SUSPEND_REASON      "suspend_reason"
#define KEY_VISITED		"visited"
#define KEY_TOPIC		"topic"
#define KEY_GREETING		"greeting"
#define KEY_MODES		"modes"
#define KEY_FLAGS		"flags"
#define KEY_USERS		"users"
#define KEY_BANS		"bans"
#define KEY_MAX			"max"
#define KEY_NOTES               "notes"
#define KEY_TOPIC_MASK          "topic_mask"

/* User data */
#define KEY_LEVEL		"level"
#define KEY_INFO		"info"
#define KEY_SEEN		"seen"

/* Ban data */
#define KEY_OWNER		"owner"
#define KEY_REASON		"reason"
#define KEY_SET			"set"
#define KEY_DURATION		"duration"
#define KEY_EXPIRES             "expires"
#define KEY_TRIGGERED		"triggered"

/* Administrative messages */
#define CSMSG_WRITE_SUCCESS	"Wrote $b$C$b database in "FMT_TIME_T".%03lu seconds."
#define CSMSG_HELP_READ_SUCCESS	"Read help database in "FMT_TIME_T".%03lu seconds."
#define CSMSG_CHANNELS_EXPIRED	"%i channels expired."

/* Channel registration */
#define CSMSG_REG_SUCCESS	"You now have ownership of $b%s$b."
#define CSMSG_PROXY_SUCCESS	"%s now has ownership of $b%s$b."
#define CSMSG_ALREADY_REGGED	"$b%s$b is registered to someone else."
#define CSMSG_MUST_BE_OPPED	"You must be a channel operator in $b%s$b to register it."
#define CSMSG_PROXY_FORBIDDEN	"You may not register a channel for someone else."
#define CSMSG_COLORS_DISALLOWED "You may not have color codes in your %s."
#define CSMSG_OWN_TOO_MANY      "%s already owns enough channels (%d); use FORCE to override."

/* Do-not-register channels */
#define CSMSG_DNR_INFO          "$b%s$b is do-not-register (by $b%s$b): %s"
#define CSMSG_END_DNR_LIST      "End of do-not-register list."
#define CSMSG_DNR_CHANNEL       "$b%s$b may only be registered by network staff."
#define CSMSG_NOREGISTER_CHANNEL "New do-not-register added for $b%s$b."
#define CSMSG_NO_SUCH_DNR       "$b%s$b is not in the do-not-register list."
#define CSMSG_DNR_REMOVED       "$b%s$b has been removed from the do-not-register list."

/* Channel unregistration */
#define CSMSG_UNREG_SUCCESS	"$b%s$b has been unregistered."
#define CSMSG_UNREG_NODELETE	"$b%s$b is protected from unregistration."
#define CSMSG_CONFIRM_UNREG	"To confirm this unregistration, you must append 'CONFIRM' to the end of your command. For example, 'unregister CONFIRM'."

/* Channel moving */
#define CSMSG_MOVE_SUCCESS	"Channel registration has been transferred to $b%s$b."
#define CSMSG_MOVE_NODELETE	"$b%s$b is protected from unregistration, and cannot be moved."

/* Channel merging */
#define CSMSG_MERGE_SUCCESS	"Channel successfully merged into $b%s$b."
#define CSMSG_MERGE_SUSPENDED	"Merging cannot be performed if the source or target channel is suspended."

/* Handle unregistration */
#define CSMSG_HANDLE_UNREGISTERED "As a result of your account unregistration, you have been deleted from all of your channels' userlists."

/* Error messages */
#define CSMSG_NOT_USER		"You lack access to $b%s$b."
#define CSMSG_NO_CHAN_USER	"%s lacks access to $b%s$b."
#define CSMSG_NO_ACCESS		"You lack sufficient access to use this command."
#define CSMSG_ACCESS_REQUIRED  	"You must have $b%s$b access in %s to use this command."
#define CSMSG_UNREGISTERED	"$b%s$b has not been registered with $b$C$b."
#define CSMSG_STRUCTURE_LIMIT	"This channel has reached the maximum %s limit of $b%d$b."
#define CSMSG_INVALID_PROPERTY	"$b%s$b is an invalid command property."
#define CSMSG_ILLEGAL_CHANNEL	"$b%s$b is an illegal channel; cannot register it."
#define CSMSG_PUBLIC_DENY	"Public commands in $b%s$b are restricted."
#define CSMSG_ALREADY_UP	"You are already %s in $b%s$b."
#define CSMSG_ALREADY_DOWN	"You are not opped or voiced in $b%s$b."
#define CSMSG_ALREADY_OPCHANNED "There has been no net.join since the opchan in $b%s$b."

/* User management */
#define CSMSG_ADDED_USER	"Added new $b%s$b %s to the %s user list."
#define CSMSG_DELETED_USER	"Deleted %s $b%s$b %s from the %s user list."
#define CSMSG_DELETED_USERS	"Deleted %ss matching '%s' from the %s user list."
#define CSMSG_TRIMMED_USERS	"Trimmed $b%d %ss$b inactive for %s or more from the %s user list."
#define CSMSG_INCORRECT_ACCESS	"%s is %s $b%s$b, not %s %s."
#define CSMSG_USER_EXISTS	"%s is already on the $b%s$b user list (as %s %s)."

#define CSMSG_CANNOT_ADD	"You do not have permission to add users."
#define CSMSG_CANNOT_DEL	"You do not have permission to delete users."
#define CSMSG_CANNOT_CLVL	"You do not have permission to modify users' access."
#define CSMSG_CANNOT_TRIM	"You must supply a duration to trim by."

#define CSMSG_NO_SELF_CLVL	"You cannot change your own access."
#define CSMSG_NO_BUMP_ACCESS	"You cannot give users more access than yourself."
#define CSMSG_USE_ADDOWNER      "Staff may not transfer ownership using $b%s$b; please use $bCLVL$b and/or $bADDOWNER$b instead."
#define CSMSG_NO_TRANSFER_SELF  "You cannot give ownership to your own account."
#define CSMSG_OWNERSHIP_GIVEN   "Ownership of $b%s$b has been transferred to account $b%s$b."

/* Ban management */
#define CSMSG_BAN_ADDED		"Permanently banned $b%s$b."
#define CSMSG_TIMED_BAN_ADDED	"Banned $b%s$b for %s."
#define CSMSG_BAN_DONE		"Banned $b%s$b."
#define CSMSG_BAN_EXTENDED	"Extended ban for $b%s$b expires in %s."
#define CSMSG_BAN_REMOVED	"Matching ban(s) for $b%s$b removed."
#define CSMSG_TRIMMED_BANS	"Trimmed %d bans inactive for %s or more from the %s ban list."
#define CSMSG_REDUNDANT_BAN    	"$b%s$b is already banned."
#define CSMSG_INVALID_DURATION  "Timed bans must last for $b%s$b %s."
#define CSMSG_LAME_MASK 	"$b%s$b is a little too liberal. Try making it more restrictive."
#define CSMSG_MASK_PROTECTED	"Sorry, ban for $b%s$b conflicts with a protected user's hostmask."
#define CSMSG_NO_MATCHING_USERS "No one in $b%s$b has a hostmask matching $b%s$b."
#define CSMSG_BAN_NOT_FOUND	"Sorry, no ban found for $b%s$b."
#define CSMSG_BANLIST_FULL	"The $b%s$b channel ban list is $bfull$b."

#define CSMSG_INVALID_TRIM	"$b%s$b isn't a valid trim target."

/* Channel management */
#define CSMSG_CHANNEL_OPENED	"$b%s$b has been opened."
#define CSMSG_CLEARED_INFO_LINE	"Your info line has been cleared."
#define CSMSG_SET_INFO_LINE    	"Your info line has been set to '%s'."
#define CSMSG_INFO_LINE    	"Your info line is: %s."
#define CSMSG_NO_INFO_LINE    	"You don't have an info line set."

#define CSMSG_TOPIC_SET		"Topic is now '%s'."
#define CSMSG_TOPIC_CLEARED   	"Topic has been reset."
#define CSMSG_NO_TOPIC		"$b%s$b does not have a default topic."
#define CSMSG_TOPIC_LOCKED	"The %s topic is locked."
#define CSMSG_TOPIC_DENY	"You lack permission to change the $b%s$b topic."
#define CSMSG_MASK_BUT_NO_TOPIC "Warning: $b%s$b does not have a default topic, but you just set the topic mask."
#define CSMSG_TOPIC_MISMATCH    "Warning: The default topic for $b%s$b does not match the topic mask; changing it anyway."

#define CSMSG_MODES_INVALID	"$b%s$b is an invalid set of channel modes."
#define CSMSG_MODES_SET		"Channel modes are now $b%s$b."
#define CSMSG_NO_MODES		"$b%s$b does not have any default modes."
#define CSMSG_MODE_LOCKED	"Modes conflicting with $b%s$b are not allowed in %s."
#define CSMSG_CANNOT_SET        "You cannot change the settings level from above your level."
#define CSMSG_BAD_SETLEVEL      "You cannot change the settings level to above your level."

#define CSMSG_USER_PROTECTED   	"Sorry, $b%s$b is protected."
#define CSMSG_OPBY_LOCKED	"You may not op unknown users."
#define CSMSG_PROCESS_FAILED	"$b$C$b could not process some of the nicks you provided."
#define CSMSG_PROCESS_USERS	"%s users in $b%s$b."

#define CSMSG_INVITED_USER	"Invited $b%s$b to join %s."
#define CSMSG_INVITING_YOU	"$b%s$b invites you to join %s."
#define CSMSG_CANNOT_INVITE	"You do not have permission to invite users."
#define CSMSG_ALREADY_PRESENT	"%s is $balready in %s$b."

#define CSMSG_KICK_DONE		"Kicked users in $b%s$b."
#define CSMSG_NO_BANS		"No channel bans found."
#define CSMSG_BANS_REMOVED	"All channel bans removed."

/* Channel userlist */
#define CSMSG_ACCESS_HEADER	"%s $b%ss:$b"
#define CSMSG_USER_HEADER	"$b%-*s%-*s%s$b"
#define CSMSG_USER_DATA		"$b%-*s%-*s$b%s"

/* Channel banlist */
#define CSMSG_BAN_HEADER     	"$b%-*s%-*s%-*s%-*s%s$b"
#define CSMSG_BAN_DATA     	"$b%-*s$b%-*s%-*s%-*s%s"

#define CSMSG_INVALID_ACCESS	"$b%s$b is an invalid access level."
#define CSMSG_CHANGED_ACCESS	"%s is now %s $b%s$b in %s."

/* Channel CTCP reactions */
#define CSMSG_BAD_CTCP_REASON   "CTCPs to this channel are forbidden."

/* Channel note list */
#define CSMSG_NOTELIST_HEADER   "Notes for $b%s$b:"
#define CSMSG_NOTE_FORMAT       "%s (set by %s): %s"
#define CSMSG_NOTELIST_END      "End of notes for $b%s$b."
#define CSMSG_NOTELIST_EMPTY    "There are no (visible) notes for $b%s$b."
#define CSMSG_NO_SUCH_NOTE      "Channel $b%s$b does not have a note named $b%s$b."
#define CSMSG_BAD_NOTE_TYPE     "Note type $b%s$b does not exist."
#define CSMSG_NOTE_SET          "Note $b%s$b set in channel $b%s$b."
#define CSMSG_NOTE_REMOVED      "Note $b%s$b removed in channel $b%s$b."

/* Channel [un]suspension */
#define CSMSG_ALREADY_SUSPENDED	"$b%s$b is already suspended."
#define CSMSG_NOT_SUSPENDED	"$b%s$b is not suspended."
#define CSMSG_SUSPENDED		"$b$C$b access to $b%s$b has been temporarily suspended."
#define CSMSG_SUSPENDED_REASON	"$b$C$b access to $b%s$b has been temporarily suspended (%s)."
#define CSMSG_UNSUSPENDED	"$b$C$b access to $b%s$b has been restored."
#define CSMSG_SUSPEND_NODELETE  "$b%s$b is protected from unregistration, and cannot be suspended."
#define CSMSG_USER_SUSPENDED    "$b%s$b's access to $b%s$b has been suspended."
#define CSMSG_USER_UNSUSPENDED  "$b%s$b's access to $b%s$b has been restored."

/* Access information */
#define CSMSG_IS_CHANSERV	"$b$C$b is the $bchannel service bot$b."
#define CSMSG_USER_HAS_ACCESS	"%s (%s) is %s $b%s$b in %s."
#define CSMSG_USER_IS_HELPING	"%s (%s) has $bsecurity override$b enabled."
#define CSMSG_LAZY_SMURF_TARGET "%s is a megalomaniacal power hungry tyrant ($bIRCOp$b; not logged in)."
#define CSMSG_SMURF_TARGET     	"%s is %s ($bIRCOp$b; on account %s)."
#define CSMSG_TEARDROP_TARGET 	"%s is %s ($bnetwork helper$b; on account %s)."
#define CSMSG_NUKE_TARGET     	"%s is %s ($bsupport helper$b; on account %s)."

/* Seen information */
#define CSMSG_NEVER_SEEN       	"%s has never been seen in $b%s$b."
#define CSMSG_USER_SEEN		"%s was last seen in $b%s$b %s ago."
#define CSMSG_USER_VACATION	"%s is currently on vacation."
#define CSMSG_USER_PRESENT     	"%s is in the channel $bright now$b."

/* Channel information */
#define CSMSG_CHANNEL_INFO	"$b%s$b Information:"
#define CSMSG_CHANNEL_TOPIC	"$bDefault Topic:       $b%s"
#define CSMSG_CHANNEL_MODES	"$bMode Lock:           $b%s"
#define CSMSG_CHANNEL_MAX	"$bMaximum Visitors:    $b%i"
#define CSMSG_CHANNEL_BANS	"$bBan Count:           $b%i"
#define CSMSG_CHANNEL_USERS	"$b%s Count:%*s$b%i"
#define CSMSG_CHANNEL_REGISTRAR "$bRegistrar:           $b%s"
#define CSMSG_CHANNEL_SUSPENDED "$bSuspended:           $bBy %s (%s)"
#define CSMSG_CHANNEL_SUSPENDED2 "$bSuspended:           $bBy %s (%s; %s remaining)"
#define CSMSG_CHANNEL_REGISTERED "$bRegistered:          $b%s ago."
#define CSMSG_CHANNEL_VISITED	"$bVisited:             $b%s ago."

#define CSMSG_PEEK_INFO		"$b%s$b Status:"
#define CSMSG_PEEK_TOPIC       	"$bTopic:          $b%s"
#define CSMSG_PEEK_MODES       	"$bModes:          $b%s"
#define CSMSG_PEEK_OPS       	"$bOps:$b"

/* Network information */
#define CSMSG_NETWORK_INFO	"Network Information:"
#define CSMSG_NETWORK_SERVERS	"$bServers:             $b%i"
#define CSMSG_NETWORK_USERS	"$bTotal Users:         $b%i"
#define CSMSG_NETWORK_BANS	"$bTotal Ban Count:     $b%i"
#define CSMSG_NETWORK_OPERS	"$bIRC Operators:       $b%i"
#define CSMSG_NETWORK_CHANNELS	"$bRegistered Channels: $b%i"
#define CSMSG_SERVICES_UPTIME	"$bServices Uptime:     $b%s"
#define CSMSG_BURST_LENGTH	"$bLast Burst Length:   $b%s"
#define CSMSG_NO_LAST_WRITE     "$bLast Database Write: $bNever"
#define CSMSG_LAST_WRITE	"$bLast Database Write: $b%s ago"

/* Command information */
#define CSMSG_COMMAND_INFO	"Command Information for $b%s$b:"
#define CSMSG_COMMAND_ACCESS	"$bAccess Restrictions: $b%s"
#define CSMSG_COMMAND_WEIGHT	"$bWeight:              $b%.1f"
#define CSMSG_COMMAND_FLAGS     "$bFlags:               $b%li"
#define CSMSG_COMMAND_USES	"$bUses:                $b%d"
#define CSMSG_COMMAND_ALIASES	"$bAliases:             $b%s"

#define CSMSG_COMMAND_IS_ALIAS	"%s is an alias for $b%s$b."

/* Staff list */
#define CSMSG_NETWORK_STAFF	"$bOnline Network Staff:$b"
#define CSMSG_STAFF_OPERS	"$bIRC Operators:$b"
#define CSMSG_STAFF_HELPERS	"$bNetwork Helpers:$b"

/* Channel searches */
#define CSMSG_ACTION_INVALID	"$b%s$b is not a recognized search action."
#define CSMSG_INVALID_CFLAG	"$b%s$b is not a recognized channel flag."
#define CSMSG_UNVISITED_HEADER	"Showing a maximum of %d channels unvisited for $b%s$b:"
#define CSMSG_UNVISITED_DATA	"%s: $b%s$b"
#define CSMSG_SEARCH_DATA	"%s"

/* Security override */
#define CSMSG_ALREADY_HELPING	"You already have security override enabled."
#define CSMSG_ALREADY_NOT_HELPING "You already have security override disabled."
#define CSMSG_NOW_HELPING	"Security override has been enabled."
#define CSMSG_NOW_NOT_HELPING	"Security override has been disabled."

/* Channel configuration */
#define CSMSG_INVALID_OPTION 	"$b%s$b is not a valid option."

#define CSMSG_CHANNEL_OPTIONS	"Channel Options:"
#define CSMSG_STRING_VALUE    	"$b%s$b%s"
#define CSMSG_NUMERIC_VALUE    	"$b%s$b%d - %s"

#define CSMSG_GREETING_TOO_LONG "Your greeting ($b%d$b characters) must be shorter than $b%d$b characters."

#define CSMSG_INVALID_STR_SET 	"$b'%s'$b is an invalid %s setting."
#define CSMSG_INVALID_NUM_SET 	"$b'%d'$b is an invalid %s setting."

#define CSMSG_UNF_RESPONSE      "I don't want to be part of your sick fantasies!"
#define CSMSG_PING_RESPONSE     "Pong!"
#define CSMSG_WUT_RESPONSE      "wut"
#define CSMSG_BAD_NUMBER        "$b%s$b is an invalid number.  Please use a number greater than 1 with this command."

#define CSMSG_TOYS_DISABLED	"Toys are disabled in %s."

/* eject_user and unban_user flags */
#define ACTION_KICK		0x0001
#define ACTION_BAN		0x0002
#define ACTION_ADD_BAN		0x0004
#define ACTION_ADD_TIMED_BAN   	0x0008
#define ACTION_UNBAN		0x0010
#define ACTION_DEL_BAN		0x0020

/* ChanServ command flags */
#define _CMD_REQUIRE_REGISTERED	0x008
#define _CMD_REQUIRE_ACCESS	0x010
#define _CMD_REQUIRE_PRIVILEGE	0x020
#define _CMD_REQUIRE_GODDABLE    0x200

#define CMD_DISABLED		0x001
#define CMD_REQUIRE_HANDLE	0x002
#define CMD_REQUIRE_CHANNEL	0x004
#define CMD_REQUIRE_REGISTERED	(CMD_REQUIRE_CHANNEL | _CMD_REQUIRE_REGISTERED)
#define CMD_REQUIRE_ACCESS	(CMD_REQUIRE_REGISTERED | CMD_REQUIRE_HANDLE | _CMD_REQUIRE_ACCESS)
#define CMD_REQUIRE_PRIVILEGE	(CMD_REQUIRE_HANDLE | _CMD_REQUIRE_PRIVILEGE)
#define CMD_REQUIRE_OPER	0x040
#define CMD_PUBLIC_OVERRIDE	0x080 /* deprecated; was for !god */
#define CMD_IGNORE_EVENT	0x100
#define CMD_REQUIRE_GODDABLE    (CMD_REQUIRE_HANDLE | _CMD_REQUIRE_GODDABLE)
#define CMD_SUSPEND_OVERRIDE	0x400
#define CMD_TOY                 0x800

#define CMD_REQUIRE_DEFAULT	(CMD_REQUIRE_ACCESS)
#define CMD_REQUIRE_LOCK	(CMD_REQUIRE_DEFAULT)

/* The 24 allows for [+-ntlksimp] and lots of fudge factor. */
#define MODELEN			24 + KEYLEN
#define PADLEN			14
#define ACCESSLEN		10

#define CSMSG_NETRIDE_DETECTED  "Netride detected - bye."
#define CSMSG_KICK_REASON	"Bye."
#define CSMSG_KICK_EXTRA	4
#define CSMSG_KICK_FORMAT	"%s (%s)"
#define CSMSG_KICK_ARGUMENTS	reason, nick
#define CSMSG_PROTECT_REASON	"That user is protected."

#define CSFUNC_ARGS		user, channel, argc, argv

static unsigned char *CHANSERV_DEFAULT_MODES = "+nt";

#define IsPrivileged(user) (IsHelping(user) || IsOper(user))

#define CHANSERV_FUNC(NAME) int NAME(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[])
typedef CHANSERV_FUNC(chanserv_func_t);

static CHANSERV_FUNC(cmd_help);
static int eject_user(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[], int action);

#define CHANSERV_SYNTAX() 	chanserv_help(user, argv[0], 0)

#define REQUIRE_PARAMS(N)	if(argc < N) { \
	chanserv_notice(user, MSG_MISSING_PARAMS, argv[0]); \
	CHANSERV_SYNTAX(); \
	return 0; }

extern unsigned int server_count;
extern time_t boot_time;
extern struct policer_params *oper_policer_params, *luser_policer_params, *god_policer_params;

struct userNode *chanserv;
struct helpfile *chanserv_helpfile;
static dict_t chanserv_func_dict;
static dict_t set_func_dict;
dict_t note_types;
static dict_t sChannels;
static dict_t plain_dnrs, mask_dnrs, handle_dnrs;
static dict_t switch_dict; /* holds struct channel_trans* */

static struct
{
    struct chanNode 	*debug_channel;
    struct chanNode     *support_channel;

    unsigned long 	db_backup_frequency;
    unsigned long 	channel_expire_frequency;

    long 		info_delay;
    unsigned int 	adjust_delay;
    long	 	channel_expire_delay;

    unsigned int	adjust_threshold;
    int			join_flood_threshold;

    unsigned int 	greeting_length;
    unsigned int        refresh_period;

    unsigned int        max_owned;
    unsigned int 	max_chan_users;
    unsigned int 	max_chan_bans;

    struct string_list  *eightball;
    unsigned char       trigger_char;
    unsigned int        lame_tricks : 1;

    char                *ctcp_short_ban_duration;
    char                *ctcp_long_ban_duration;

    /* only used if TRANSITION enabled */
    const char		*old_chanserv_name;
    unsigned int	max_switch_load;
    unsigned int	switch_timeout;
    regex_t		access_regex;
    regex_t		auth_regex;
} chanserv_conf;

struct chanserv_cmd
{
    chanserv_func_t	*func;
    enum userLevel	access;
    double		weight;
    long		flags;
    unsigned int	uses;
    unsigned int        refs;
};

struct listData
{
    struct userData	**users;
    unsigned int	matches;

    enum userLevel	lowest;
    enum userLevel	highest;

    unsigned int	handleLength;
};

static unsigned int registered_channels;
static unsigned int userCount[USER_LEVEL_HIGHEST + 1];
static unsigned int banCount;

char *accessNames[] =
{
    "nothing",
    "peon",
    "op",
    "master",
    "coowner",
    "owner",
    "helper"
};

char *accessTitles[] =
{
    "BUG:",
    "Peon",
    "Op",
    "Master",
    "Coowner",
    "Owner",
    "BUG:"
};

char accessChars[] =
{
    'X',
    '+',
    '@',
    '%',
    '*',
    '!',
    'X'
};

int levelMap[] =
{
    0,
    100,
    200,
    300,
    400,
    500,
    600
};

const char *CSMSG_VERSION_ID = "$Id: chanserv.c,v 1.161.2.112.2.87 2002/05/04 01:35:15 entrope Exp $";

struct userData *helperList;
struct chanData *channelList;
static time_t lastwrite;

#define chanserv_notice(target, format...) send_message(target , chanserv , ## format)

#define GetChannelUser(channel, handle) _GetChannelUser(channel, handle, 1, 0)
#define GetChannelAccess(channel, handle) _GetChannelUser(channel, handle, 0, 0)
#define GetTrueChannelAccess(channel, handle) _GetChannelUser(channel, handle, 0, 1)

static enum userLevel
user_level_name(char *name)
{
    enum userLevel level;

    for(level = USER_LEVEL_LOWEST; level <= USER_LEVEL_HIGHEST; level++)
    {
	if(!strcasecmp(name, accessNames[level]))
	{
	    return level;
	}
    }

    return ulNone;
}

static void
chanserv_help(struct userNode *user, const char *topic, int show_aliases)
{
    char *aliases, *resolved = resolve_alias("ChanServ", (char *)topic);

    send_help(user, chanserv, chanserv_helpfile, resolved);
    if(show_aliases && (aliases = command_aliases("ChanServ", resolved)))
    {
	char *c = aliases;

	for(c = aliases; *c; c++)
	{
	    *c = toupper(*c);
	}

	chanserv_notice(user, CSMSG_COMMAND_ALIASES, aliases);
	free(aliases);
    }
}

static struct userData*
_GetChannelUser(struct chanData *channel, struct handle_info *handle, int override, int allow_suspended)
{
    struct userData *uData, **head;

    if(!channel || !handle) return NULL;

    if(override && HANDLE_FLAGGED(handle, HELPING)
       && ((handle->opserv_level > 0) || !IsProtected(channel)))
    {
	for(uData = helperList;
	    uData && uData->handle != handle;
	    uData = uData->next);

	if(!uData)
	{
	    uData = malloc(sizeof(struct userData));
	    uData->handle = handle;

	    uData->access = ulHelper;
	    uData->seen = 0;

	    uData->info = NULL;

	    uData->prev = NULL;
	    uData->next = helperList;
	    if(helperList)
	    {
		helperList->prev = uData;
	    }
	    helperList = uData;
	}

	head = &helperList;
    }
    else
    {
	for(uData = channel->users; uData; uData = uData->next)
        {
            if((uData->handle == handle) && (allow_suspended || !(uData->flags & USER_SUSPENDED))) break;
        }

	head = &(channel->users);
    }

    if(uData && (uData != *head))
    {
	/* Shuffle the user to the head of whatever list he was in. */
	if(uData->next) uData->next->prev = uData->prev;
	if(uData->prev) uData->prev->next = uData->next;

	uData->prev = NULL;
	uData->next = *head;

	if(*head)
	{
	    (**head).prev = uData;
	}
	*head = uData;
    }

    return uData;
}

/* Returns non-zero if user has at least VALUE access (where VALUE is
 * one of 'a', 'p', 'o', 'm', 'c', 'w' or 'n').  exempt_owner is set
 * when handling !set, so the owner can set things to/from 'n'.
 */
static int check_user_level(struct userNode *user, struct chanNode *channel, char value, int use_override, int exempt_owner)
{
    struct userData *uData;
    if(value == 'a') return 1;
    uData = _GetChannelUser(channel->channel_info, user->handle_info, use_override, 0);
    if(!uData) return 0;
    switch(value) {
    case 'c': return uData->access >= ulCoowner;
    case 'm': return uData->access >= ulMaster;
    case 'n': return exempt_owner ? (uData->access >= ulOwner) : 0;
    case 'o': return uData->access >= ulOp;
    case 'p': return uData->access >= ulPeon;
    case 'w': return uData->access >= ulOwner;
    default:  return 1;
    }
}

static struct userData*
require_access(struct chanNode *channel, struct userNode *user, enum userLevel access)
{
    struct userData *uData;

    if(!(uData = GetChannelUser(channel->channel_info, user->handle_info)))
    {
	chanserv_notice(user, CSMSG_NOT_USER, channel->name);
	return NULL;
    }

    if(uData->access < access)
    {
	chanserv_notice(user, CSMSG_NO_ACCESS);
	return NULL;
    }

    return uData;
}

/* Scan for other users authenticated to the same handle
   still in the channel. If so, keep them listed as present.

   user is optional, if not null, it skips checking that userNode
   (for the handle_part function) */
static void
scan_handle_presence(struct chanNode *channel, struct handle_info *handle, struct userNode *user)
{
    struct userData *uData;

    if(!channel->channel_info || IsSuspended(channel->channel_info)) return;

    uData = GetTrueChannelAccess(channel->channel_info, handle);
    if(uData)
    {
        struct modeNode *mn = find_handle_in_channel(channel, handle, user);
        uData->present = mn ? 1 : 0;
    }
}

/* This function is called when a channel command (or
 * channel-qualified privmsg) is directed towards ChanServ. */
static void
chanserv_chanmsg(struct userNode* user, struct chanNode *channel, char *text)
{
    struct userData *uData = NULL;
    struct chanserv_cmd *cmd;
    unsigned char *argv[MAXNUMPARAMS], *tmp, *cmdname, buffer[MAXLEN+200];
    int argc, v = 0, index = 0, argument;

    if(!*text)
    {
	return;
    }

    argc = split_line(text, false, ArrayLength(argv) - 1, argv + 1);
    if(!argc) return;

    if(IsChannelName(argv[1]))
    {
	if(argc < 2)
	{
	    return;
	}

	cmdname = argv[2];
    }
    else
    {
	cmdname = argv[1];
    }


    if(!isalnum(cmdname[0]))
    {
	return;
    }

    cmd = dict_find(chanserv_func_dict, cmdname, NULL);

    if(!cmd)
    {
	chanserv_notice(user, MSG_COMMAND_UNKNOWN, cmdname);
	return;
    }

    if(cmd->flags & CMD_DISABLED)
    {
	chanserv_notice(user, MSG_COMMAND_DISABLED, cmdname);
	return;
    }

    if(!policer_conforms(user->command_policer, now, cmd->weight) && !IsOper(user))
    {
	chanserv_notice(user, MSG_COMMAND_FLOOD);
	return;
    }

    if((cmd->flags & CMD_REQUIRE_HANDLE) && !user->handle_info)
    {
	chanserv_notice(user, MSG_AUTHENTICATE);
	return;
    }

    if(((cmd->flags & _CMD_REQUIRE_PRIVILEGE) && !IsPrivileged(user))
       || ((cmd->flags & CMD_REQUIRE_OPER) && !IsOper(user))
       || ((cmd->flags & _CMD_REQUIRE_GODDABLE) && !IsNetworkHelper(user) && !IsOper(user)))
    {
	chanserv_notice(user, MSG_COMMAND_PRIVILEGED, cmdname);
	return;
    }

    /* We have to put the PubCmd check here, before channel is
       set to the channel in the argument, if any. */
    if(channel && channel->channel_info
       && !check_user_level(user, channel, channel->channel_info->options[optPubCmd], 1, 0))
    {
        chanserv_notice(user, CSMSG_PUBLIC_DENY, channel->name);
        return;
    }

/* This is a funny macro because IsChannelName recognizes +channels,
 * but people sometimes say "!mode +s" in a channel.  Using the normal
 * IsChannelName tends to produce a weird error in this case, so
 * ChanServ does not recognize +channels.
 * Defining it to IsChannelName(N) restores the "weird" behavior.
 */
#define IsCSChannelName(N) (IsChannelName(N) && ((N)[0] != '+'))
    if(argc > 1 && IsCSChannelName(argv[1]))
    {
	/* We have the #channel command form. */
	channel = GetChannel(argv[1]);

	if(channel)
	{
	    argc--;
	    v = 2;
	}
	else
	{
	    tmp = alloca(strlen(argv[1]) + 1);
	    strcpy(tmp, argv[1]);

	    strcpy(argv[1], argv[2]);
	    argv[2] = argv[1] + strlen(argv[1]) + 1;

	    strcpy(argv[2], tmp);

	    v = 1;
	}
    }
    else if(argc > 1 && IsCSChannelName(argv[2]))
    {
	channel = GetChannel(argv[2]);

	if(channel)
	{
	    tmp = alloca(strlen(argv[1]) + 1);
	    strcpy(tmp, argv[1]);

	    memmove(argv[1], argv[2], strlen(argv[2]) + 1);
	    argv[2] = argv[1] + strlen(argv[1]) + 1;

	    strcpy(argv[2], tmp);

	    argc--;
	    v = 2;
	}
	else
	{
	    v = 1;
	}
    }
    else
    {
	argv[0] = 0;
	v += 1;
    }

    if(!channel && (cmd->flags & CMD_REQUIRE_CHANNEL))
    {
	chanserv_notice(user, MSG_INVALID_CHANNEL);
	return;
    }

    if((cmd->flags & _CMD_REQUIRE_REGISTERED) && !channel->channel_info)
    {
	chanserv_notice(user, CSMSG_UNREGISTERED, channel->name);
	return;
    }

    if(!(cmd->flags & CMD_SUSPEND_OVERRIDE) && (cmd->flags & _CMD_REQUIRE_REGISTERED) && IsSuspended(channel->channel_info))
    {
	chanserv_notice(user, CSMSG_SUSPENDED, channel->name);
	return;
    }

    if((cmd->flags & _CMD_REQUIRE_ACCESS) && channel)
    {
	if(!(uData = require_access(channel, user, cmd->access)))
	{
	    return;
	}

	uData->seen = now;
    }

    if((cmd->flags & CMD_TOY))
    {
        if(channel && channel->channel_info && channel->channel_info->options[optToys] == 'd') {
            chanserv_notice(user, CSMSG_TOYS_DISABLED, channel->name);
            return;
        }
        if((v == 2)
           || !channel
           || !channel->channel_info
           || !GetChannelUser(channel->channel_info, user->handle_info)
           || (channel->channel_info->options[optToys] == 'n'))
        {
            channel = NULL;
        }
    }

    if(!(cmd->flags & CMD_IGNORE_EVENT))
    {
	/* Write a log entry, truncated to fit in the 512 byte buffer.
	   It is assumed truncation only needs to take place after the
	   basic elements have been printed.

	   TODO: It might be useful if commands done in public weren't
	   stored in the log buffer (they would still be logged to disk).

	   Example: (#channel) nick[handle]: ban *!*@* EYE R EVUL [override]
	*/

	if(channel)
	{
	    index = sprintf(buffer, "(%s) ", channel->name);
	}

	index += sprintf(buffer + index, "[%s", user->nick);

	if(cmd->flags & CMD_REQUIRE_HANDLE)
	{
	    index += sprintf(buffer + index, ":%s]", user->handle_info->handle);
	}
        else
        {
            buffer[index++] = ']';
        }

	buffer[index++] = ':';
	buffer[index++] = ' ';

	for(argument = v; (argument < argc + v) && (index < MAXLEN); argument++)
	{
	    index += sprintf(buffer + index, "%s ", argv[argument]);
	}
	index--;

	if(cmd->flags & _CMD_REQUIRE_ACCESS && IsHelping(user))
	{
	    /* Make sure the user doesn't have access to the channel
	       before flagging the entry. We don't use uData to check
	       access levels because require_access uses GetChannelUser(). */
	    uData = GetChannelAccess(channel->channel_info, user->handle_info);

	    if(!uData || (uData->access < cmd->access))
	    {
		index += sprintf(buffer + index, " [override]");
	    }
	}

	buffer[index++] = '\n';
        buffer[index] = '\0';
    }

    cmd->uses++;

    if(cmd->func(user, channel, argc, argv + v) && !(cmd->flags & CMD_IGNORE_EVENT))
    {
	log(CS_LOG, LOG_INFO, "%s", buffer);
    }
}

static void
chanserv_privmsg(struct userNode* user, struct userNode* target, char *command, int server_qualified)
{
    (void)server_qualified; (void)target;
    chanserv_chanmsg(user, 0, command);
}

static void
chanserv_ctcp_check(struct userNode *user, struct chanNode *channel, char *text)
{
    size_t tlen;
    char *spc;
    struct userData *uData;
    enum userLevel minlvl;
    unsigned int eflags, argc;
    unsigned char *argv[4];

    /* Bail early if channel is inactive or doesn't restrict CTCPs, or sender is a service */
    if(!channel->channel_info || IsSuspended(channel->channel_info)) return;
    if(channel->channel_info->options[optCTCPUsers] == 'a') return;
    if(IsService(user)) return;
    /* Parse out CTCP type, and exit if it's an ACTION */
    tlen = strlen(text);
    spc = memchr(text, ' ', tlen);
    if(spc) *spc = 0;
    else if(text[tlen-1] == '\001') text[tlen-1] = 0;
    if(!strcasecmp(text, "ACTION")) return;
    /* Figure out the minimum level needed to CTCP the channel */
    switch(channel->channel_info->options[optCTCPUsers]) {
    default: case 'p': minlvl = ulPeon; break;
    case 'o': minlvl = ulOp; break;
    case 'm': minlvl = ulMaster; break;
    case 'c': minlvl = ulCoowner; break;
    case 'w': minlvl = ulOwner; break;
    case 'n': minlvl = INT_MAX; break;
    }
    /* Exit if they're allowed to CTCP the channel. */
    uData = GetChannelUser(channel->channel_info, user->handle_info);
    if(uData && (uData->access >= minlvl)) return;
    /* We need to enforce against them; do so. */
    eflags = 0;
    argv[0] = text;
    argv[1] = user->nick;
    argc = 2;
    if(GetUserMode(channel, user)) eflags |= ACTION_KICK;
    switch(channel->channel_info->options[optCTCPReaction]) {
    default: case 'k': /* just do the kick */ break;
    case 'b':
        eflags |= ACTION_BAN;
        break;
    case 't':
        eflags |= ACTION_BAN | ACTION_ADD_BAN | ACTION_ADD_TIMED_BAN;
        argv[argc++] = chanserv_conf.ctcp_short_ban_duration;
        break;
    case 'T':
        eflags |= ACTION_BAN | ACTION_ADD_BAN | ACTION_ADD_TIMED_BAN;
        argv[argc++] = chanserv_conf.ctcp_long_ban_duration;
        break;
    }
    argv[argc++] = alloca(strlen(CSMSG_BAD_CTCP_REASON)+1);
    strcpy(argv[argc-1], CSMSG_BAD_CTCP_REASON);
    eject_user(chanserv, channel, argc, argv, eflags);
}

static struct note_type *
create_note_type(const char *name)
{
    struct note_type *ntype = calloc(1, sizeof(*ntype) + strlen(name));
    strcpy(ntype->name, name);
    ntype->refs = 1;
    dict_insert(note_types, ntype->name, ntype);
    return ntype;
}

static void
chanserv_deref_note_type(void *data)
{
    struct note_type *ntype = data;

    if(--ntype->refs > 0) return;
    free(ntype);
}

void
chanserv_flush_note_type(struct note_type *ntype)
{
    struct chanData *cData;
    for (cData = channelList; cData; cData = cData->next) {
        dict_remove(cData->notes, ntype->name);
    }
}

static struct note *
chanserv_add_channel_note(struct chanData *channel, struct note_type *type, const char *setter, const char *text)
{
    struct note *note;
    unsigned int len = strlen(text);

    if(len > type->max_length) len = type->max_length;
    note = calloc(1, sizeof(*note) + len);
    note->type = type;
    strncpy(note->setter, setter, sizeof(note->setter)-1);
    memcpy(note->note, text, len);
    note->note[len] = 0;
    dict_insert(channel->notes, type->name, note);
    type->refs++;
    return note;
}

static void
chanserv_free_note(void *data)
{
    struct note *note = data;

    chanserv_deref_note_type(note->type);
    free(note);
}

static int
parse_mode_lock(unsigned char *mode[], int count, struct chanData *channel)
{
    int lock = -1, argo = 1, pos = 0, limit = 0, key = 0;
    long on = 0, off = 0, conflict = 0;

    if(count < 1) return 0;

#define lock_mode(FLAG) do { if(lock == 1) on |= FLAG; else if(lock == 0) off |= FLAG; } while(0)

    while(mode[0][pos])
    {
	switch(mode[0][pos++])
	{
	case '+':
	    lock = 1;
	    break;
	case '-':
	    lock = 0;
	    break;
	case 'n': lock_mode(MODE_NOPRIVMSGS); break;
	case 't': lock_mode(MODE_TOPICLIMIT); break;
	case 's': lock_mode(MODE_SECRET); break;
	case 'p': lock_mode(MODE_PRIVATE); break;
	case 'i': lock_mode(MODE_INVITEONLY); break;
	case 'm': lock_mode(MODE_MODERATED); break;
        case 'D': lock_mode(MODE_DELAYJOINS); break;
	case 'l':
	    if(lock)
	    {
		if(argo >= count) return 0;

		on |= MODE_LIMIT;
		limit = argo++;
	    }
	    else
	    {
		off |= MODE_LIMIT;
	    }
	    break;
	case 'k':
	    if(lock)
	    {
		if(argo >= count) return 0;

		on |= MODE_KEY;
		key = argo++;
	    }
	    else
	    {
		off |= MODE_KEY;
	    }
	    break;
	default:
	    return 0;
	}
    }

    if(lock == -1) return 0;

    if(key) safestrncpy(channel->key, mode[key], sizeof(channel->key));
    if(limit)
    {
	channel->limit = atoi(mode[limit]);

	if(channel->limit < 1)
	{
	    return 0;
	}
    }

    /* strip out any conflicting mode locks ('+p-p', for example). */
    if((conflict = on & off))
    {
	on &= ~conflict;
	off &= ~conflict;
    }

    channel->mode_on = on;
    channel->mode_off = off;

    return 1;
}

static void
create_mode_lock(unsigned char *buffer, struct chanData *channel)
{
    int pos = 0, n;
    long mode = channel->mode_on;

    for(n=0; n<2; n++)
    {
	if(!mode)
	{
	    mode = channel->mode_off;
	    continue;
	}

	buffer[pos++] = n == 0 ? '+' : '-';

	if(mode & MODE_NOPRIVMSGS) buffer[pos++] = 'n';
	if(mode & MODE_TOPICLIMIT) buffer[pos++] = 't';
	if(mode & MODE_SECRET) buffer[pos++] = 's';
	if(mode & MODE_PRIVATE) buffer[pos++] = 'p';
	if(mode & MODE_INVITEONLY) buffer[pos++] = 'i';
	if(mode & MODE_MODERATED) buffer[pos++] = 'm';
        if(mode & MODE_DELAYJOINS) buffer[pos++] = 'D';

	/* parse_mode_lock() restricts limit to positive numbers. */
 	if(mode & MODE_LIMIT) buffer[pos++] = 'l';
	if(mode & MODE_KEY) buffer[pos++] = 'k';

	mode = channel->mode_off;
    }

    if((channel->mode_on & MODE_LIMIT) && channel->limit)
    {
	buffer[pos++] = ' ';

	pos += sprintf(buffer + pos, "%d", channel->limit);
    }

    if((channel->mode_on & MODE_KEY) && channel->key)
    {
	int len = strlen(channel->key);
	buffer[pos++] = ' ';

	memcpy(buffer + pos, channel->key, len);
	pos += len;
    }

    buffer[pos] = 0;
}

static int
append_entry(const char *key, void *data, void *extra)
{
    struct helpfile_expansion *exp = extra;
    int row;

    (void)data;
    row = exp->value.table.length++;
    exp->value.table.contents[row] = calloc(1, sizeof(char*));
    exp->value.table.contents[row][0] = key;
    return 0;
}

static char *
chanserv_expand_command(char *variable)
{
    struct chanserv_cmd *cmd;
    static char value[80];
    extern struct userNode *nickserv;
    char *property;

    cmd = dict_find(chanserv_func_dict, variable, NULL);
    if(!cmd) return "";

    property = variable + strlen(variable) + 1;

    value[0] = 0;
    if(!strcasecmp(property, "weight"))
    {
	snprintf(value, sizeof(value), "%.1f", cmd->weight);
    }
    else if(!strcasecmp(property, "access"))
    {
	if(cmd->flags & CMD_DISABLED)
	{
	    return "This command is disabled.";
	}
	else if(cmd->flags & CMD_REQUIRE_OPER)
	{
	    return "You must be an IRC Operator to use this command.";
	}
	else if(cmd->flags & _CMD_REQUIRE_PRIVILEGE)
	{
	    return "You must be a helper or IRC Operator to use this command.";
	}
        else if(cmd->flags & _CMD_REQUIRE_GODDABLE)
        {
            snprintf(value, sizeof(value), "You must be a helper to use this command.");
        }
	else if(cmd->flags & _CMD_REQUIRE_ACCESS)
	{
	    snprintf(value, sizeof(value), "You must have %s access or greater to use this command.", accessNames[cmd->access]);
	}
	else if(cmd->flags & CMD_REQUIRE_HANDLE)
	{
	    snprintf(value, sizeof(value), "You must be authenticated to %s to use this command.", nickserv->nick);
	}
	else
	{
	    return "Anyone may use this command.";
	}
    }

    return value;
}

static char max_length_text[MAXLEN+1][16];

static struct helpfile_expansion
chanserv_expand_variable(char *variable)
{
    struct helpfile_expansion exp;
    char *start;
    int count;

    if(!strcasecmp(variable, "index"))
    {
        exp.type = HF_TABLE;
        exp.value.table.length = 1;
        exp.value.table.width = 1;
        exp.value.table.flags = TABLE_REPEAT_ROWS;
        exp.value.table.contents = calloc(dict_size(chanserv_func_dict)+1, sizeof(char**));
        exp.value.table.contents[0] = calloc(1, sizeof(char*));
        exp.value.table.contents[0][0] = "Commands:";
        dict_foreach(chanserv_func_dict, append_entry, &exp);
        return exp;
    }
    else if(!strcasecmp(variable, "notes"))
    {
        dict_iterator_t it;
        exp.type = HF_TABLE;
        exp.value.table.length = 1;
        exp.value.table.width = 3;
        exp.value.table.flags = TABLE_NO_FREE;
        exp.value.table.contents = calloc(dict_size(note_types)+1, sizeof(char**));
        exp.value.table.contents[0] = calloc(exp.value.table.width, sizeof(char*));
        exp.value.table.contents[0][0] = "Note Type";
        exp.value.table.contents[0][1] = "Visibility";
        exp.value.table.contents[0][2] = "Max Length";
        for (it=dict_first(note_types); it; it=iter_next(it))
        {
            struct note_type *ntype = iter_data(it);
            int row;

            if(ntype->visible_type == NOTE_VIS_PRIVILEGED) continue;
            row = exp.value.table.length++;
            exp.value.table.contents[row] = calloc(exp.value.table.width, sizeof(char*));
            exp.value.table.contents[row][0] = ntype->name;
            exp.value.table.contents[row][1] = (ntype->visible_type == NOTE_VIS_ALL) ? "all" :
                (ntype->visible_type == NOTE_VIS_CHANNEL_USERS) ? "chan users" :
                "unknown";
            if(!max_length_text[ntype->max_length][0])
            {
                snprintf(max_length_text[ntype->max_length], sizeof(max_length_text[ntype->max_length]),
                         "%u", ntype->max_length);
            }
            exp.value.table.contents[row][2] = max_length_text[ntype->max_length];
        }
        return exp;
    }

    exp.type = HF_STRING;
    exp.value.str = NULL;
    start = variable;
    count = 0;
    while(*variable)
    {
        while(*variable != '/' && *variable) variable++;
	while(*variable == '/') *variable++ = 0;
	if(!*variable) break;
	count++;
    }

    if(!strcasecmp(start, "command"))
    {
	exp.value.str = strdup(chanserv_expand_command(start + strlen(start) + 1));
    }

    return exp;
}

static struct chanData*
register_channel(char *registrar, char *topic, char *greeting, char *modes, char *options, long flags, struct chanNode *cNode)
{
    struct chanData *channel;

    channel = calloc(1, sizeof(struct chanData));

    channel->registrar = registrar ? strdup(registrar) : NULL;
    channel->greeting = greeting ? strdup(greeting) : NULL;
    channel->registered = now;
    channel->visited = channel->registered;
    channel->limitAdjusted = channel->registered;
    channel->flags = flags ? flags : CHANNEL_DEFAULT_FLAGS;
    channel->notes = dict_new();
    dict_set_free_data(channel->notes, chanserv_free_note);

    safestrncpy(channel->options, options ? options : CHANNEL_DEFAULT_OPTIONS, sizeof(channel->options));

    if(strlen(channel->options) < (OPTION_STRING_SIZE - 1))
    {
	/* Use a sane default from the DEFAULT_OPTIONS in the
	   case of missing options. */
	memcpy(channel->options + strlen(channel->options), CHANNEL_DEFAULT_OPTIONS + strlen(channel->options), (OPTION_STRING_SIZE - 1) - strlen(channel->options));
    }

    if(topic)
    {
	safestrncpy(channel->topic, topic, sizeof(channel->topic));
    }

    if(modes)
    {
	unsigned char *split_modes[3];
	int count;

	count = split_line(modes, 0, 3, split_modes);
	parse_mode_lock(split_modes, count, channel);
    }

    channel->prev = NULL;
    channel->next = channelList;

    if(channelList)
    {
	channelList->prev = channel;
    }
    channelList = channel;

    channel->channel = cNode;

    if(cNode)
    {
	cNode->channel_info = channel;
    }

    registered_channels++;

    return channel;
}

static struct userData*
add_channel_user(struct chanData *channel, struct handle_info *handle, enum userLevel access, time_t seen, const char *info)
{
    struct userData *ud;
    struct chanList *cl;

    if(access > USER_LEVEL_HIGHEST || access < USER_LEVEL_LOWEST)
    {
	return NULL;
    }

    ud = calloc(1, sizeof(*ud));
    ud->handle = handle;
    ud->seen = seen;
    ud->access = access;
    ud->info = info ? strdup(info) : NULL;

    ud->prev = NULL;
    ud->next = channel->users;
    if(channel->users)
    {
	channel->users->prev = ud;
    }
    channel->users = ud;

    channel->userCount[0]++;
    channel->userCount[access]++;
    userCount[0]++;
    userCount[access]++;

    cl = malloc(sizeof(struct chanList));
    cl->channel = channel;
    cl->user = ud;

    cl->prev = NULL;
    cl->next = ud->handle->channels;
    if(cl->next)
    {
	cl->next->prev = cl;
    }
    ud->handle->channels = cl;

    return ud;
}

static struct chanList *
find_channel_entry(struct chanData *channel, struct userData *user)
{
    struct chanList *entry = NULL;
	
    for(entry = user->handle->channels; entry; entry = entry->next)
    {
	if(entry->channel == channel)
	{
	    return entry;
	}
    }

    return NULL;
}

void
del_channel_user(struct chanData *channel, struct userData *user)
{
    struct chanList *entry = NULL;

    if(!channel || !user) return;

    channel->userCount[0]--;
    channel->userCount[user->access]--;
    userCount[0]--;
    userCount[user->access]--;

    entry = find_channel_entry(channel, user);
    if(entry)
    {
	if(entry->prev) entry->prev->next = entry->next;
	else user->handle->channels = entry->next;

	if(entry->next) entry->next->prev = entry->prev;

	free(entry);
    }
    else
    {
	log(CS_LOG, LOG_ERROR, "Missing access list entry for %s.\n", user->handle->handle);
    }

    if(user->prev) user->prev->next = user->next;
    else channel->users = user->next;

    if(user->next) user->next->prev = user->prev;

    if(user->info) free(user->info);
    free(user);
}

static void expire_ban(void *data);

static struct banData*
add_channel_ban(struct chanData *channel, const char *mask, char *owner, time_t set, time_t triggered, time_t expires, char *reason)
{
    struct banData *bd;

    if(!mask) return NULL;

    bd = malloc(sizeof(struct banData));

    bd->channel = channel;
    bd->set = set;
    bd->triggered = triggered;
    bd->expires = expires;

    safestrncpy(bd->mask, mask, sizeof(bd->mask));
    if(owner) safestrncpy(bd->owner, owner, sizeof(bd->owner));
    bd->reason = reason ? strdup(reason) : NULL;

    if(expires)
    {
	timeq_add(expires, expire_ban, bd);
    }

    bd->prev = NULL;
    bd->next = channel->bans;
    if(channel->bans)
    {
	channel->bans->prev = bd;
    }
    channel->bans = bd;
    channel->banCount++;
    banCount++;

    return bd;
}

static void
del_channel_ban(struct banData *ban)
{
    if(!ban) return;

    ban->channel->banCount--;
    banCount--;

    if(ban->prev) ban->prev->next = ban->next;
    else ban->channel->bans = ban->next;

    if(ban->next) ban->next->prev = ban->prev;

    if(ban->expires)
    {
	timeq_del(0, expire_ban, ban, TIMEQ_IGNORE_WHEN);
    }

    if(ban->reason) free(ban->reason);

    free(ban);
}

static void
expire_ban(void *data)
{
    struct banData *bd;
    struct banList bans;
    struct chanData *channel;
    struct string_list *removed;
    unsigned int i;
    char *ban;

    bd = data;
    channel = bd->channel;

    if(IsSuspended(channel))
    {
	bd->expires = 0;
	del_channel_ban(bd);

	return;
    }

    bans = channel->channel->banlist;
    removed = alloc_string_list(2);
    for(i=0; i<bans.used; i++)
    {
	if(!strcmp(bans.list[i]->ban, bd->mask))
	{
	    ban = strdup(bans.list[i]->ban);
	    string_list_append(removed, ban);
	}
    }

    bd->expires = 0;
    del_channel_ban(bd);

    if(removed->used)
    {
	DelChannelBan(removed->used, removed->list, channel->channel, chanserv, 1);
    }

    free_string_list(removed);
}

static void chanserv_expire_suspension(void *data);

static void
unregister_channel(struct chanData *channel)
{
    /* After channel unregistration, the following must be cleaned
       up:
       - Channel information.
       - Channel users.
       - Channel bans.
       - Channel suspension data.
       - Timeq entries. (Except timed bans, which are handled elsewhere.)
    */

    if(!channel)
    {
	return;
    }

    timeq_del(0, NULL, channel, TIMEQ_IGNORE_FUNC | TIMEQ_IGNORE_WHEN);

    while(channel->users)
    {
	del_channel_user(channel, channel->users);
    }

    while(channel->bans)
    {
	del_channel_ban(channel->bans);
    }

    if(channel->registrar) free(channel->registrar);
    if(channel->greeting) free(channel->greeting);

    if(channel->prev) channel->prev->next = channel->next;
    else channelList = channel->next;

    if(channel->next) channel->next->prev = channel->prev;

    if(channel->suspended)
    {
	struct chanNode *cNode = dict_find(channels, channel->suspended->name, NULL);
	struct suspended *suspended = channel->suspended;

	free(suspended->name);
	free(suspended->suspender);
        if(suspended->reason) free(suspended->reason);
        if(suspended->expires) timeq_del(suspended->expires, chanserv_expire_suspension, suspended, 0);
	free(suspended);

	if(cNode)
	{
	    cNode->channel_info = NULL;
	}
    }
    else
    {
	channel->channel->channel_info = NULL;
    }

    if(channel->notes)
    {
        dict_delete(channel->notes);
    }

    free(channel);

    registered_channels--;
}

static void
expire_channels(void *data)
{
    struct chanData *channel, *next;
    struct chanNode *save;
    char delay[INTERVALLEN], reason[INTERVALLEN + 64];
    (void)data;

    intervalString(delay, chanserv_conf.channel_expire_delay);
    sprintf(reason, "Channel registration automatically expired after %s of disuse.", delay);

    for(channel = channelList; channel; channel = next)
    {
	next = channel->next;

	if(IsProtected(channel))
	{
	    continue;
	}

	if((now - channel->visited) > chanserv_conf.channel_expire_delay)
	{
            int suspended = IsSuspended(channel);
	    log(CS_LOG, LOG_INFO, "(%s) Channel registration expired.\n", (suspended ? channel->suspended->name : channel->channel->name));

	    /* Because we have a chanData * and not a chanNode *, save
	       the chanNode *, then call unregister_channel(). */
	    save = channel->channel;
	    unregister_channel(channel);

	    if(!suspended)
	    {
		DelChannelUser(chanserv, save, reason, 0);
	    }
	}
    }

    if(chanserv_conf.channel_expire_frequency)
    {
	timeq_add(now + chanserv_conf.channel_expire_frequency, expire_channels, NULL);
    }
}

static int
protect_user(struct userNode *victim, struct userNode *aggressor, struct chanData *channel)
{
    char protect = channel->options[optProtect];
    struct userData *cs_victim, *cs_aggressor;

    if(protect == 'n' || victim == aggressor || IsOper(aggressor))
    {
	return 0;
    }

    if(!(cs_victim = GetChannelAccess(channel, victim->handle_info))) return 0;
    if(!(cs_aggressor = GetChannelAccess(channel, aggressor->handle_info))) return 1;

    switch(protect)
    {
    case 'l':
	if(cs_victim->access > cs_aggressor->access) return 1;
	break;
    case 'e':
	if(cs_victim->access >= cs_aggressor->access) return 1;
	break;
    }

    return 0;
}

static int
validate_op(struct userNode *user, struct chanNode *channel, struct userNode *victim)
{
    struct userData *cs_victim;

    if((cs_victim = GetChannelUser(channel->channel_info, victim->handle_info))
       && (cs_victim->access >= ulOp)) return 1;
    if(!check_user_level(user, channel, channel->channel_info->options[optStrictOp], 1, 0))
    {
	chanserv_notice(user, CSMSG_OPBY_LOCKED);
	return 0;
    }

    return 1;
}

static int
validate_auto_op(struct userNode *user, struct chanNode *channel)
{
    struct userData *uData;

    if(!check_user_level(user, channel, channel->channel_info->options[optAutoOp], 0, 0)) return 0;
    uData = GetChannelAccess(channel->channel_info, user->handle_info);
    return uData && !(uData->flags & USER_AUTO_OP);
}

static int
validate_deop(struct userNode *user, struct chanNode *channel, struct userNode *victim)
{
    if(IsService(victim))
    {
	chanserv_notice(user, MSG_SERVICE_IMMUNE, victim->nick);
	return 0;
    }

    if(protect_user(victim, user, channel->channel_info))
    {
	chanserv_notice(user, CSMSG_USER_PROTECTED, victim->nick);
	return 0;
    }

    return 1;
}

#define chanserv_get_handle_info(USER, NAME) smart_get_handle_info(chanserv, USER, NAME)

static struct do_not_register *
chanserv_add_dnr(const char *chan_name, const char *setter, const char *reason)
{
    struct do_not_register *dnr = calloc(1, sizeof(*dnr)+strlen(reason));
    strcpy(dnr->chan_name, chan_name);
    strcpy(dnr->setter, setter);
    strcpy(dnr->reason, reason);
    if(dnr->chan_name[0] == '*')
    {
        dict_insert(handle_dnrs, dnr->chan_name+1, dnr);
    }
    else if(strpbrk(dnr->chan_name, "*?"))
    {
        dict_insert(mask_dnrs, dnr->chan_name, dnr);
    }
    else
    {
        dict_insert(plain_dnrs, dnr->chan_name, dnr);
    }
    return dnr;
}

static struct do_not_register *
chanserv_is_dnr(const char *chan_name, struct handle_info *handle)
{
    struct do_not_register *dnr;
    dict_iterator_t it;

    if(handle && (dnr = dict_find(handle_dnrs, handle->handle, NULL))) return dnr;
    if((dnr = dict_find(plain_dnrs, chan_name, NULL))) return dnr;
    for(it = dict_first(mask_dnrs); it; it = iter_next(it))
    {
        if(match_ircglob(chan_name, iter_key(it))) return iter_data(it);
    }
    return NULL;
}

static CHANSERV_FUNC(cmd_noregister)
{
    const char *chan_name = NULL;
    unsigned int arg = 1;
    struct do_not_register *dnr;

    if(channel)
    {
        if((argc > 1) && (argv[1][0] == '*')) {
            chan_name = argv[1];
            arg = 2;
        } else {
            chan_name = channel->name;
        }
    }
    else if(argc > 1)
    {
        if(!IsCSChannelName(argv[1]) && (argv[1][0] != '*'))
        {
            chanserv_notice(user, MSG_NOT_CHANNEL_NAME);
            return 0;
        }
        chan_name = argv[1];
        arg = 2;
    }
    if(chan_name && (arg < argc))
    {
        unsplit_string(argv+arg, argc-arg);
        if((argv[arg][0] == '*') && !get_handle_info(argv[arg]+1))
        {
            chanserv_notice(user, MSG_HANDLE_UNKNOWN, argv[arg]+1);
        }
        else
        {
            chanserv_add_dnr(chan_name, user->handle_info->handle, argv[arg]);
            chanserv_notice(user, CSMSG_NOREGISTER_CHANNEL, chan_name);
        }
    }
    else if((dnr = chanserv_is_dnr(chan_name, 0)))
    {
        chanserv_notice(user, CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
    }
    else
    {
        dict_iterator_t it;

        for (it = dict_first(handle_dnrs); it; it = iter_next(it))
        {
            dnr = iter_data(it);
            chanserv_notice(user, CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
        }
        for (it = dict_first(plain_dnrs); it; it = iter_next(it))
        {
            dnr = iter_data(it);
            chanserv_notice(user, CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
        }
        for (it = dict_first(mask_dnrs); it; it = iter_next(it))
        {
            dnr = iter_data(it);
            chanserv_notice(user, CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
        }
        chanserv_notice(user, CSMSG_END_DNR_LIST);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_allowregister)
{
    const char *chan_name = NULL;
    int arg = 1;

    if(argc > 1)
    {
        chan_name = argv[1];
        arg = 2;
    }
    else if(channel)
    {
        chan_name = channel->name;
    }
    if(!chan_name)
    {
        chanserv_notice(user, MSG_MISSING_PARAMS, argv[0]);
        CHANSERV_SYNTAX();
        return 0;
    }
    if((chan_name[0] == '*') && dict_find(handle_dnrs, chan_name+1, NULL))
    {
        dict_remove(handle_dnrs, chan_name+1);
        chanserv_notice(user, CSMSG_DNR_REMOVED, chan_name);
    }
    else if(dict_find(plain_dnrs, chan_name, NULL))
    {
        dict_remove(plain_dnrs, chan_name);
        chanserv_notice(user, CSMSG_DNR_REMOVED, chan_name);
    }
    else if(dict_find(mask_dnrs, chan_name, NULL))
    {
        dict_remove(mask_dnrs, chan_name);
        chanserv_notice(user, CSMSG_DNR_REMOVED, chan_name);
    }
    else
    {
        chanserv_notice(user, CSMSG_NO_SUCH_DNR, chan_name);
        return 0;
    }
    return 1;
}

static CHANSERV_FUNC(cmd_register)
{
    struct handle_info *handle;
    struct chanData *cData;
    struct chanList *cList;
    struct modeNode *mn;
    char *reason, *chan_name;
    unsigned int new_channel, force=0, owned;
    struct do_not_register *dnr;

    if(channel)
    {
        if(channel->channel_info)
        {
            chanserv_notice(user, CSMSG_ALREADY_REGGED, channel->name);
            return 0;
        }

        if(channel->bad_channel)
        {
            chanserv_notice(user, CSMSG_ILLEGAL_CHANNEL, channel->name);
            return 0;
        }

        if(!IsPrivileged(user) && (!(mn = GetUserMode(channel, user)) || !(mn->modes & MODE_CHANOP)))
        {
            chanserv_notice(user, CSMSG_MUST_BE_OPPED, channel->name);
            return 0;
        }

        new_channel = 0;
        chan_name = channel->name;
    }
    else
    {
        struct suspended *suspended;

        if(!IsCSChannelName(argv[1]))
        {
            chanserv_notice(user, MSG_NOT_CHANNEL_NAME);
            return 0;
        }

        if(opserv_bad_channel(argv[1]))
        {
            chanserv_notice(user, CSMSG_ILLEGAL_CHANNEL, argv[1]);
            return 0;
        }

        if((suspended = dict_find(sChannels, argv[1], NULL)))
        {
            char *reason = suspended->reason ? suspended->reason : "no reason";
            chanserv_notice(user, CSMSG_SUSPENDED_REASON, argv[1], reason);
            return 0;
        }

        new_channel = 1;
        chan_name = argv[1];
    }

    if(argc >= (new_channel+2))
    {
	if(!IsPrivileged(user))
	{
	    chanserv_notice(user, CSMSG_PROXY_FORBIDDEN);
	    return 0;
	}

	if(!(handle = chanserv_get_handle_info(user, argv[new_channel+1]))) return 0;
        force = (argc > (new_channel+2)) && !strcasecmp(argv[new_channel+2], "force");
        dnr = chanserv_is_dnr(chan_name, handle);
    }
    else
    {
	handle = user->handle_info;
        dnr = chanserv_is_dnr(chan_name, handle);
    }
    if(dnr && !force)
    {
        if(IsPrivileged(user))
        {
            chanserv_notice(user, CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
        }
        else
        {
                chanserv_notice(user, CSMSG_DNR_CHANNEL, chan_name);
        }
        return 0;
    }


    for(owned=0, cList=handle->channels; cList; cList=cList->next)
    {
        if(cList->user->access == ulOwner) owned++;
    }
    if((owned >= chanserv_conf.max_owned) && !force)
    {
        chanserv_notice(user, CSMSG_OWN_TOO_MANY, handle->handle, chanserv_conf.max_owned);
        return 0;
    }

    if(new_channel) channel = AddChannel(argv[1], now, "", "", 0, chanserv->numeric, NULL);

    cData = register_channel(user->handle_info->handle, NULL, NULL, NULL, NULL, 0, channel);
    add_channel_user(cData, handle, USER_LEVEL_HIGHEST, 0, NULL);
    scan_handle_presence(channel, handle, NULL);

    if(!new_channel) AddChannelUser(chanserv, channel);
    AddChannelOp(1, &chanserv, channel, chanserv, !new_channel);

    parse_mode_lock(&CHANSERV_DEFAULT_MODES, 1, cData);
    mod_chanmode(channel, CHANSERV_DEFAULT_MODES, NULL, 0);
    irc_mode(chanserv, channel, CHANSERV_DEFAULT_MODES);

    /* Initialize the channel's max user record. */
    cData->max = channel->members.used;

    if(handle != user->handle_info)
    {
        chanserv_notice(user, CSMSG_PROXY_SUCCESS, handle->handle, channel->name);
    }
    else
    {
        chanserv_notice(user, CSMSG_REG_SUCCESS, channel->name);
    }

    reason = alloca(strlen(user->nick) + strlen(channel->name) + strlen(handle->handle) + 32);
    sprintf(reason, "%s registered to %s by %s.", channel->name, handle->handle, user->nick);
    global_message(MESSAGE_RECIPIENT_OPERS | MESSAGE_RECIPIENT_HELPERS, reason);
    return 1;
}

static CHANSERV_FUNC(cmd_unregister)
{
    char *reason, *name;
    int length, offset;
    struct chanData *cData;

    if(IsProtected(channel->channel_info))
    {
	chanserv_notice(user, CSMSG_UNREG_NODELETE, channel->name);
	return 0;
    }

    if(!IsPrivileged(user))
    {
	struct chanserv_cmd *command = dict_find(chanserv_func_dict, argv[0], NULL);
	enum userLevel level = command ? command->access : ulOwner;

	if(!require_access(channel, user, level))
	{
	    return 0;
	}

	if(argc < 2 || strcmp(argv[1], "CONFIRM"))
	{
	    chanserv_notice(user, CSMSG_CONFIRM_UNREG);
	    return 0;
	}
    }

    length = strlen(channel->name);
    offset = length > 7 ? length : 7;

    reason = alloca(strlen(user->nick) + offset + 32);
    sprintf(reason, "%s unregistered by %s.", "Channel", user->nick);

    name = alloca(length + 1);
    safestrncpy(name, channel->name, length + 1);

    cData = channel->channel_info;
    if(IsSuspended(cData))
    {
	dict_remove(sChannels, channel->name);
    }
    else
    {
	DelChannelUser(chanserv, channel, reason, 0);
    }
    unregister_channel(cData);
    chanserv_notice(user, CSMSG_UNREG_SUCCESS, name);

    sprintf(reason, "%s unregistered by %s.", name, user->nick);
    global_message(MESSAGE_RECIPIENT_OPERS | MESSAGE_RECIPIENT_HELPERS, reason);

    return 1;
}

static CHANSERV_FUNC(cmd_move)
{
    struct chanNode *target;
    struct modeNode *mn;
    int new_channel = 0;
    char reason[MAXLEN];

    REQUIRE_PARAMS(2);

    if(IsProtected(channel->channel_info))
    {
	chanserv_notice(user, CSMSG_MOVE_NODELETE, channel->name);
	return 0;
    }

    if(!(target = GetChannel(argv[1])))
    {
        if(!IsCSChannelName(argv[1]))
        {
            chanserv_notice(user, MSG_NOT_CHANNEL_NAME);
            return 0;
        }

        if(opserv_bad_channel(argv[1]))
        {
            chanserv_notice(user, CSMSG_ILLEGAL_CHANNEL, argv[1]);
            return 0;
        }

        if(chanserv_is_dnr(argv[1], NULL))
        {
            chanserv_notice(user, CSMSG_DNR_CHANNEL, argv[1]);
            return 0;
        }

        target = AddChannel(argv[1], now, "", "", 0, chanserv->numeric, NULL);;
        new_channel = 1;
    }

    if(target->channel_info)
    {
	chanserv_notice(user, CSMSG_ALREADY_REGGED, target->name);
	return 0;
    }

    if(!new_channel && (!(mn = GetUserMode(target, user)) || !(mn->modes && MODE_CHANOP)) && !IsPrivileged(user))
    {
	chanserv_notice(user, CSMSG_MUST_BE_OPPED, target->name);
	return 0;
    }

    /* Move the channel_info to the target channel; it
       shouldn't be necessary to clear timeq callbacks
       for the old channel. */
    target->channel_info = channel->channel_info;
    target->channel_info->channel = target;
    channel->channel_info = NULL;

    if(IsSuspended(target->channel_info))
    {
	/* Transfer the suspension information. */
	struct suspended *suspended = target->channel_info->suspended;

	free(suspended->name);
	suspended->name = strdup(target->name);
    }
    else
    {
	if(!new_channel) AddChannelUser(chanserv, target);
	AddChannelOp(1, &chanserv, target, chanserv, 1);
    }

    chanserv_notice(user, CSMSG_MOVE_SUCCESS, target->name);

    if(!IsSuspended(target->channel_info))
    {
	sprintf(reason, "Channel moved to %s by %s.", target->name, user->nick);
	DelChannelUser(chanserv, channel, reason, 0);
    }

    sprintf(reason, "%s moved to %s by %s.", channel->name, target->name, user->nick);
    global_message(MESSAGE_RECIPIENT_OPERS | MESSAGE_RECIPIENT_HELPERS, reason);

    return 1;
}

static void
merge_users(struct chanData *source, struct chanData *target)
{
    struct userData *suData, *tuData, *next;
    dict_iterator_t it;
    dict_t merge;

    merge = dict_new();

    /* Insert the source's users into the scratch area. */
    for(suData = source->users; suData; suData = suData->next)
    {
	dict_insert(merge, suData->handle->handle, suData);
    }

    /* Iterate through the target's users, looking for
       users common to both channels. The lower access is
       removed from either the scratch area or target user
       list. */
    for(tuData = target->users; tuData; tuData = next)
    {
	struct userData *choice;

	next = tuData->next;

	/* If a source user exists with the same handle as a target
	   channel's user, resolve the conflict by removing one. */
	suData = dict_find(merge, tuData->handle->handle, NULL);
	if(!suData)
	{
	    continue;
	}

	/* Pick the data we want to keep. */
	if(suData->access == tuData->access)
	{
	    /* If the access is the same, use the later seen time. */
	    choice = (suData->seen > tuData->seen) ? suData : tuData;
	}
	else
	{
	    /* Otherwise, keep the higher access level. */
	    choice = (suData->access > tuData->access) ? suData : tuData;
	}

	/* Remove the user that wasn't picked. */
	if(choice == tuData)
	{
	    dict_remove(merge, suData->handle->handle);
	    del_channel_user(source, suData);
	}
	else
	{
	    del_channel_user(target, tuData);
	}
    }

    /* Move the remaining users to the target channel. */
    for(it = dict_first(merge); it; it = iter_next(it))
    {
	struct chanList *entry;

	suData = iter_data(it);

	/* Insert the user into the target channel's linked list. */
	suData->prev = NULL;
	suData->next = target->users;

	if(target->users)
	{
	    target->users->prev = suData;
	}
	target->users = suData;

	/* Update the user counts for the target channel; the
	   source counts are left alone. */
	target->userCount[0]++;
	target->userCount[suData->access]++;

	/* Update the handle-channel link to point to the target. */
	entry = find_channel_entry(source, suData);
	if(entry)
	{
	    entry->channel = target;
	}
	else
	{
	    log(CS_LOG, LOG_ERROR, "Missing access list entry for %s.\n", suData->handle->handle);
	}
    }

    /* Possible to assert (source->users == NULL) here. */
    source->users = NULL;
    dict_delete(merge);
}

static void
merge_bans(struct chanData *source, struct chanData *target)
{
    struct banData *sbData, *tbData, *sNext, *tNext;

    /* Perform a totally expensive O(n*m) merge, ick. */
    for(sbData = source->bans; sbData; sbData = sNext)
    {
	sNext = sbData->next;

	for(tbData = target->bans; tbData; tbData = tNext)
	{
	    tNext = tbData->next;

	    /* Perform two comparisons between each source
	       and target ban, conflicts are resolved by
	       removing the stricter ban, but keeping the
	       later expiration and triggered time. */
	    if(match_ircglobs(tbData->mask, sbData->mask))
	    {
		/* There is a broader ban in the target channel that
		   overrides one in the source channel; remove the 
		   source ban. */
		if(sbData->expires > tbData->expires)
		{
		    tbData->expires = sbData->expires;
		}
		if(sbData->triggered > tbData->triggered)
		{
		    tbData->triggered = sbData->triggered;
		}
		del_channel_ban(sbData);
		break;
	    }
	    else if(match_ircglobs(sbData->mask, tbData->mask))
	    {
		/* There is a broader ban in the source channel that
		   overrides one in the target channel; remove the
		   target ban and move the source over. */
		if(tbData->expires > sbData->expires)
		{
		    sbData->expires = tbData->expires;
		}
		if(tbData->triggered > sbData->triggered)
		{
		    sbData->triggered = tbData->triggered;
		}
		del_channel_ban(tbData);

		/* Modify the ban's associated channel. */
		sbData->channel = target;

		/* Insert the ban into the target channel's linked list. */
		sbData->prev = NULL;
		sbData->next = target->bans;

		if(target->bans)
		{
		    target->bans->prev = sbData;
		}
		target->bans = sbData;

		/* Update the user counts for the target channel. */
		target->banCount++;
	    }
	}
    }

    /* Possible to assert (source->bans == NULL) here. */
    source->bans = NULL;
}

static void
merge_data(struct chanData *source, struct chanData *target)
{
    if(source->visited > target->visited)
    {
	target->visited = source->visited;
    }
}

static void
merge_channel(struct chanData *source, struct chanData *target)
{
    merge_users(source, target);
    merge_bans(source, target);
    merge_data(source, target);
}

static CHANSERV_FUNC(cmd_merge)
{
    struct chanNode *target;
    struct userData *uData;
    char *reason;

    REQUIRE_PARAMS(2);

    /* Make sure the target channel exists and is registered to the user
       performing the command. */
    if(!(target = GetChannel(argv[1])))
    {
	chanserv_notice(user, MSG_INVALID_CHANNEL);
	return 0;
    }

    if(!target->channel_info)
    {
	chanserv_notice(user, CSMSG_UNREGISTERED, target->name);
	return 0;
    }

    if(IsSuspended(target->channel_info))
    {
	chanserv_notice(user, CSMSG_MERGE_SUSPENDED);
	return 0;
    }

    if(!(uData = GetChannelUser(target->channel_info, user->handle_info)) || (uData->access < ulOwner))
    {
	chanserv_notice(user, CSMSG_ACCESS_REQUIRED, "owner", target->name);
	return 0;
    }

    /* Merge the channel structures and associated data. */
    merge_channel(channel->channel_info, target->channel_info);

    reason = alloca(strlen(target->name) + strlen(user->nick) + 32);
    sprintf(reason, "Channel merged into %s by %s.", target->name, user->nick);

    unregister_channel(channel->channel_info);
    DelChannelUser(chanserv, channel, reason, 0);

    chanserv_notice(user, CSMSG_MERGE_SUCCESS, target->name);
    return 1;
}

static CHANSERV_FUNC(cmd_opchan)
{
    (void)argv; (void)argc;
    if(!IsPrivileged(user) && !channel->channel_info->may_opchan)
    {
        chanserv_notice(user, CSMSG_ALREADY_OPCHANNED, channel->name);
        return 0;
    }
    channel->channel_info->may_opchan = 0;
    AddChannelOp(1, &chanserv, channel, NULL, 1);
    return 1;
}

static int
cmd_add_user(struct userNode *user, struct chanNode *channel, enum userLevel access, unsigned int argc, unsigned char *argv[])
{
    struct userData *actee;
    struct userData *actor;
    struct handle_info *handle;

    REQUIRE_PARAMS(2);

    actor = GetChannelUser(channel->channel_info, user->handle_info);
    if(actor->access < ulMaster)
    {
	chanserv_notice(user, CSMSG_CANNOT_ADD);
	return 0;
    }

    if(channel->channel_info->userCount[0] >= chanserv_conf.max_chan_users)
    {
	chanserv_notice(user, CSMSG_STRUCTURE_LIMIT, "user", chanserv_conf.max_chan_users);
	return 0;
    }

    if(!(handle = chanserv_get_handle_info(user, argv[1]))) return 0;

    if((actee = GetTrueChannelAccess(channel->channel_info, handle)))
    {
        const char *title = accessNames[actee->access];
	chanserv_notice(user, CSMSG_USER_EXISTS, handle->handle, channel->name, preposition(title), title);
	return 0;
    }

    if(actor->access <= access)
    {
	chanserv_notice(user, CSMSG_NO_BUMP_ACCESS);
	return 0;
    }

    actee = add_channel_user(channel->channel_info, handle, access, 0, NULL);
    scan_handle_presence(channel, handle, NULL);

    chanserv_notice(user, CSMSG_ADDED_USER, accessNames[access], handle->handle, channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_addowner)
{
    return cmd_add_user(user, channel, ulOwner, argc, argv);
}
static CHANSERV_FUNC(cmd_addcoowner)
{
    return cmd_add_user(user, channel, ulCoowner, argc, argv);
}
static CHANSERV_FUNC(cmd_addmaster)
{
    return cmd_add_user(user, channel, ulMaster, argc, argv);
}
static CHANSERV_FUNC(cmd_addop)
{
    return cmd_add_user(user, channel, ulOp, argc, argv);
}
static CHANSERV_FUNC(cmd_addpeon)
{
    return cmd_add_user(user, channel, ulPeon, argc, argv);
}
static CHANSERV_FUNC(cmd_adduser)
{
    enum userLevel level;

    REQUIRE_PARAMS(3);

    level = user_level_name(argv[1]);
    if(level == ulNone)
    {
	chanserv_notice(user, CSMSG_INVALID_ACCESS, argv[1]);
	return 0;
    }

    return cmd_add_user(user, channel, level, argc - 1, argv + 1);
}

static CHANSERV_FUNC(cmd_clvl)
{
    struct handle_info *handle;
    struct userData *victim;
    struct userData *actor;
    enum userLevel access, new_access = ulNone;
    int privileged = IsHelping(user);

    REQUIRE_PARAMS(3);

    actor = GetChannelUser(channel->channel_info, user->handle_info);
    if(actor->access < ulMaster)
    {
	chanserv_notice(user, CSMSG_CANNOT_CLVL);
	return 0;
    }

    if(!(handle = chanserv_get_handle_info(user, argv[1]))) return 0;

    if(handle == user->handle_info && !privileged)
    {
	chanserv_notice(user, CSMSG_NO_SELF_CLVL);
	return 0;
    }

    if(!(victim = GetTrueChannelAccess(channel->channel_info, handle)))
    {
	chanserv_notice(user, CSMSG_NO_CHAN_USER, handle->handle, channel->name);
	return 0;
    }

    if(actor->access <= victim->access && !privileged)
    {
	chanserv_notice(user, MSG_USER_OUTRANKED, handle->handle);
	return 0;
    }

    for(access = USER_LEVEL_LOWEST; access <= USER_LEVEL_HIGHEST; access++)
    {
	if(!strcasecmp(argv[2], accessNames[access]))
	{
	    new_access = access;
	    break;
	}
    }

    if(new_access == ulNone)
    {
	chanserv_notice(user, CSMSG_INVALID_ACCESS, argv[2]);
	return 0;
    }

    if(new_access >= actor->access && !privileged)
    {
	chanserv_notice(user, CSMSG_NO_BUMP_ACCESS);
	return 0;
    }

    channel->channel_info->userCount[victim->access]--;
    userCount[victim->access]--;

    victim->access = new_access;

    channel->channel_info->userCount[victim->access]++;
    userCount[victim->access]++;

    chanserv_notice(user, CSMSG_CHANGED_ACCESS, handle->handle, preposition(accessNames[new_access]), accessNames[new_access], channel->name);

    return 1;
}

static int
cmd_del_user(struct userNode *user, struct chanNode *channel, enum userLevel access, unsigned int argc, unsigned char *argv[])
{
    struct handle_info *handle;
    struct userData *victim;
    struct userData *actor;

    REQUIRE_PARAMS(2);

    actor = GetChannelUser(channel->channel_info, user->handle_info);
    if(actor->access < ulMaster)
    {
	chanserv_notice(user, CSMSG_CANNOT_DEL);
	return 0;
    }

    if(!(handle = chanserv_get_handle_info(user, argv[1]))) return 0;

    if(!(victim = GetTrueChannelAccess(channel->channel_info, handle)))
    {
	chanserv_notice(user, CSMSG_NO_CHAN_USER, handle->handle, channel->name);
	return 0;
    }

    if((actor->access <= victim->access) && !IsHelping(user))
    {
	chanserv_notice(user, MSG_USER_OUTRANKED, victim->handle->handle);
	return 0;
    }

    if(access == ulNone)
    {
	access = victim->access;
    }
    else
    {
	if(access != victim->access)
	{
	    chanserv_notice(user, CSMSG_INCORRECT_ACCESS, handle->handle, preposition(accessNames[victim->access]), accessNames[victim->access], preposition(accessNames[access]), accessNames[access]);
	    return 0;
	}
    }

    del_channel_user(channel->channel_info, victim);
    chanserv_notice(user, CSMSG_DELETED_USER, preposition(accessNames[access]), accessNames[access], handle->handle, channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_delowner)
{
    return cmd_del_user(user, channel, ulOwner, argc, argv);
}

static CHANSERV_FUNC(cmd_delcoowner)
{
    return cmd_del_user(user, channel, ulCoowner, argc, argv);
}

static CHANSERV_FUNC(cmd_delmaster)
{
    return cmd_del_user(user, channel, ulMaster, argc, argv);
}

static CHANSERV_FUNC(cmd_delop)
{
    return cmd_del_user(user, channel, ulOp, argc, argv);
}

static CHANSERV_FUNC(cmd_delpeon)
{
    return cmd_del_user(user, channel, ulPeon, argc, argv);
}

static CHANSERV_FUNC(cmd_deluser)
{
    return cmd_del_user(user, channel, ulNone, argc, argv);
}

static int
cmd_mdel_user(struct userNode *user, struct chanNode *channel, enum userLevel access, unsigned int argc, unsigned char *argv[])
{
    struct userData *actor, *uData, *next;

    REQUIRE_PARAMS(2);

    actor = GetChannelUser(channel->channel_info, user->handle_info);
    if(actor->access < ulMaster)
    {
	chanserv_notice(user, CSMSG_CANNOT_DEL);
	return 0;
    }

    if((actor->access <= access) && !IsHelping(user))
    {
	chanserv_notice(user, CSMSG_NO_ACCESS);
	return 0;
    }

    for(uData = channel->channel_info->users; uData; uData = next)
    {
	next = uData->next;

	if(uData->access == access && match_ircglob(uData->handle->handle, argv[1]))
	{
	    del_channel_user(channel->channel_info, uData);
	}
    }

    chanserv_notice(user, CSMSG_DELETED_USERS, accessNames[access], argv[1], channel->name);

    return 1;
}

static CHANSERV_FUNC(cmd_mdelowner)
{
    return cmd_mdel_user(user, channel, ulOwner, argc, argv);
}

static CHANSERV_FUNC(cmd_mdelcoowner)
{
    return cmd_mdel_user(user, channel, ulCoowner, argc, argv);
}

static CHANSERV_FUNC(cmd_mdelmaster)
{
    return cmd_mdel_user(user, channel, ulMaster, argc, argv);
}

static CHANSERV_FUNC(cmd_mdelop)
{
    return cmd_mdel_user(user, channel, ulOp, argc, argv);
}

static CHANSERV_FUNC(cmd_mdelpeon)
{
    return cmd_mdel_user(user, channel, ulPeon, argc, argv);
}

static int
cmd_trim_bans(struct userNode *user, struct chanNode *channel, unsigned long duration)
{
    struct banData *bData, *next;
    char interval[INTERVALLEN];
    unsigned int count;
    time_t limit;

    count = 0;
    limit = now - duration;
    for(bData = channel->channel_info->bans; bData; bData = next)
    {
	next = bData->next;

	if(bData->triggered < limit)
	{
	    del_channel_ban(bData);
	    count++;
	}
    }

    intervalString(interval, duration);
    chanserv_notice(user, CSMSG_TRIMMED_BANS, count, interval, channel->name);
    return 1;
}

static int
cmd_trim_users(struct userNode *user, struct chanNode *channel, enum userLevel access, unsigned long duration)
{
    struct userData *actor, *uData, *next;
    char interval[INTERVALLEN];
    unsigned int count;
    time_t limit;

    actor = GetChannelUser(channel->channel_info, user->handle_info);
    if((actor->access <= access) && !IsHelping(user))
    {
	chanserv_notice(user, CSMSG_NO_ACCESS);
	return 0;
    }

    count = 0;
    limit = now - duration;
    for(uData = channel->channel_info->users; uData; uData = next)
    {
	next = uData->next;

	if(uData->seen > limit)
	{
	    continue;
	}

	if(uData->access == access || (access == ulNone && (uData->access < actor->access)))
	{
	    del_channel_user(channel->channel_info, uData);
	    count++;
	}
    }

    intervalString(interval, duration);
    chanserv_notice(user, CSMSG_TRIMMED_USERS, count, (access == ulNone) ? "user" : accessNames[access], interval, channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_trim)
{
    unsigned long duration;
    enum userLevel level;

    REQUIRE_PARAMS(3);

    duration = ParseInterval(argv[2]);
    if(!duration)
    {
	chanserv_notice(user, CSMSG_CANNOT_TRIM);
	return 0;
    }

    if(!strcmp(argv[1], "bans"))
    {
	cmd_trim_bans(user, channel, duration);
	return 1;
    }

    if(!strcmp(argv[1], "users"))
    {
	cmd_trim_users(user, channel, ulNone, duration);
	return 1;
    }

    level = user_level_name(argv[1]);
    if(level != ulNone)
    {
	cmd_trim_users(user, channel, level, duration);
	return 1;
    }

    chanserv_notice(user, CSMSG_INVALID_TRIM, argv[1]);
    return 0;
}

/* If argc is 0 in cmd_up or cmd_down, no notices will be sent
   to the user. cmd_all takes advantage of this. */
static CHANSERV_FUNC(cmd_up)
{
    struct userData *uData;
    struct modeNode *mn;
    (void)argv; /* weeeeeeeeeeeeeeeeeeeeeeee */

    uData = GetChannelUser(channel->channel_info, user->handle_info);

    mn = GetUserMode(channel, user);
    if(!mn && argc)
    {
	chanserv_notice(user, MSG_CHANNEL_ABSENT, channel->name);
	return 0;
    }

    if(uData->access >= ulOp)
    {
	if((mn->modes & MODE_CHANOP) && argc)
	{
	    chanserv_notice(user, CSMSG_ALREADY_UP, "opped", channel->name);
	    return 0;
	}

	AddChannelOp(1, &user, channel, chanserv, 1);
    }
    else
    {
	if((mn->modes & MODE_VOICE) && argc)
	{
	    chanserv_notice(user, CSMSG_ALREADY_UP, "voiced", channel->name);
	    return 0;
	}

	AddChannelVoice(1, &user, channel, chanserv, 1);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_down)
{
    char change[6 + (2 * COMBO_NUMERIC_LEN)];
    struct modeNode *mn;
    (void)argv;(void)argc;

    mn = GetUserMode(channel, user);
    if(!mn && argc)
    {
	chanserv_notice(user, MSG_CHANNEL_ABSENT, channel->name);
	return 0;
    }

    if(!(mn->modes & (MODE_CHANOP | MODE_VOICE)) && argc)
    {
	chanserv_notice(user, CSMSG_ALREADY_DOWN, channel->name);
	return 0;
    }

    DelChannelOp(1, &user, channel, chanserv, 0);
    DelChannelVoice(1, &user, channel, chanserv, 0);

    sprintf(change, "-ov %s %s", user->numeric, user->numeric);
    irc_mode(chanserv, channel, change);

    return 1;
}

static int cmd_all(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[], chanserv_func_t cmd)
{
    struct chanList *cList;
    (void)channel; (void)argc; (void)argv;

    for(cList = user->handle_info->channels; cList; cList = cList->next)
    {
	if(!cList->channel->channel || IsSuspended(cList->channel) || !GetUserMode(cList->channel->channel, user))
	{
	    continue;
	}

	cmd(user, cList->channel->channel, 0, NULL);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_upall)
{
    return cmd_all(CSFUNC_ARGS, cmd_up);
}

static CHANSERV_FUNC(cmd_downall)
{
    return cmd_all(CSFUNC_ARGS, cmd_down);
}

typedef int validate_func_t(struct userNode *user, struct chanNode *channel, struct userNode *victim);
typedef void process_func_t(unsigned int num, struct userNode **newops, struct chanNode *channel, struct userNode *who, int announce);

static int
modify_users(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[], validate_func_t validate, process_func_t process, char *action)
{
    unsigned int i = 0, valid = 0, invalid = 0;
    struct userNode *victim, **victims;

    REQUIRE_PARAMS(2);

    victims = alloca(sizeof(struct userNode *) * (argc - 1));

    while(++i < argc)
    {
	victim = GetUserH(argv[i]);

	if(!victim
           || (validate && !validate(user, channel, victim))
           || !GetUserMode(channel, victim))
	{
	    invalid++;
	    continue;
	}

	victims[valid++] = victim;
    }

    if(invalid)
    {
	chanserv_notice(user, CSMSG_PROCESS_FAILED);

	if(invalid == (argc - 1))
	{
	    return 0;
	}
    }

    process(valid, victims, channel, chanserv, 1);
    if(valid && action)
    {
	chanserv_notice(user, CSMSG_PROCESS_USERS, action, channel->name);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_op)
{
    return modify_users(CSFUNC_ARGS, validate_op, AddChannelOp, "Opped");
}

static CHANSERV_FUNC(cmd_deop)
{
    return modify_users(CSFUNC_ARGS, validate_deop, DelChannelOp, "Deopped");
}

static CHANSERV_FUNC(cmd_voice)
{
    return modify_users(CSFUNC_ARGS, NULL, AddChannelVoice, "Voiced");
}

static CHANSERV_FUNC(cmd_devoice)
{
    return modify_users(CSFUNC_ARGS, NULL, DelChannelVoice, "Devoiced");
}

static int
bad_channel_ban(struct chanNode *channel, struct userNode *user, const char *ban, int *victimCount, struct userNode **victims)
{
    
    unsigned int ii;

    if(victimCount) *victimCount = 0;
    for(ii=0; ii<channel->members.used; ii++)
    {
        struct modeNode *mn = channel->members.list[ii];

        if(IsService(mn->user))
        {
            continue;
        }

        if(!user_matches_glob(mn->user, ban, 1))
        {
            continue;
        }

        if(protect_user(mn->user, user, channel->channel_info))
        {
            return 1;
        }

        if(victims)
        {
            victims[(*victimCount)++] = mn->user;
        }
    }
    return 0;
}

static int
eject_user(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[], int action)
{
    struct userNode *victim, **victims;
    unsigned int offset, n, victimCount = 0, duration = 0, present;
    char *reason = CSMSG_KICK_REASON, *ban;

    offset = (action & ACTION_ADD_TIMED_BAN) ? 3 : 2;
    REQUIRE_PARAMS(offset);
    if(argc > offset)
    {
	unsplit_string(argv + offset, argc - offset);
	reason = argv[offset];
    }

    if((victim = GetUserH(argv[1])))
    {
        /* XXX: The comparison with ACTION_KICK is just because all
         * other actions can work on users outside the channel, and we
         * want to allow those (e.g.  unbans) in that case.  If we add
         * some other ejection action for in-channel users, change
         * this too. */
	present = GetUserMode(channel, victim) ? 1 : 0;

        if((action == ACTION_KICK) && !present)
        {
            chanserv_notice(user, MSG_CHANNEL_USER_ABSENT, victim->nick, channel->name);
            return 0;
        }

	if(IsService(victim))
	{
	    chanserv_notice(user, MSG_SERVICE_IMMUNE, victim->nick);
	    return 0;
	}

	if(protect_user(victim, user, channel->channel_info))
	{
	    chanserv_notice(user, CSMSG_USER_PROTECTED, argv[1]);
	    return 0;
	}

	victims = &victim;
	victimCount = 1;

	ban = generate_hostmask(victim, GENMASK_STRICT_HOST|GENMASK_ANY_IDENT);
    }
    else
    {
	if(!is_ircmask(argv[1]))
	{
	    chanserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	    return 0;
	}

	victims = alloca(sizeof(struct userNode *) * channel->members.used);

        if(bad_channel_ban(channel, user, argv[1], &victimCount, victims))
        {
	    chanserv_notice(user, CSMSG_MASK_PROTECTED, argv[1]);
	    return 0;
	}

        if((victimCount > 4) && ((victimCount * 3) > channel->members.used) && !IsOper(user))
        {
            chanserv_notice(user, CSMSG_LAME_MASK, argv[1]);
            return 0;
        }

        if((action == ACTION_KICK) && (victimCount == 0))
        {
            chanserv_notice(user, CSMSG_NO_MATCHING_USERS, channel->name, argv[1]);
            return 0;
        }

	ban = strdup(argv[1]);
	present = victimCount ? 1 : 0;
    }

    /* Truncate the ban in place if necessary; we must ensure
       that 'ban' is a valid ban mask before sanitizing it. */
    sanitize_ircmask(ban);

    if(action & ACTION_ADD_BAN)
    {
	struct banData *bData, *next;

	if(channel->channel_info->banCount >= chanserv_conf.max_chan_bans)
	{
	    chanserv_notice(user, CSMSG_STRUCTURE_LIMIT, "ban", chanserv_conf.max_chan_bans);
	    free(ban);
	    return 0;
	}

	if(action & ACTION_ADD_TIMED_BAN)
	{
	    duration = ParseInterval(argv[2]);

	    if(duration < 15)
	    {
		chanserv_notice(user, CSMSG_INVALID_DURATION, "at least", "15 seconds");

		free(ban);
		return 0;
	    }
	    else if(duration > (86400 * 365 * 2))
	    {
		chanserv_notice(user, CSMSG_INVALID_DURATION, "less than", "2 years");

		free(ban);
		return 0;
	    }
	}
	else
	{
	    duration = 0;
	}

	for(bData = channel->channel_info->bans; bData; bData = next)
	{
	    if(match_ircglobs(bData->mask, ban))
	    {
		/* The ban is redundant; there is already a ban
		   with the same effect in place. */
		if(bData->expires && !strcasecmp(bData->mask, ban))
		{
		    int reset = 0;

		    /* If the ban matches an existing one exactly,
		       extend the expiration time if the provided
		       duration is longer. */
		    if(duration && ((time_t)(now + duration) > bData->expires))
		    {
			bData->expires = now + duration;
			reset = 1;
		    }
		    else if(!duration)
		    {
			bData->expires = 0;
			reset = 1;
		    }

		    if(reset)
		    {
			char length[INTERVALLEN];

			/* Delete the expiration timeq entry and
			   requeue if necessary. */
			timeq_del(0, expire_ban, bData, TIMEQ_IGNORE_WHEN);

			if(bData->expires)
			{
			    timeq_add(bData->expires, expire_ban, bData);
			}

			if(duration)
			{
			    intervalString(length, duration);
			    chanserv_notice(user, CSMSG_BAN_EXTENDED, ban, length);
			}
			else
			{
			    chanserv_notice(user, CSMSG_BAN_ADDED, ban);
			}

			free(ban);
			return 0;
		    }
		}
		chanserv_notice(user, CSMSG_REDUNDANT_BAN, ban);

		free(ban);
		return 0;
	    }

	    next = bData->next;
	    if(match_ircglobs(ban, bData->mask))
	    {
		/* The ban we are adding makes previously existing
		   bans redundant; silently remove them. */
		del_channel_ban(bData);
	    }
	}

	add_channel_ban(channel->channel_info, ban, user->nick, now, (present ? now : 0), (duration ? now + duration : 0), reason);
    }

    if(action & ACTION_BAN)
    {
	if(channel->banlist.used >= MAXBANS)
	{
	    chanserv_notice(user, CSMSG_BANLIST_FULL, channel->name);

	    free(ban);
	    return 0;
	}

	if(action == ACTION_BAN && ChannelBanExists(channel, ban))
	{
	    chanserv_notice(user, CSMSG_REDUNDANT_BAN, ban);

	    free(ban);
	    return 0;
	}

	/* De-op/voice the users if we know they're in the channel. */
	if(present)
	{
	    DelChannelOp(victimCount, victims, channel, chanserv, 1);
	    DelChannelVoice(victimCount, victims, channel, chanserv, 1);
	}
	AddChannelBan(1, &ban, channel, chanserv, now, 1);
    }

    if(action & ACTION_KICK)
    {
	char *nick = user->nick, *kick_reason = alloca(strlen(reason) + strlen(nick) + CSMSG_KICK_EXTRA);
	sprintf(kick_reason, CSMSG_KICK_FORMAT, CSMSG_KICK_ARGUMENTS);

	for (n = 0; n < victimCount; n++)
	{
	    KickChannelUser(victims[n], channel, chanserv, kick_reason);
	}
    }

    if(action & ACTION_ADD_BAN)
    {
	if(duration)
	{
	    char interval[INTERVALLEN];

	    intervalString(interval, duration);
	    chanserv_notice(user, CSMSG_TIMED_BAN_ADDED, ban, interval);
	}
	else
	{
	    chanserv_notice(user, CSMSG_BAN_ADDED, ban);
	}
    }
    else if(action & ACTION_BAN)
    {
	chanserv_notice(user, CSMSG_BAN_DONE, ban);
    }
    else if(action & ACTION_KICK && victimCount)
    {
	chanserv_notice(user, CSMSG_KICK_DONE, channel->name);
    }

    free(ban);
    return 1;
}

static CHANSERV_FUNC(cmd_kickban)
{
    return eject_user(CSFUNC_ARGS, ACTION_KICK | ACTION_BAN);
}

static CHANSERV_FUNC(cmd_kick)
{
    return eject_user(CSFUNC_ARGS, ACTION_KICK);
}

static CHANSERV_FUNC(cmd_ban)
{
    return eject_user(CSFUNC_ARGS, ACTION_BAN);
}

static CHANSERV_FUNC(cmd_addban)
{
    return eject_user(CSFUNC_ARGS, ACTION_KICK | ACTION_BAN | ACTION_ADD_BAN);
}

static CHANSERV_FUNC(cmd_addtimedban)
{
    return eject_user(CSFUNC_ARGS, ACTION_KICK | ACTION_BAN | ACTION_ADD_BAN | ACTION_ADD_TIMED_BAN);
}

static int
unban_user(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[], int action)
{
    struct userNode *actee;
    char *mask = NULL;
    int acted = 0;

    REQUIRE_PARAMS(2);

    /* may want to allow a comma delimited list of users... */
    if(!(actee = GetUserH(argv[1])))
    {
	if(!is_ircmask(argv[1]))
	{
	    chanserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	    return 0;
	}

	mask = strdup(argv[1]);
    }

    /* We don't sanitize the mask here because ircu
       doesn't do it. */
    if(action & ACTION_UNBAN)
    {
	unsigned int i;
	struct banNode **ary = channel->banlist.list;
	struct string_list *removed;
	char *ban;

	removed = alloc_string_list(2);
	for (i=0; i<channel->banlist.used; i++)
	{
	    if(actee)
	    {
		if(user_matches_glob(actee, ary[i]->ban, 1))
		{
		    ban = strdup(ary[i]->ban);
		    string_list_append(removed, ban);
		}
	    }
	    else
	    {
		if(match_ircglobs(mask, ary[i]->ban))
		{
		    ban = strdup(ary[i]->ban);
		    string_list_append(removed, ban);
		}
	    }
	}

	if(removed->used)
	{
	    DelChannelBan(removed->used, removed->list, channel, chanserv, 1);
	    acted = 1;
	}

	free_string_list(removed);
    }

    if(action & ACTION_DEL_BAN)
    {
	struct banData *ban, *next;

	ban = channel->channel_info->bans;
	while(ban)
	{
	    if(actee)
	    {
		for( ; ban && !user_matches_glob(actee, ban->mask, 1);
		     ban = ban->next);
	    }
	    else
	    {
		for( ; ban && !match_ircglobs(mask, ban->mask);
		     ban = ban->next);
	    }

	    if(!ban) break;

	    next = ban->next;
	    del_channel_ban(ban);
	    ban = next;
	    acted = 1;
	}
    }

    if(!acted)
    {
	chanserv_notice(user, CSMSG_BAN_NOT_FOUND, actee ? actee->nick : mask);
    }
    else
    {
	chanserv_notice(user, CSMSG_BAN_REMOVED, actee ? actee->nick : mask);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_unban)
{
    return unban_user(CSFUNC_ARGS, ACTION_UNBAN);
}

static CHANSERV_FUNC(cmd_delban)
{
    /* it doesn't necessarily have to remove the channel ban - may want
       to make that an option. */
    return unban_user(CSFUNC_ARGS, ACTION_UNBAN | ACTION_DEL_BAN);
}

static CHANSERV_FUNC(cmd_unbanme)
{
    struct userData *uData = GetChannelUser(channel->channel_info, user->handle_info);
    long flags = ACTION_UNBAN;

    (void)argc;
    /* remove permanent bans if the user has the proper access. */
    if(uData->access >= ulMaster)
    {
	flags |= ACTION_DEL_BAN;
    }

    argv[1] = user->nick;
    return unban_user(user, channel, 2, argv, flags);
}

static CHANSERV_FUNC(cmd_unbanall)
{
    unsigned int i;
    struct banNode **ary = channel->banlist.list;
    struct string_list *removed;
    char *ban;
    (void)argc; (void)argv;

    if(!channel->banlist.used)
    {
	chanserv_notice(user, CSMSG_NO_BANS);
	return 0;
    }

    removed = alloc_string_list(channel->banlist.used);
    for (i=0; i<channel->banlist.used; i++)
    {
	ban = strdup(ary[i]->ban);
	string_list_append(removed, ban);
    }
    DelChannelBan(removed->used, removed->list, channel, chanserv, 1);
    free_string_list(removed);

    chanserv_notice(user, CSMSG_BANS_REMOVED);
    return 1;
}

static CHANSERV_FUNC(cmd_open)
{
    char *change;

    cmd_unbanme(CSFUNC_ARGS);

    if(channel->modes & MODE_KEY)
    {
	change = alloca(strlen(channel->key) + 6);
	sprintf(change, "-ilk %s", channel->key);
    }
    else
    {
	change = "-il";
    }

    mod_chanmode(channel, change, NULL, 0);
    irc_mode(chanserv, channel, change);

    chanserv_notice(user, CSMSG_CHANNEL_OPENED, channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_togop)
{
    struct userData *uData;

    (void)argc; (void)argv;
    uData = GetChannelAccess(channel->channel_info, user->handle_info);
    if(!uData)
    {
        chanserv_notice(user, CSMSG_NOT_USER, channel->name);
        return 0;
    }
    uData->flags ^= USER_AUTO_OP;
    return 1;
}

static CHANSERV_FUNC(cmd_access)
{
    struct userNode *target;
    struct userData *uData;

    if(argc < 2)
    {
	target = user;
    }
    else if(!(target = GetUserH(argv[1])))
    {
	chanserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	return 0;
    }

    if(target == chanserv)
    {
	chanserv_notice(user, CSMSG_IS_CHANSERV);
	return 1;
    }

    if(!target->handle_info)
    {
        if(IsOper(target))
        {
            chanserv_notice(user, CSMSG_LAZY_SMURF_TARGET, target->nick);
            return 0;
        }
	if(argc > 1)
	{
	    chanserv_notice(user, MSG_USER_AUTHENTICATE, argv[1]);
	    return 0;
	}
	else
	{
	    chanserv_notice(user, MSG_AUTHENTICATE);
	    return 0;
	}
    }

    if(!(uData = _GetChannelUser(channel->channel_info, target->handle_info, 1, 1)))
    {
	chanserv_notice(user, CSMSG_NO_CHAN_USER, target->nick, channel->name);
	return 0;
    }

    if(uData->access == ulHelper)
    {
	chanserv_notice(user, CSMSG_USER_IS_HELPING, target->nick, target->handle_info->handle);
	uData = GetTrueChannelAccess(channel->channel_info, target->handle_info);
    }

    if(uData)
    {
	chanserv_notice(user, CSMSG_USER_HAS_ACCESS, target->nick, target->handle_info->handle, preposition(accessNames[uData->access]), accessNames[uData->access], channel->name);
    }
    else
    {
	chanserv_notice(user, CSMSG_NO_CHAN_USER, target->nick, channel->name);
    }

    if(IsOper(target))
    {
        const char *epithet;
        if(target->handle_info->epithet)
        {
            epithet = target->handle_info->epithet;
        }
        else
        {
            epithet = "a megalomaniacal power hungry tyrant";
        }
	chanserv_notice(user, CSMSG_SMURF_TARGET, target->nick, epithet, target->handle_info->handle);
    }
    else if(IsNetworkHelper(target))
    {
        const char *epithet;
        if(target->handle_info->epithet)
        {
            epithet = target->handle_info->epithet;
        }
        else
        {
            epithet = "a wannabe tyrant";
        }
	chanserv_notice(user, CSMSG_TEARDROP_TARGET, target->nick, epithet, target->handle_info->handle);

    }
    else if(IsSupportHelper(target))
    {
        const char *epithet;
        if(target->handle_info->epithet)
        {
            epithet = target->handle_info->epithet;
        }
        else
        {
            epithet = "a wannabe tyrant";
        }
	chanserv_notice(user, CSMSG_NUKE_TARGET, target->nick, epithet, target->handle_info->handle);
    }

    if(uData)
    {
        if((GetUserMode(channel, user) || target->handle_info == user->handle_info) && uData->info)
        {
            chanserv_notice(user, "[%s] %s", target->nick, uData->info);
        }
        if((uData->flags & USER_SUSPENDED)
           && (GetChannelUser(channel->channel_info, user->handle_info)
               || (user->handle_info == uData->handle)))
        {
            chanserv_notice(user, CSMSG_USER_SUSPENDED, target->nick, channel->name);
        }
    }

    return 1;
}

static void
zoot_list(struct userNode *user, struct chanNode *channel, struct listData *list)
{
    struct userData **users, *found;
    enum userLevel display;
    char buf[INTERVALLEN], *seen;
    unsigned int i;

    users = list->users;
    for(display = list->highest; display >= list->lowest; display--)
    {
	found = NULL;

	for(i = 0; i < list->matches; i++)
	{
	    if(users[i]->access == display)
	    {
		found = users[i];
		break;
	    }
	}

	if(!found)
	{
	    if(list->highest == list->lowest)
	    {
		chanserv_notice(user, CSMSG_ACCESS_HEADER, channel->name, accessNames[display]);
		chanserv_notice(user, "$bNone.$b");
		return;
	    }
	    continue;
	}

	chanserv_notice(user, CSMSG_ACCESS_HEADER, channel->name, accessNames[display]);
	chanserv_notice(user, CSMSG_USER_HEADER, 0, "", list->handleLength+2, "Handle:", "Last Seen:");

	while(i < list->matches)
	{
	    if(users[i]->access != display)
	    {
		i++;
		continue;
	    }

	    if(!users[i]->present)
	    {
		if(users[i]->seen)
		{
		    intervalString(buf, now - users[i]->seen);
		    seen = buf;
		}
		else
		{
		    seen = "Never.";
		}
	    }
	    else
	    {
		seen = "Here.";
	    }

	    chanserv_notice(user, CSMSG_USER_DATA, 2, "", list->handleLength, users[i]->handle->handle, seen);

	    i++;
	}
    }
}

static void
def_list(struct userNode *user, struct chanNode *channel, struct listData *list)
{
    struct userData **users, *found;
    enum userLevel display;
    char buf[INTERVALLEN], *seen;
    unsigned int i;

    chanserv_notice(user, CSMSG_ACCESS_HEADER, channel->name, list->highest == list->lowest ? accessNames[list->lowest] : "user");
    chanserv_notice(user, CSMSG_USER_HEADER, ACCESSLEN, "Access:", list->handleLength, "Handle:", "Last Seen:");

    users = list->users;
    for(display = list->highest; display >= list->lowest; display--)
    {
	found = NULL;

	for(i = 0; i < list->matches; i++)
	{
	    if(users[i]->access == display)
	    {
		found = users[i];
		break;
	    }
	}

	if(!found)
	{
	    if(list->highest == list->lowest)
	    {
		chanserv_notice(user, "$bNone.$b");
		return;
	    }
	    continue;
	}

	while(i < list->matches)
	{
	    if(users[i]->access != display)
	    {
		i++;
		continue;
	    }

	    if(!users[i]->present)
	    {
		if(users[i]->seen)
		{
		    intervalString(buf, now - users[i]->seen);
		    seen = buf;
		}
		else
		{
		    seen = "Never.";
		}
	    }
	    else
	    {
		seen = "Here.";
	    }

	    chanserv_notice(user, CSMSG_USER_DATA, ACCESSLEN, accessTitles[users[i]->access], list->handleLength, users[i]->handle->handle, seen);

	    i++;
	}
    }
}

static int
cmd_list_users(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[], enum userLevel display_lowest, enum userLevel display_highest)
{
    void (*send_list)(struct userNode *, struct chanNode *, struct listData *);
    struct userData *uData, **users;
    struct chanData *cData;
    struct listData lData;
    enum userLevel display;
    unsigned int handleLength = 0, matches = 0;
    char *search = NULL;

    if(argc > 1)
    {
	search = argv[1];
    }

    /* Maybe I should make it harder to change the default. :> */
    send_list = zoot_list;

    if(user->handle_info)
    {
	switch(user->handle_info->userlist_style)
	{
	case HI_STYLE_DEF: send_list = def_list; break;
	}
    }

    cData = channel->channel_info;
    for(display = display_highest; display >= display_lowest; display--)
    {
	matches += cData->userCount[display];
    }

    users = alloca(matches * sizeof(struct userData *));
    matches = 0;

    for(uData = cData->users; uData; uData = uData->next)
    {
	if(uData->access < display_lowest || uData->access > display_highest)
	    continue;

	if(search && !match_ircglob(uData->handle->handle, search))
	    continue;

	users[matches++] = uData;

	if((unsigned int)strlen(uData->handle->handle) > handleLength)
	    handleLength = strlen(uData->handle->handle);
    }

    lData.users = users;
    lData.matches = matches;
    lData.lowest = display_lowest;
    lData.highest = display_highest;
    lData.handleLength = handleLength < 10 ? 10 : handleLength + 3;

    send_list(user, channel, &lData);

    return 1;
}

static CHANSERV_FUNC(cmd_users)
{
    return cmd_list_users(CSFUNC_ARGS, USER_LEVEL_LOWEST, USER_LEVEL_HIGHEST);
}

static CHANSERV_FUNC(cmd_wlist)
{
    return cmd_list_users(CSFUNC_ARGS, ulOwner, ulOwner);
}

static CHANSERV_FUNC(cmd_clist)
{
    return cmd_list_users(CSFUNC_ARGS, ulCoowner, ulCoowner);
}
static CHANSERV_FUNC(cmd_mlist)
{
    return cmd_list_users(CSFUNC_ARGS, ulMaster, ulMaster);
}
static CHANSERV_FUNC(cmd_olist)
{
    return cmd_list_users(CSFUNC_ARGS, ulOp, ulOp);
}
static CHANSERV_FUNC(cmd_plist)
{
    return cmd_list_users(CSFUNC_ARGS, ulPeon, ulPeon);
}

static CHANSERV_FUNC(cmd_bans)
{
    unsigned int banLength = 0, ownerLength = 0, durationLength = 0;
    unsigned int matches = 0, timed = 0, i;
    char t_buffer[INTERVALLEN], *triggered = NULL;
    char e_buffer[INTERVALLEN], *expires = NULL, *search = NULL;
    struct banData *ban, **bans;

    if(argc > 1)
    {
	search = argv[1];
    }

    if(!channel->channel_info->bans)
    {
	chanserv_notice(user, CSMSG_BAN_HEADER, 6, "Mask:", 8, "Set By:", 11, "Triggered:", 9, "Expires:", "Reason:");
	chanserv_notice(user, "$bNone.$b");
	return 0;
    }

    bans = alloca(channel->channel_info->banCount * sizeof(struct banData *));

    for(ban = channel->channel_info->bans; ban; ban = ban->next)
    {
	if(search && !match_ircglobs(search, ban->mask))
	    continue;

	bans[matches++] = ban;

	if((unsigned int)strlen(ban->mask) > banLength) banLength = strlen(ban->mask);
	if((unsigned int)strlen(ban->owner) > ownerLength) ownerLength = strlen(ban->owner);

	if(ban->expires) timed = 1;
    }

    if(!matches)
    {
	chanserv_notice(user, CSMSG_BAN_HEADER, 6, "Mask:", 8, "Set By:", 11, "Triggered:", 9, "Expires:", "Reason:");
	chanserv_notice(user, "$bNone.$b");
	return 0;
    }

    banLength = banLength < 6 ? 6 : banLength + 1;
    ownerLength = ownerLength < 8 ? 8 : ownerLength + 1;
    durationLength = timed ? INTERVALLEN : 0;

    chanserv_notice(user, CSMSG_BAN_HEADER, banLength, "Mask:", ownerLength, "Set By:", INTERVALLEN, "Triggered:", durationLength, timed ? "Expires:" : "", "Reason:");
    for(i = 0; i < matches; i++)
    {
	ban = bans[i];

	if(!timed)
	{
	    expires = "";
	}
	else if(ban->expires)
	{
	    intervalString(e_buffer, ban->expires - now);
	    expires = e_buffer;
	}
	else
	{
	    expires = "Never.";
	}

	if(ban->triggered)
	{
	    intervalString(t_buffer, now - ban->triggered);
	    triggered = t_buffer;
	}
	else
	{
	    triggered = "Never.";
	}

	chanserv_notice(user, CSMSG_BAN_DATA, banLength, ban->mask, ownerLength, ban->owner, INTERVALLEN, triggered, durationLength, expires, ban->reason);
    }

    return 1;
}

static int
bad_topic(struct chanNode *channel, struct userNode *user, const char *new_topic)
{
    struct chanData *cData = channel->channel_info;
    if (check_user_level(user, channel, cData->options[optEnfTopic], 1, 0)) return 0;
    if (cData->topic_mask)
    {
        return !match_ircglob(new_topic, cData->topic_mask);
    }
    else if (cData->topic[0])
    {
        return strcasecmp(new_topic, cData->topic);
    }
    else return 0;
}

static CHANSERV_FUNC(cmd_topic)
{
    if(argc < 2)
    {
	if(channel->channel_info->topic[0])
	{
	    SetChannelTopic(channel, chanserv, channel->channel_info->topic, 1);
	    chanserv_notice(user, CSMSG_TOPIC_SET, channel->channel_info->topic);
	}
	else
	{
	    chanserv_notice(user, CSMSG_NO_TOPIC, channel->name);
            return 0;
	}
    } else {
        unsplit_string(argv + 1, argc - 1);
        if(bad_topic(channel, user, argv[1]))
        {
            chanserv_notice(user, CSMSG_TOPIC_LOCKED, channel->name);
            return 0;
        }

        SetChannelTopic(channel, chanserv, argv[1], 1);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_mode)
{
    char modes[MODELEN];
    (void)argc; (void)argv;

    if(argc < 2)
    {
	if(channel->channel_info->mode_on || channel->channel_info->mode_off)
	{
	    create_mode_lock(modes, channel->channel_info);

	    mod_chanmode(channel, modes, NULL, 0);
	    irc_mode(chanserv, channel, modes);
	}
	else
	{
	    chanserv_notice(user, CSMSG_NO_MODES, channel->name);
	}

	return 1;
    }

    unsplit_string(argv + 1, argc - 1);
    if(!verify_chanmode(argv[1]))
    {
	chanserv_notice(user, CSMSG_MODES_INVALID, argv[1]);
	return 0;
    }

    mod_chanmode(channel, argv[1], NULL, 0);

    if(!check_user_level(user, channel, channel->channel_info->options[optEnfModes], 1, 0))
    {
	chan_mode_t on = channel->channel_info->mode_on;

	if((on && ((channel->modes & on) != on)) || (channel->modes & channel->channel_info->mode_off))
	{
	    create_mode_lock(modes, channel->channel_info);

	    mod_chanmode(channel, modes, NULL, 0);
	    irc_mode(chanserv, channel, modes);

	    chanserv_notice(user, CSMSG_MODE_LOCKED, modes, channel->name);
	    return 0;
	}
    }

    irc_mode(chanserv, channel, argv[1]);
    chanserv_notice(user, CSMSG_MODES_SET, argv[1]);
    return 1;
}

static CHANSERV_FUNC(cmd_invite)
{
    struct userData *uData;
    struct userNode *invite;
    (void)argc; (void)argv;

    if(argc > 1)
    {
        if(!(invite = GetUserH(argv[1])))
	{
            chanserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
            return 0;
        }
    }
    else
    {
        invite = user;
    }

    if(GetUserMode(channel, invite))
    {
	chanserv_notice(user, CSMSG_ALREADY_PRESENT, invite->nick, channel->name);
	return 0;
    }

    uData = GetChannelUser(channel->channel_info, user->handle_info);
    if(invite != user)
    {
	if(uData->access < ulMaster)
	{
	    chanserv_notice(user, CSMSG_CANNOT_INVITE);
	    return 0;
	}

	chanserv_notice(invite, CSMSG_INVITING_YOU, user->nick, channel->name);
    }

    irc_invite(chanserv, invite, channel);
    if(argc > 1)
    {
	chanserv_notice(user, CSMSG_INVITED_USER, argv[1], channel->name);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_version)
{
    (void)channel;

    chanserv_notice(user, MSG_VERSION);

    if(argc > 1 && !strcasecmp(argv[1], "cvs"))
    {
	chanserv_notice(user, CSMSG_VERSION_ID);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_info)
{
    char modes[MODELEN], buffer[INTERVALLEN];
    struct userData *uData;
    struct chanData *cData;
    struct suspended *sChan = NULL;
    struct do_not_register *dnr;
    enum userLevel access;
    int i;
    (void)argc;(void)argv;

    if(channel)
    {
        cData = channel->channel_info;
    }
    else
    {
        sChan = dict_find(sChannels, argv[1], NULL);
        if(!sChan)
        {
            chanserv_notice(user, CSMSG_UNREGISTERED, argv[1]);
            return 0;
        }
        cData = sChan->cData;
    }

    if(!cData)
    {
        chanserv_notice(user, CSMSG_UNREGISTERED, channel ? (unsigned char *)channel->name : argv[1]);
        return 0;
    }

    chanserv_notice(user, CSMSG_CHANNEL_INFO, channel ? channel->name : sChan->name);

    uData = GetChannelUser(cData, user->handle_info);
    if(uData && uData->access >= ulOp)
    {
	chanserv_notice(user, CSMSG_CHANNEL_TOPIC, cData->topic);
        create_mode_lock(modes, cData);
	chanserv_notice(user, CSMSG_CHANNEL_MODES, modes);
    }

    chanserv_notice(user, CSMSG_CHANNEL_BANS, cData->banCount);
    chanserv_notice(user, CSMSG_CHANNEL_MAX, cData->max);

    for(access = USER_LEVEL_HIGHEST; access >= USER_LEVEL_LOWEST; access--)
    {
	i = PADLEN - strlen(accessTitles[access]);
 
	chanserv_notice(user, CSMSG_CHANNEL_USERS, accessTitles[access], i, "", cData->userCount[access]);
    }
    chanserv_notice(user, CSMSG_CHANNEL_USERS, "Total User", PADLEN - 10, "", cData->userCount[0]);

    if(cData->registrar)
    {
	chanserv_notice(user, CSMSG_CHANNEL_REGISTRAR, cData->registrar);
    }

    if(IsSuspended(cData))
    {
        struct suspended *suspended = cData->suspended;
        char *reason = suspended->reason ? suspended->reason : "No reason";
        if(suspended->expires)
        {
            intervalString(buffer, suspended->expires - now);
            chanserv_notice(user, CSMSG_CHANNEL_SUSPENDED2, cData->suspended->suspender, reason, buffer);
        }
        else
        {
            chanserv_notice(user, CSMSG_CHANNEL_SUSPENDED, cData->suspended->suspender, reason);
        }
    }

    intervalString(buffer, now - cData->registered);
    chanserv_notice(user, CSMSG_CHANNEL_REGISTERED, buffer);

    intervalString(buffer, now - cData->visited);
    chanserv_notice(user, CSMSG_CHANNEL_VISITED, buffer);

    if(IsPrivileged(user) && (dnr = chanserv_is_dnr(channel->name, NULL)))
    {
        chanserv_notice(user, CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_netinfo)
{
    extern unsigned long burst_length;
    char interval[INTERVALLEN];
    enum userLevel access;
    int i;
    (void)argc;(void)argv;(void)channel;

    chanserv_notice(user, CSMSG_NETWORK_INFO);

    chanserv_notice(user, CSMSG_NETWORK_SERVERS, server_count);
    chanserv_notice(user, CSMSG_NETWORK_USERS, dict_size(clients));
    chanserv_notice(user, CSMSG_NETWORK_OPERS, curr_opers.used);

    chanserv_notice(user, CSMSG_NETWORK_CHANNELS, registered_channels);

    chanserv_notice(user, CSMSG_NETWORK_BANS, banCount);
    for(access = USER_LEVEL_HIGHEST; access >= USER_LEVEL_LOWEST; access--)
    {
	i = PADLEN - strlen(accessTitles[access]);
 
	chanserv_notice(user, CSMSG_CHANNEL_USERS, accessTitles[access], i, "", userCount[access]);
    }
    chanserv_notice(user, CSMSG_CHANNEL_USERS, "Total User", PADLEN - 10, "", userCount[0]);

    intervalString(interval, now - boot_time);
    chanserv_notice(user, CSMSG_SERVICES_UPTIME, interval);

    intervalString(interval, burst_length);
    chanserv_notice(user, CSMSG_BURST_LENGTH, interval);

    if(lastwrite)
    {
	intervalString(interval, now - lastwrite);
	chanserv_notice(user, CSMSG_LAST_WRITE, interval);
    }
    else
    {
	chanserv_notice(user, CSMSG_NO_LAST_WRITE);
    }

    return 1;
}

static void
send_staff_list(struct userNode *to, struct userList *list, int skip_flags)
{
    struct helpfile_table table;
    unsigned int nn;
    struct userNode *user;
    unsigned char *nick;

    table.length = 0;
    table.width = 1;
    table.flags = TABLE_REPEAT_ROWS | TABLE_NO_FREE | TABLE_NO_HEADERS;
    table.contents = alloca(list->used*sizeof(*table.contents));
    for(nn=0; nn<list->used; nn++)
    {
        user = list->list[nn];
        if(user->modes & skip_flags) continue;
        table.contents[table.length] = alloca(table.width*sizeof(**table.contents));
        if(IsAway(user))
        {
            nick = alloca(strlen(user->nick)+3);
            sprintf(nick, "(%s)", user->nick);
        }
        else
        {
            nick = user->nick;
        }
        table.contents[table.length][0] = nick;
        table.length++;
    }
    table_send(chanserv, to->numeric, 0, NULL, table);
}

static CHANSERV_FUNC(cmd_ircops)
{
    (void)argc;(void)argv;(void)channel;
    chanserv_notice(user, CSMSG_STAFF_OPERS);
    send_staff_list(user, &curr_opers, FLAGS_SERVICE);
    return 1;
}

static CHANSERV_FUNC(cmd_helpers)
{
    (void)argc;(void)argv;(void)channel;
    chanserv_notice(user, CSMSG_STAFF_HELPERS);
    send_staff_list(user, &curr_helpers, FLAGS_OPER);
    return 1;
}

static CHANSERV_FUNC(cmd_staff)
{
    (void)argc;(void)argv;(void)channel;

    chanserv_notice(user, CSMSG_NETWORK_STAFF);
    cmd_ircops(CSFUNC_ARGS);
    cmd_helpers(CSFUNC_ARGS);

    return 1;
}

static CHANSERV_FUNC(cmd_peek)
{
    struct modeNode *mn;
    char modes[MODELEN];
    unsigned int n;
    struct helpfile_table table;

    (void)argc;(void)argv;
    irc_make_chanmode(channel, modes);

    chanserv_notice(user, CSMSG_PEEK_INFO, channel->name);
    chanserv_notice(user, CSMSG_PEEK_TOPIC, channel->topic);
    chanserv_notice(user, CSMSG_PEEK_MODES, modes);
    chanserv_notice(user, CSMSG_PEEK_OPS);

    table.length = 0;
    table.width = 1;
    table.flags = TABLE_REPEAT_ROWS | TABLE_NO_FREE | TABLE_NO_HEADERS;
    table.contents = alloca(channel->members.used*sizeof(*table.contents));
    for(n = 0; n < channel->members.used; n++)
    {
	mn = channel->members.list[n];
	if(!(mn->modes & MODE_CHANOP) || IsLocal(mn->user)) continue;
        table.contents[table.length] = alloca(sizeof(**table.contents));
        table.contents[table.length][0] = mn->user->nick;
        table.length++;
    }
    table_send(chanserv, user->numeric, 0, NULL, table);

    return 1;
}

static CHANSERV_FUNC(cmd_setinfo)
{
    struct userData *uData;

    uData = GetChannelAccess(channel->channel_info, user->handle_info);

    if(!uData)
    {
	/* If they got past the command restrictions (which require access)
	   but fail this test, we have someone with security override on. */
	chanserv_notice(user, CSMSG_NOT_USER, channel->name);
	return 0;
    }

    if(argc < 2)
    {
	if(uData->info)
	{
	    chanserv_notice(user, CSMSG_INFO_LINE, uData->info);
	}
	else
	{
	    chanserv_notice(user, CSMSG_NO_INFO_LINE);
	}

	return 1;
    }

    unsplit_string(argv + 1, argc - 1);
    if(uData->info) free(uData->info);
    if(argv[1][0] == '*' && argv[1][1] == 0)
    {
	uData->info = NULL;
	chanserv_notice(user, CSMSG_CLEARED_INFO_LINE);
    }
    else
    {
	uData->info = strdup(argv[1]);
	chanserv_notice(user, CSMSG_SET_INFO_LINE, argv[1]);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_seen)
{
    struct userData *uData;
    struct handle_info *handle;
    char seen[INTERVALLEN];

    REQUIRE_PARAMS(2);

    if(!strcasecmp(argv[1], chanserv->nick))
    {
	chanserv_notice(user, CSMSG_IS_CHANSERV);
	return 1;
    }

    if(!(handle = get_handle_info(argv[1])))
    {
	chanserv_notice(user, MSG_HANDLE_UNKNOWN, argv[1]);
	return 0;
    }

    if(!(uData = GetTrueChannelAccess(channel->channel_info, handle)))
    {
	chanserv_notice(user, CSMSG_NO_CHAN_USER, handle->handle, channel->name);
	return 0;
    }

    if(uData->present)
    {
	chanserv_notice(user, CSMSG_USER_PRESENT, handle->handle);
    }
    else
    {
	if(uData->seen)
	{
	    intervalString(seen, now - uData->seen);
	    chanserv_notice(user, CSMSG_USER_SEEN, handle->handle, channel->name, seen);
	}
	else
	{
	    chanserv_notice(user, CSMSG_NEVER_SEEN, handle->handle, channel->name);
	}

        if(HANDLE_FLAGGED(handle, FROZEN))
        {
            chanserv_notice(user, CSMSG_USER_VACATION, handle->handle);
        }
    }

    return 1;
}

static int
note_type_visible_to_user(struct chanData *channel, struct note_type *ntype, struct userNode *user)
{
    switch(ntype->visible_type)
    {
    case NOTE_VIS_ALL: return 1;
    case NOTE_VIS_CHANNEL_USERS: return !channel || (user->handle_info && GetChannelUser(channel, user->handle_info));
    case NOTE_VIS_PRIVILEGED: default: return IsPrivileged(user);
    }
}

static int
note_type_settable_by_user(struct chanNode *channel, struct note_type *ntype, struct userNode *user)
{
    struct userData *uData;

    switch(ntype->set_access_type)
    {
    case NOTE_SET_CHANNEL_ACCESS:
        if(!user->handle_info) return 0;
        uData = GetChannelUser(channel->channel_info, user->handle_info);
        return uData->access >= ntype->set_access.min_ulevel;
    case NOTE_SET_CHANNEL_SETTER:
        return check_user_level(user, channel, channel->channel_info->options[optSetters], 1, 0);
    case NOTE_SET_PRIVILEGED: default: return IsPrivileged(user);
    }
}

static CHANSERV_FUNC(cmd_note)
{
    struct chanData *cData = channel->channel_info;
    struct note *note;
    struct note_type *ntype;

    /* If no arguments, show all visible notes for the channel. */
    if(argc < 2)
    {
        dict_iterator_t it;
        unsigned int count;

        for(count=0, it=dict_first(cData->notes); it; it=iter_next(it))
        {
            note = iter_data(it);
            if(!note_type_visible_to_user(cData, note->type, user)) continue;
            if(!count++) chanserv_notice(user, CSMSG_NOTELIST_HEADER, channel->name);
            chanserv_notice(user, CSMSG_NOTE_FORMAT, iter_key(it), note->setter, note->note);
        }
        if(count) chanserv_notice(user, CSMSG_NOTELIST_END, channel->name);
        else chanserv_notice(user, CSMSG_NOTELIST_EMPTY, channel->name);
    }
    /* If one argument, show the named note. */
    else if(argc < 3)
    {
        if((note = dict_find(cData->notes, argv[1], NULL))
           && note_type_visible_to_user(cData, note->type, user))
        {
            chanserv_notice(user, CSMSG_NOTE_FORMAT, note->type->name, note->setter, note->note);
        }
        else if((ntype = dict_find(note_types, argv[1], NULL))
                && note_type_visible_to_user(NULL, ntype, user))
        {
            chanserv_notice(user, CSMSG_NO_SUCH_NOTE, channel->name, ntype->name);
            return 0;
        }
        else
        {
            chanserv_notice(user, CSMSG_BAD_NOTE_TYPE, argv[1]);
            return 0;
        }
    }
    /* Assume they're trying to set a note. */
    else
    {
        ntype = dict_find(note_types, argv[1], NULL);
        if(!ntype)
        {
            chanserv_notice(user, CSMSG_BAD_NOTE_TYPE, argv[1]);
            return 0;
        }
        else if(note_type_settable_by_user(channel, ntype, user))
        {
            unsplit_string(argv+2, argc-2);
            chanserv_add_channel_note(cData, ntype, user->handle_info->handle, argv[2]);
            chanserv_notice(user, CSMSG_NOTE_SET, ntype->name, channel->name);
        }
        else
        {
            chanserv_notice(user, CSMSG_NO_ACCESS);
            return 0;
        }
    }
    return 1;
}

static CHANSERV_FUNC(cmd_delnote)
{
    struct note *note;

    REQUIRE_PARAMS(2);
    if(!(note = dict_find(channel->channel_info->notes, argv[1], NULL))
       || !note_type_settable_by_user(channel, note->type, user))
    {
        chanserv_notice(user, CSMSG_NO_SUCH_NOTE, channel->name, argv[1]);
        return 0;
    }
    dict_remove(channel->channel_info->notes, note->type->name);
    chanserv_notice(user, CSMSG_NOTE_REMOVED, channel->name, argv[1]);
    return 1;
}

static void
event_helper_func(struct logEntry *match, void *data)
{
    char *severity;
    static char timestamp[24];
    struct tm tm;

    localtime_r(&match->time, &tm);
    strftime(timestamp, sizeof(timestamp), "[%H:%M:%S %m/%d/%Y]", &tm);

    switch(match->severity)
    {
    case LOG_INFO: severity = "(info) "; break;
    case LOG_WARNING: severity = "(warning) "; break;
    case LOG_ERROR: severity = "(error) "; break;
    case LOG_OTHER: default: severity = ""; break;
    }

    chanserv_notice(data, "%s %s%s", timestamp, severity, match->text);
}

static int
chanserv_events(struct userNode *user, unsigned int limit, char *glob, char *label)
{
    struct logSearch discrim;
    unsigned int matches;

    if(limit < 1 || limit > 200)
    {
	limit = 10;
    }

    discrim.limit = limit;
    discrim.glob = glob;

    discrim.type = CS_LOG;
    discrim.severity = 0;

    chanserv_notice(user, MSG_SEARCH_RESULTS, label);
    matches = log_entry_search(&discrim, event_helper_func, user);

    if(matches)
    {
	chanserv_notice(user, MSG_MATCH_COUNT, matches);
    }
    else
    {
	chanserv_notice(user, MSG_NO_MATCHES);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_events)
{
    unsigned int limit;
    char *glob;

    limit = (argc > 1) ? atoi(argv[1]) : 10;

    if(argc > 2)
    {
	glob = alloca(4 + strlen(channel->name) + strlen(argv[2]));
	sprintf(glob, "(%s)*%s", channel->name, argv[2]);
    }
    else
    {
        glob = alloca(4 + strlen(channel->name));
        sprintf(glob, "(%s)*", channel->name);
    }

    return chanserv_events(user, limit, glob, "channel events");
}

static CHANSERV_FUNC(cmd_last)
{
    unsigned int limit;
    char *glob;
    (void)channel;

    limit = argc > 1 ? atoi(argv[1]) : 10;
    glob = argc > 2 ? argv[2] : NULL;

    return chanserv_events(user, limit, glob, "$b$C$b events");
}

static CHANSERV_FUNC(cmd_say)
{
    (void)user;

    REQUIRE_PARAMS(2);

    unsplit_string(argv + 1, argc - 1);
    send_channel_message(channel, chanserv, "%s", argv[1]);

    return 1;
}

static CHANSERV_FUNC(cmd_emote)
{
    (void)user;

    REQUIRE_PARAMS(2);

    /* CTCP is so annoying. */
    unsplit_string(argv + 1, argc - 1);
    send_channel_message(channel, chanserv, "\001ACTION %s\001", argv[1]);

    return 1;
}

struct chanNode *
chanserv_support_channel(void)
{
    return chanserv_conf.support_channel;
}

static CHANSERV_FUNC(cmd_god)
{
    int helping = 0;
    (void)channel;

#ifdef DISTRUST_OPERS
    if(user->handle_info->opserv_level < 800)
    {
        return 0;
    }
#endif

    if(argc > 1)
    {
	if(enabled_string(argv[1]))
	{
	    helping = 1;
	}
	else if(disabled_string(argv[1]))
	{
	    helping = 0;
	}
	else
	{
	    chanserv_notice(user, MSG_INVALID_BINARY, argv[1]);
	    return 0;
	}
    }
    else
    {
	helping = !IsHelping(user);
    }

    if(HANDLE_FLAGGED(user->handle_info, HELPING))
    {
	if(helping)
	{
	    chanserv_notice(user, CSMSG_ALREADY_HELPING);
	}
	else
	{
	    HANDLE_CLEAR_FLAG(user->handle_info, HELPING);
	    chanserv_notice(user, CSMSG_NOW_NOT_HELPING);
	}
    }
    else
    {
	if(helping)
	{
	    HANDLE_SET_FLAG(user->handle_info, HELPING);
	    chanserv_notice(user, CSMSG_NOW_HELPING);
	}
	else
	{
	    chanserv_notice(user, CSMSG_ALREADY_NOT_HELPING);
	}
    }

    if(IsHelping(user))
    {
	policer_set_params(user->command_policer, god_policer_params);
    }
    else
    {
	policer_set_params(user->command_policer, IsOper(user) ? oper_policer_params : luser_policer_params);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_expire)
{
    int channel_count = registered_channels;
    (void)channel;(void)argc;(void)argv;

    expire_channels(NULL);
    chanserv_notice(user, CSMSG_CHANNELS_EXPIRED, channel_count - registered_channels);

    return 1;
}

static void
chanserv_expire_suspension(void *data)
{
    struct suspended *suspended = data;
    struct chanNode *channel;

    dict_remove(sChannels, suspended->name);
    channel = AddChannel(suspended->name, now, NULL, NULL, 0, chanserv->numeric, NULL);
    AddChannelOp(1, &chanserv, channel, chanserv, 1);
    channel->channel_info = suspended->cData;
    suspended->cData->channel = channel;
    suspended->cData->flags &= ~CHANNEL_SUSPENDED;
    suspended->cData->suspended = NULL;
    free(suspended->name);
    free(suspended->suspender);
    free(suspended->reason);
    free(suspended);
}

static CHANSERV_FUNC(cmd_csuspend)
{
    struct suspended *suspended;
    char *reason;
    time_t expiry;

    if(IsProtected(channel->channel_info))
    {
        chanserv_notice(user, CSMSG_SUSPEND_NODELETE, channel->name);
        return 0;
    }

    if(IsSuspended(channel->channel_info))
    {
	chanserv_notice(user, CSMSG_ALREADY_SUSPENDED, channel->name);
	return 0;
    }

    REQUIRE_PARAMS(3);

    if(!strcmp(argv[1], "0"))
    { 
        expiry = 0;
    }
    else
    {
        unsigned int duration;
        if(!(duration = ParseInterval(argv[1])))
        {
            chanserv_notice(user, MSG_INVALID_DURATION, argv[1]);
        }
        expiry = now + duration;
    }

    unsplit_string(argv + 2, argc - 2);

    suspended = malloc(sizeof(struct suspended));

    suspended->name = strdup(channel->name);
    suspended->suspender = strdup(user->nick);
    suspended->expires = expiry;
    if(suspended->expires)
    {
        timeq_add(suspended->expires, chanserv_expire_suspension, suspended);
    }
    reason = suspended->reason = strdup(argv[2]);

    suspended->cData = channel->channel_info;

    suspended->cData->suspended = suspended;
    suspended->cData->channel = NULL;

    /* Mark the channel as suspended, then part. */
    channel->channel_info->flags |= CHANNEL_SUSPENDED;
    dict_insert(sChannels, suspended->name, suspended);
    DelChannelUser(chanserv, channel, reason, 0);
    chanserv_notice(user, CSMSG_SUSPENDED, suspended->name);

    reason = alloca(strlen(user->nick) + strlen(channel->name) + 32);
    sprintf(reason, "%s suspended by %s.", suspended->name, user->nick);
    global_message(MESSAGE_RECIPIENT_OPERS | MESSAGE_RECIPIENT_HELPERS, reason);
    return 1;
}

static CHANSERV_FUNC(cmd_cunsuspend)
{
    struct suspended *suspended;
    char *message;

    if(!channel && argc < 2)
    {
	chanserv_notice(user, MSG_INVALID_CHANNEL);
	return 0;
    }

    if(channel)
    {
	if(!channel->channel_info || !IsSuspended(channel->channel_info) || !channel->channel_info->suspended)
	{
	    chanserv_notice(user, CSMSG_NOT_SUSPENDED, channel->name);
	    return 0;
	}
	suspended = channel->channel_info->suspended;
    }
    else
    if(!(suspended = dict_find(sChannels, argv[1], NULL)))
    {
        chanserv_notice(user, CSMSG_NOT_SUSPENDED, argv[1]);
        return 0;
    }

    timeq_del(suspended->expires, chanserv_expire_suspension, suspended, 0);
    chanserv_expire_suspension(suspended);

    chanserv_notice(user, CSMSG_UNSUSPENDED, channel->name);
    message = alloca(strlen(user->nick) + strlen(channel->name) + 32);
    sprintf(message, "%s unsuspended by %s.", channel->name, user->nick);
    global_message(MESSAGE_RECIPIENT_OPERS | MESSAGE_RECIPIENT_HELPERS, message);
    return 1;
}

static CHANSERV_FUNC(cmd_command)
{
    struct chanserv_cmd *command;
    char *cmd, *access;
    (void)channel;

    if(argc < 2 || argc == 3)
    {
	chanserv_notice(user, MSG_MISSING_PARAMS, argv[0]);
	CHANSERV_SYNTAX();
	return 0;
    }

    cmd = resolve_alias("ChanServ", argv[1]);
    command = dict_find(chanserv_func_dict, cmd, NULL);

    if(!command)
    {
	chanserv_notice(user, MSG_COMMAND_UNKNOWN, cmd);
	return 0;
    }
    access = alloca(strlen(cmd) + 9);
    sprintf(access, "%s%caccess", cmd, 0);

    if(argc < 3)
    {
	char *aliases;

	/* Display information about the command. This is available to
	   all users. */

	if(cmd != (char *)argv[1])
	{
	    chanserv_notice(user, CSMSG_COMMAND_IS_ALIAS, argv[1], cmd);
	}

	chanserv_notice(user, CSMSG_COMMAND_INFO, cmd);
	chanserv_notice(user, CSMSG_COMMAND_ACCESS, chanserv_expand_command(access));
	/* I really don't think users care too much about either the
         * weight or the flags. */
	if(IsPrivileged(user))
	{
	    chanserv_notice(user, CSMSG_COMMAND_FLAGS, command->flags);
	    chanserv_notice(user, CSMSG_COMMAND_WEIGHT, command->weight);
	}
	chanserv_notice(user, CSMSG_COMMAND_USES, command->uses);

	if((aliases = command_aliases("ChanServ", cmd)))
	{
	    char *c = aliases;

	    for(c = aliases; *c; c++)
	    {
		*c = toupper(*c);
	    }

	    chanserv_notice(user, CSMSG_COMMAND_ALIASES, aliases);
	    free(aliases);
	}

	return 1;
    }

    if(!IsOper(user))
    {
	chanserv_notice(user, MSG_COMMAND_PRIVILEGED, argv[0]);
	return 0;
    }

    if(!strcasecmp(argv[2], "flags"))
    {
	command->flags = atoi(argv[3]) | (command->flags & CMD_REQUIRE_LOCK);
	chanserv_notice(user, CSMSG_COMMAND_ACCESS, chanserv_expand_command(access));
    }
    else if(!strcasecmp(argv[2], "weight"))
    {
	command->weight = atof(argv[3]);
	chanserv_notice(user, CSMSG_COMMAND_WEIGHT, command->weight);
    }
    else if(!strcasecmp(argv[2], "access"))
    {
	enum userLevel level;

	level = user_level_name(argv[3]);

	if(level >= USER_LEVEL_LOWEST && level <= USER_LEVEL_HIGHEST)
	{
	    command->access = level;
	    chanserv_notice(user, CSMSG_COMMAND_ACCESS, chanserv_expand_command(access));
	}
	else
	{
	    chanserv_notice(user, CSMSG_INVALID_ACCESS, argv[3]);
	    return 0;
	}
    }
    else
    {
	chanserv_notice(user, CSMSG_INVALID_PROPERTY, argv[2]);
	return 0;
    }

    return 1;
}

typedef struct chanservSearch
{
    char *name;
    char *registrar;

    time_t unvisited;
    time_t registered;

    unsigned long flags;
    unsigned int limit;
} *search_t;

typedef void (*channel_search_func)(struct chanData *channel, void *data);

static search_t
chanserv_search_create(struct userNode *user, unsigned int argc, unsigned char *argv[])
{
    search_t search;
    unsigned int i;

    search = malloc(sizeof(struct chanservSearch));
    (void)memset(search, 0, sizeof(*search));
    search->limit = 25;

    for(i = 0; i < argc; i++)
    {
	/* Assume all criteria require arguments. */
	if(i == (argc - 1))
	{
	    chanserv_notice(user, MSG_MISSING_PARAMS, argv[i]);
	    return NULL;
	}

	if(!strcasecmp(argv[i], "name"))
	{
	    search->name = argv[++i];
	}
	else if(!strcasecmp(argv[i], "registrar"))
	{
	    search->registrar = argv[++i];
	}
	else if(!strcasecmp(argv[i], "unvisited"))
	{
	    search->unvisited = ParseInterval(argv[++i]);
	}
	else if(!strcasecmp(argv[i], "registered"))
	{
	    search->registered = ParseInterval(argv[++i]);
	}
	else if(!strcasecmp(argv[i], "flags"))
	{
	    i++;
	    if(!strcasecmp(argv[i], "nodelete"))
	    {
		search->flags |= CHANNEL_NODELETE;
	    }
	    else if(!strcasecmp(argv[i], "suspended"))
	    {
		search->flags |= CHANNEL_SUSPENDED;
	    }
	    else
	    {
		chanserv_notice(user, CSMSG_INVALID_CFLAG, argv[i]);
		goto fail;
	    }
	}
	else if(!strcasecmp(argv[i], "limit"))
	{
	    search->limit = strtoul(argv[++i], NULL, 10);
	}
	else
	{
	    chanserv_notice(user, MSG_INVALID_CRITERIA, argv[i]);
	    goto fail;
	}
    }

    if(search->name && !strcmp(search->name, "*"))
    {
	search->name = 0;
    }
    if(search->registrar && !strcmp(search->registrar, "*"))
    {
	search->registrar = 0;
    }

    return search;
  fail:
    free(search);
    return NULL;
}

static int
chanserv_channel_match(struct chanData *channel, search_t search)
{
    const char *name = channel->channel ? channel->channel->name : channel->suspended->name;
    if((search->name && !match_ircglob(name, search->name)) ||
       (search->registrar && !channel->registrar) ||
       (search->registrar && !match_ircglob(channel->registrar, search->registrar)) ||
       (search->unvisited && (now - channel->visited) < search->unvisited) ||
       (search->registered && (now - channel->registered) > search->registered) ||
       (search->flags && ((search->flags & channel->flags) != search->flags)))
    {
	return 0;
    }

    return 1;
}

static unsigned int
chanserv_channel_search(search_t search, channel_search_func smf, void *data)
{
    struct chanData *channel;
    unsigned int matches = 0;

    for(channel = channelList; channel && matches < search->limit; channel = channel->next)
    {
	if(!chanserv_channel_match(channel, search)) continue;
	matches++;
	smf(channel, data);
    }

    return matches;
}

static void
search_count(struct chanData *channel, void *data)
{
    (void)channel;(void)data;
}

static void
search_print(struct chanData *channel, void *data)
{
    const char *name = IsSuspended(channel) ? channel->suspended->name : channel->channel->name;
    chanserv_notice(data, CSMSG_SEARCH_DATA, name);
}

static CHANSERV_FUNC(cmd_search)
{
    search_t search;
    unsigned int matches;
    channel_search_func action;

    (void)channel;

    REQUIRE_PARAMS(3);

    if(!strcasecmp(argv[1], "count"))
    {
	action = search_count;
    }
    else if(!strcasecmp(argv[1], "print"))
    {
	action = search_print;
    }
    else
    {
	chanserv_notice(user, CSMSG_ACTION_INVALID, argv[1]);
	return 0;
    }

    search = chanserv_search_create(user, argc - 2, argv + 2);
    if(!search) return 0;

    if(action == search_count)
    {
	search->limit = INT_MAX;
    }

    if(action == search_print)
    {
	chanserv_notice(user, MSG_SEARCH_RESULTS, "channels");
    }

    matches = chanserv_channel_search(search, action, user);

    if(matches)
    {
	chanserv_notice(user, MSG_MATCH_COUNT, matches);
    }
    else
    {
	chanserv_notice(user, MSG_NO_MATCHES);
    }

    free(search);
    return 1;
}

static CHANSERV_FUNC(cmd_unvisited)
{
    struct chanData *cData;
    time_t interval = chanserv_conf.channel_expire_delay;
    char buffer[INTERVALLEN];
    unsigned int limit = 25, matches = 0;
    (void)channel;

    if(argc > 1)
    {
	interval = ParseInterval(argv[1]);
	if(argc > 2) limit = atoi(argv[2]);
    }

    intervalString(buffer, interval);
    chanserv_notice(user, CSMSG_UNVISITED_HEADER, limit, buffer);

    for(cData = channelList; cData && matches < limit; cData = cData->next)
    {
	if((now - cData->visited) < interval) continue;

	intervalString(buffer, now - cData->visited);
	chanserv_notice(user, CSMSG_UNVISITED_DATA, cData->channel->name, buffer);
	matches++;
    }

    return 1;
}

#define OPTION_FUNC(NAME) int NAME(struct userNode *user, struct chanNode *channel, int argc, unsigned char *argv[])
typedef OPTION_FUNC(option_func_t);

#define BINARY_OPTION(args...) return binary_option(args, CSFUNC_ARGS);
#define MULTIPLE_OPTION(args...) return multiple_option(args, values, ArrayLength(values), CSFUNC_ARGS);

static OPTION_FUNC(opt_defaulttopic)
{
    if(argc > 0)
    {
        if(!check_user_level(user, channel, channel->channel_info->options[optEnfTopic], 1, 0))
        {
            chanserv_notice(user, CSMSG_TOPIC_LOCKED, channel->name);
            return 0;
        }

	unsplit_string(argv, argc);

	if(argv[0][0] == '*' && argv[0][1] == 0)
	{
	    channel->channel_info->topic[0] = 0;
	}
	else
	{
	    safestrncpy(channel->channel_info->topic, argv[0], sizeof(channel->topic));
            if(channel->channel_info->topic_mask
               && !match_ircglob(channel->channel_info->topic, channel->channel_info->topic_mask))
            {
                chanserv_notice(user, CSMSG_TOPIC_MISMATCH, channel->name);
            }
	}
        SetChannelTopic(channel, chanserv, channel->channel_info->topic, 1);
    }

    chanserv_notice(user, CSMSG_STRING_VALUE, "DefaultTopic ", channel->channel_info->topic[0] ? channel->channel_info->topic : "None.");
    return 1;
}

static OPTION_FUNC(opt_topicmask)
{
    if(argc > 0)
    {
        struct chanData *cData = channel->channel_info;

        if(!check_user_level(user, channel, cData->options[optEnfTopic], 1, 0))
        {
            chanserv_notice(user, CSMSG_TOPIC_LOCKED, channel->name);
            return 0;
        }

	unsplit_string(argv, argc);

        if(cData->topic_mask) free(cData->topic_mask);
	if(argv[0][0] == '*' && argv[0][1] == 0)
	{
	    cData->topic_mask = 0;
	}
	else
	{
            cData->topic_mask = strdup(argv[0]);
            if(!cData->topic[0])
            {
                chanserv_notice(user, CSMSG_MASK_BUT_NO_TOPIC, channel->name);
            }
            else if(!match_ircglob(cData->topic, cData->topic_mask))
            {
                chanserv_notice(user, CSMSG_TOPIC_MISMATCH, channel->name);
            }
	}
    }

    chanserv_notice(user, CSMSG_STRING_VALUE, "TopicMask    ", channel->channel_info->topic_mask ? channel->channel_info->topic_mask : "None.");
    return 1;
}

static OPTION_FUNC(opt_greeting)
{
    if(argc > 0)
    {
	if(!require_access(channel, user, ulCoowner))
	{
	    return 0;
	}

	unsplit_string(argv, argc);

	if(argv[0][0] == '*' && argv[0][1] == 0)
	{
	    free(channel->channel_info->greeting);
	    channel->channel_info->greeting = NULL;
	}
	else
	{
	    unsigned int length;

	    length = strlen(argv[0]);
	    if(length > chanserv_conf.greeting_length)
	    {
		chanserv_notice(user, CSMSG_GREETING_TOO_LONG, length, chanserv_conf.greeting_length);
		return 0;
	    }

	    channel->channel_info->greeting = strdup(argv[0]);
	}
    }

    chanserv_notice(user, CSMSG_STRING_VALUE, "Greeting     ", channel->channel_info->greeting ? channel->channel_info->greeting : "None.");
    return 1;
}

static OPTION_FUNC(opt_modes)
{
    char modes[MODELEN];

    if(argc > 0)
    {
        if(!check_user_level(user, channel, channel->channel_info->options[optEnfModes], 1, 0))
        {
            chanserv_notice(user, CSMSG_NO_ACCESS);
            return 0;
        }
	if(argv[0][0] == '*' && argv[0][1] == 0)
	{
	    channel->channel_info->mode_on = 0;
	    channel->channel_info->mode_off = 0;
	    modes[0] = 0;
	}
	else
	{
	    if(!parse_mode_lock(argv, argc, channel->channel_info))
	    {
		chanserv_notice(user, CSMSG_INVALID_STR_SET, argv[0], "mode lock");
		return 0;
	    }

	    create_mode_lock(modes, channel->channel_info);
	    mod_chanmode(channel, modes, NULL, 0);
	    irc_mode(chanserv, channel, modes);
	}
    }
    else
    {
	create_mode_lock(modes, channel->channel_info);
    }

    chanserv_notice(user, CSMSG_STRING_VALUE, "Modes        ", strlen(modes) ? modes : "None.");
    return 1;
}

struct valueData
{
    char *description;
    char value;
};

static int
binary_option(char *name, unsigned long mask, struct userNode *user, struct chanNode *channel, int argc, unsigned char *argv[])
{
    struct chanData *cData = channel->channel_info;
    int value;

    if(argc > 0)
    {
	/* Set flag according to value. */
	if(enabled_string(argv[0]))
	{
	    cData->flags |= mask;
	    value = 1;
	}
	else if(disabled_string(argv[0]))
	{
	    cData->flags &= ~mask;
	    value = 0;
	}
	else
	{
	    chanserv_notice(user, MSG_INVALID_BINARY, argv[0]);
	    return 0;
	}
    }
    else
    {
	/* Find current option value. */
	value = (cData->flags & mask) ? 1 : 0;
    }

    chanserv_notice(user, CSMSG_STRING_VALUE, name, value ? "Enabled." : "Disabled.");
    return 1;
}

static int
multiple_option(char *name, char *description, enum opt option, int check_level, struct valueData *values, int count, struct userNode *user, struct chanNode *channel, int argc, unsigned char *argv[])
{
    struct chanData *cData = channel->channel_info;
    int index;

    if(argc > 0)
    {
	/* Set flag according to value. */
        if((check_level != -1) && !check_user_level(user, channel, cData->options[check_level], 1, 1))
        {
            chanserv_notice(user, CSMSG_CANNOT_SET);
            return 0;
        }

        index = atoi(argv[0]);

	if(index < 0 || index >= count)
	{
	    chanserv_notice(user, CSMSG_INVALID_NUM_SET, index, description);
            /* Show possible values. */
            for(index = 0; index < count; index++)
            {
                chanserv_notice(user, CSMSG_NUMERIC_VALUE, name, index, values[index].description);
            }
	    return 0;
	}

        if((check_level != -1) && !check_user_level(user, channel, values[index].value, 1, 1))
        {
            chanserv_notice(user, CSMSG_BAD_SETLEVEL);
            return 0;
        }

	cData->options[option] = values[index].value;
    }
    else
    {
	/* Find current option value. */
	for(index = 0; index < count && cData->options[option] != values[index].value; index++);
    }

    chanserv_notice(user, CSMSG_NUMERIC_VALUE, name, index, values[index].description);
    return 1;
}

static OPTION_FUNC(opt_nodelete)
{
    if((argc > 0) && !IsOper(user))
    {
	chanserv_notice(user, MSG_SETTING_PRIVILEGED, argv[-1]);
	return 0;
    }

    BINARY_OPTION("NoDelete     ", CHANNEL_NODELETE);
}

static OPTION_FUNC(opt_userinfo)
{
    BINARY_OPTION("UserInfo     ", CHANNEL_INFO_LINES);
}

static OPTION_FUNC(opt_voice)
{
    BINARY_OPTION("Voice        ", CHANNEL_VOICE_ALL);
}

static OPTION_FUNC(opt_greet)
{
    BINARY_OPTION("Greet        ", CHANNEL_GREET_ALL);
}

static OPTION_FUNC(opt_greetusers)
{
    BINARY_OPTION("GreetUsers   ", CHANNEL_GREET_USERS);
}

static OPTION_FUNC(opt_dynlimit)
{
    BINARY_OPTION("DynLimit     ", CHANNEL_DYNAMIC_LIMIT);
}

static OPTION_FUNC(opt_topicsnarf)
{
    if((argc > 0) && !check_user_level(user, channel, channel->channel_info->options[optEnfTopic], 1, 0))
    {
        chanserv_notice(user, CSMSG_TOPIC_LOCKED, channel->name);
        return 0;
    }
    BINARY_OPTION("TopicSnarf   ", CHANNEL_TOPIC_SNARF);
}

static OPTION_FUNC(opt_strictop)
{
    struct valueData values[] =
    {
	{"Anyone may op unknown users.", 'a'},
	{"Ops and above may op unknown users.", 'o'},
	{"Masters and above may op unknown users.", 'm'},
	{"Coowners and above may op unknown users.", 'c'},
	{"Owners may op unknown users.", 'w'},
	{"No one may op unknown users.", 'n'}
    };

    MULTIPLE_OPTION("StrictOp     ", "strict op", optStrictOp, optStrictOp);
}

static OPTION_FUNC(opt_autoop)
{
    struct valueData values[] =
    {
	{"Everyone will be automatically opped.", 'a'},
	{"Ops and above will be automatically opped.", 'o'},
	{"Masters and above will be automatically opped.", 'm'},
	{"Coowners and above will be automatically opped.", 'c'},
	{"Owners will be automatically opped.", 'w'},
	{"No one will be automatically opped.", 'n'}
    };

    MULTIPLE_OPTION("AutoOp       ", "auto op", optAutoOp, optAutoOp);
}

static OPTION_FUNC(opt_protect)
{
    struct valueData values[] =
    {
	{"Users will be protected from those of equal or lower access.", 'e'},
	{"Users will be protected from those of lower access.", 'l'},
	{"No users will be protected.", 'n'}
    };

    MULTIPLE_OPTION("Protect      ", "user protection", optProtect, -1);
}

static OPTION_FUNC(opt_enfmodes)
{
    struct valueData values[] =
    {
	{"Anyone may override the mode lock.", 'a'},
	{"Ops and above may override the mode lock.", 'o'},
	{"Masters and above may override the mode lock.", 'm'},
	{"Coowners and above may override the mode lock.", 'c'},
	{"Owners may override the mode lock.", 'w'},
	{"No one may override the mode lock.", 'n'}
    };

    MULTIPLE_OPTION("EnfModes     ", "mode lock enforcement", optEnfModes, optEnfModes);
}

static OPTION_FUNC(opt_enftopic)
{
    struct valueData values[] =
    {
	{"Anyone may override the topic lock.", 'a'},
	{"Ops and above may override the topic lock.", 'o'},
	{"Masters and above may override the topic lock.", 'm'},
	{"Coowners and above may override the topic lock.", 'c'},
	{"Owners may override the topic lock.", 'w'},
	{"No one may override the topic lock.", 'n'}
    };

    MULTIPLE_OPTION("EnfTopic     ", "topic lock enforcement", optEnfTopic, optEnfTopic);
}

static OPTION_FUNC(opt_pubcmd)
{
    struct valueData values[] =
    {
	{"Anyone may use public commands.", 'a'},
        {"Peons and above may use public commands.", 'p'},
	{"Ops and above may use public commands.", 'o'},
	{"Masters and above may use public commands.", 'm'},
	{"Coowners and above may use public commands.", 'c'},
	{"Owners may use public commands.", 'w'},
	{"No one may use public commands.", 'n'}
    };

    MULTIPLE_OPTION("PubCmd       ", "public command", optPubCmd, optPubCmd);
}

static OPTION_FUNC(opt_toys)
{
    struct valueData values[] =
    {
	{"Toys are completely disabled.", 'd'},
	{"Toys will only reply privately.", 'n'},
	{"Toys will reply publicly.", 'p'}
    };

    MULTIPLE_OPTION("Toys         ", "toy control", optToys, -1);
}

static OPTION_FUNC(opt_setters)
{
    struct valueData values[] =
    {
	{"Masters and above may change settings.", 'm'},
	{"Co-owners and above may change settings.", 'c'},
	{"Owner may change settings.", 'w'}
    };

    MULTIPLE_OPTION("Setters      ", "setters", optSetters, optSetters);
}

static OPTION_FUNC(opt_ctcpusers)
{
    struct valueData values[] =
    {
	{"Anyone may send CTCPs to the channel.", 'a'},
        {"Peons and above may send CTCPs to the channel.", 'p'},
	{"Ops and above may send CTCPs to the channel.", 'o'},
	{"Masters and above may send CTCPs to the channel.", 'm'},
	{"Coowners and above may send CTCPs to the channel.", 'c'},
	{"Owners may send CTCPs to the channel.", 'w'},
	{"No one may send CTCPs to the channel.", 'n'}
    };

    MULTIPLE_OPTION("CTCPUsers    ", "CTCP allowed to", optCTCPUsers, optCTCPUsers);
}

static OPTION_FUNC(opt_ctcpreaction)
{
    struct valueData values[] =
    {
        {"Kick on disallowed CTCPs", 'k'},
        {"Kickban on disallowed CTCPs", 'b'},
        {"Short timed ban on disallowed CTCPs", 't'},
        {"Long timed ban on disallowed CTCPs", 'T'}
    };

    MULTIPLE_OPTION("CTCPReaction ", "CTCP reaction", optCTCPReaction, -1);
}

static OPTION_FUNC(opt_topicrefresh)
{
    struct valueData values[] =
    {
        {"Never refresh topic.", 'n'},
        {"Refresh every 3 hours.", '1'},
        {"Refresh every 6 hours.", '2'},
        {"Refresh every 12 hours.", '3'},
        {"Refresh every 24 hours.", '4'}
    };

    MULTIPLE_OPTION("TopicRefresh ", "topic refresh", optTopicRefresh, -1);
}

static CHANSERV_FUNC(cmd_set)
{
    option_func_t *opt;

    if(argc < 2)
    {
	unsigned int i = 0;
	option_func_t *set_display[] = {
            /* multiple-choice options */
	    opt_defaulttopic, opt_topicmask, opt_greeting, opt_modes, opt_pubcmd,
            opt_strictop, opt_autoop, opt_enfmodes, opt_enftopic, opt_protect,
            opt_toys, opt_setters, opt_topicrefresh, opt_ctcpusers,
            opt_ctcpreaction,
            /* binary options */
            opt_voice, opt_greet, opt_greetusers, opt_userinfo, opt_dynlimit,
            opt_topicsnarf, opt_nodelete
	};

	chanserv_notice(user, CSMSG_CHANNEL_OPTIONS);

	/* Do this so options are presented in a consistent order. */
	while(i < ArrayLength(set_display))
	{
	    set_display[i++](user, channel, 0, NULL);
	}

	return 1;
    }

    if(!(opt = dict_find(set_func_dict, argv[1], NULL)))
    {
	chanserv_notice(user, CSMSG_INVALID_OPTION, argv[1]);
	return 0;
    }

    if((argc > 2)
       && !check_user_level(user, channel, channel->channel_info->options[optSetters], 1, 0))
    {
        chanserv_notice(user, CSMSG_NO_ACCESS);
        return 0;
    }

    return opt(user, channel, argc - 2, argv + 2);
}

static CHANSERV_FUNC(cmd_giveownership)
{
    struct handle_info *new_owner_hi;
    struct userData *new_owner, *curr_user;
    REQUIRE_PARAMS(2);
    curr_user = GetChannelAccess(channel->channel_info, user->handle_info);
    if(!curr_user || (curr_user->access != ulOwner))
    {
        chanserv_notice(user, CSMSG_USE_ADDOWNER, argv[0]);
        return 0;
    }
    if(!(new_owner_hi = chanserv_get_handle_info(user, argv[1]))) return 0;
    if(new_owner_hi == user->handle_info)
    {
        chanserv_notice(user, CSMSG_NO_TRANSFER_SELF);
        return 0;
    }
    new_owner = GetChannelAccess(channel->channel_info, new_owner_hi);
    if(!new_owner)
    {
        chanserv_notice(user, CSMSG_NO_CHAN_USER, new_owner_hi->handle, channel->name);
        return 0;
    }
    new_owner->access = ulOwner;
    curr_user->access = ulCoowner;
    chanserv_notice(user, CSMSG_OWNERSHIP_GIVEN, channel->name, new_owner_hi->handle);
    return 1;
}

static CHANSERV_FUNC(cmd_suspend)
{
    struct handle_info *hi;
    struct userData *self, *target;

    REQUIRE_PARAMS(2);
    if(!(hi = chanserv_get_handle_info(user, argv[1]))) return 0;
    self = GetChannelUser(channel->channel_info, user->handle_info);
    if(!(target = GetTrueChannelAccess(channel->channel_info, hi)))
    {
        chanserv_notice(user, CSMSG_NO_CHAN_USER, hi->handle, channel->name);
        return 0;
    }
    if(target->access >= self->access)
    {
        chanserv_notice(user, MSG_USER_OUTRANKED, hi->handle);
        return 0;
    }
    if(target->flags & USER_SUSPENDED)
    {
        chanserv_notice(user, CSMSG_ALREADY_SUSPENDED, hi->handle);
        return 0;
    }
    target->flags |= USER_SUSPENDED;
    chanserv_notice(user, CSMSG_USER_SUSPENDED, hi->handle, channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_unsuspend)
{
    struct handle_info *hi;
    struct userData *self, *target;

    REQUIRE_PARAMS(2);
    if(!(hi = chanserv_get_handle_info(user, argv[1]))) return 0;
    self = GetChannelUser(channel->channel_info, user->handle_info);
    if(!(target = GetTrueChannelAccess(channel->channel_info, hi)))
    {
        chanserv_notice(user, CSMSG_NO_CHAN_USER, hi->handle, channel->name);
        return 0;
    }
    if(target->access >= self->access)
    {
        chanserv_notice(user, MSG_USER_OUTRANKED, hi->handle);
        return 0;
    }
    if(!(target->flags & USER_SUSPENDED))
    {
        chanserv_notice(user, CSMSG_NOT_SUSPENDED, hi->handle);
        return 0;
    }
    target->flags &= ~USER_SUSPENDED;
    chanserv_notice(user, CSMSG_USER_UNSUSPENDED, hi->handle, channel->name);
    return 1;
}

static void
chanserv_refresh_topics(void *data)
{
    unsigned int refresh_num = (now - self->link) / chanserv_conf.refresh_period;
    struct chanData *cData;
    char opt;

    for(cData = channelList; cData; cData = cData->next)
    {
        if(IsSuspended(cData)) continue;
        opt = cData->options[optTopicRefresh];
        if(opt == 'n') continue;
        if((refresh_num - cData->last_refresh) < (unsigned int)(1 << (opt - '1'))) continue;
        if(cData->topic[0]) SetChannelTopic(cData->channel, chanserv, cData->topic, 1);
        cData->last_refresh = refresh_num;
    }
    timeq_add(now + chanserv_conf.refresh_period, chanserv_refresh_topics, data);
}

#ifndef NO_LAME_TRICKS
static CHANSERV_FUNC(cmd_unf)
{
    (void)argc; (void)argv;

    if(channel)
    {
        char *response;
        response = alloca(strlen(user->nick)+strlen(CSMSG_UNF_RESPONSE)+5);
        sprintf(response, "%s: %s", user->nick, CSMSG_UNF_RESPONSE);
        irc_privmsg(chanserv->numeric, channel->name, response);
    }
    else
    {
        chanserv_notice(user, CSMSG_UNF_RESPONSE);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_ping)
{
    (void)argc; (void)argv;

    if(channel)
    {
        char *response;
        response = alloca(strlen(user->nick)+strlen(CSMSG_PING_RESPONSE)+5);
        sprintf(response, "%s: %s", user->nick, CSMSG_PING_RESPONSE);
        irc_privmsg(chanserv->numeric, channel->name, response);
    }
    else
    {
        irc_privmsg(chanserv->numeric, channel->name, CSMSG_PING_RESPONSE);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_wut)
{
    (void)argc; (void)argv;

    if(channel)
    {
        char *response;
        response = alloca(strlen(user->nick)+strlen(CSMSG_WUT_RESPONSE)+5);
        sprintf(response, "%s: %s", user->nick, CSMSG_WUT_RESPONSE);
        irc_privmsg(chanserv->numeric, channel->name, response);
    }
    else
    {
        chanserv_notice(user, CSMSG_WUT_RESPONSE);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_8ball)
{
    unsigned int i, j, accum;
    const char *resp;

    REQUIRE_PARAMS(2);
    accum = 0;
    for (i=1; i<argc; i++) {
        for (j=0; argv[i][j]; j++) {
            accum = (accum << 5) - accum + toupper(argv[i][j]);
        }
    }
    resp = chanserv_conf.eightball->list[accum % chanserv_conf.eightball->used];
    if(channel)
    {
        char *response;
        response = alloca(strlen(user->nick)+strlen(resp)+5);
        sprintf(response, "%s: %s", user->nick, resp);
        irc_privmsg(chanserv->numeric, channel->name, response);
    }
    else
    {
        chanserv_notice(user, "%s", resp);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_d)
{
    int sides, rolled;

    REQUIRE_PARAMS(2);
    if((sides = atoi(argv[1])) <= 1)
    {
        chanserv_notice(user, CSMSG_BAD_NUMBER, argv[1]);
        return 0;
    }
    rolled = (rand() % sides)+1;

    if(channel)
    {
        char *response = alloca(strlen(user->nick)+48);
        sprintf(response, "%s: A %d shows on the %d-sided die.", user->nick, rolled, sides);
        irc_privmsg(chanserv->numeric, channel->name, response);
    }
    else
    {
        chanserv_notice(user, "A %d shows on the %d-sided die.", rolled, sides);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_huggle)
{
    (void)argc; (void)argv;
    if(channel)
    {
        char response[MAXLEN];
        sprintf(response, "\001ACTION huggles %s\001", user->nick);
        irc_privmsg(chanserv->numeric, channel->name, response);
    }
    else
    {
        irc_privmsg(chanserv->numeric, user->numeric, "\001ACTION huggles you\001");
    }
    return 1;
}
#endif /* ndef NO_LAME_TRICKS */

static void
chanserv_adjust_limit(void *data)
{
    struct chanData *cData = data;
    struct chanNode *channel;
    unsigned int limit;
    char change[10];

    if(IsSuspended(cData))
    {
	return;
    }

    channel = cData->channel;
    limit = channel->members.used + chanserv_conf.adjust_threshold + 5;
    cData->limitAdjusted = now;

    if(cData->limit)
    {
	if(limit > cData->limit)
	{
	    limit = cData->limit;
	}
	else if(limit == cData->limit)
	{
	    return;
	}
    }

    sprintf(change, "+l %d", limit);

    channel->limit = limit;
    irc_mode(chanserv, channel, change);
}

static void
handle_new_channel(struct chanNode *channel)
{
    struct chanData *cData;

    if(!(cData = channel->channel_info)) return;

    if(cData->mode_on || cData->mode_off)
    {
	char modes[MODELEN];
	create_mode_lock(modes, cData);
	mod_chanmode(channel, modes, NULL, 0);
	irc_mode(chanserv, channel, modes);
    }

    if(self->uplink && !self->uplink->burst && channel->channel_info->topic[0])
    {
        SetChannelTopic(channel, chanserv, channel->channel_info->topic, 1);
    }
}

/* welcome to my worst nightmare. warning: read the code below
   at your own risk. */
static void
handle_join(struct userNode *user, struct chanNode *channel)
{
    struct chanData *cData;
    struct userData *uData;
    struct banData *bData;
    struct handle_info *handle;
    unsigned int ii;

    if(!channel->channel_info)
    {
	if(channel->members.used < 3)
	{
	    struct suspended *suspended = dict_find(sChannels, channel->name, NULL);

	    if(!suspended)
	    {
		return;
	    }

	    channel->channel_info = suspended->cData;
	    suspended->cData->channel = channel;
	}

	return;
    }

    if(IsLocal(user) || IsSuspended(channel->channel_info)) return;

    cData = channel->channel_info;

    if(channel->members.used > cData->max)
    {
	cData->max = channel->members.used;
    }

    /* Check for bans.  If they're joining through a ban, one of two
     * cases applies:
     *   1: Join during a netburst, by riding the break.  Kick them.
     *   2: They're allowed to join through the ban (an invite in
     *   ircu2.10, or a +e on Hybrid, or something).
     * If they're not joining through a ban, and the banlist is not
     * full, see if they're on the banlist for the channel.  If so,
     * kickban them.
     */
    for(ii=0; ii<channel->banlist.used; ii++)
    {
        if(user_matches_glob(user, channel->banlist.list[ii]->ban, 1)) break;
    }
    if(ii<channel->banlist.used)
    {
        if(self->uplink->burst || user->uplink->burst)
        {
            /* Riding a netburst.  Naughty. */
            KickChannelUser(user, channel, chanserv, CSMSG_NETRIDE_DETECTED);
            return;
        }
    }
    else if(channel->banlist.used < MAXBANS)
    {
        /* Not joining through a ban. */
	for(bData = cData->bans;
	    bData && !user_matches_glob(user, bData->mask, 1);
	    bData = bData->next);

        if(bData)
        {
            char *nick = bData->owner, *reason = bData->reason, *ban;
            char *kick_reason = alloca(strlen(reason) + strlen(nick) + CSMSG_KICK_EXTRA);
            sprintf(kick_reason, CSMSG_KICK_FORMAT, CSMSG_KICK_ARGUMENTS);

	    bData->triggered = now;
            if(bData != cData->bans)
            {
                /* Shuffle the ban to the head of the list. */
                if(bData->next) bData->next->prev = bData->prev;
                if(bData->prev) bData->prev->next = bData->next;

                bData->prev = NULL;
                bData->next = cData->bans;

                if(cData->bans)
	        {
                    cData->bans->prev = bData;
                }
                cData->bans = bData;
            }

            ban = bData->mask;
            AddChannelBan(1, &ban, channel, chanserv, now, 1);
            KickChannelUser(user, channel, chanserv, kick_reason);
            return;
        }
    }

    /* Users aren't going to be authenticated to NickServ during a burst,
       so we might as well short-circuit all of this stuff during bursts.
    */
    if(self->uplink->burst || user->uplink->burst)
    {
	return;
    }

    if((channel == chanserv_conf.support_channel) && IsSupportHelper(user))
    {
        HANDLE_SET_FLAG(user->handle_info, HELPING);
    }

    if(cData->flags & CHANNEL_DYNAMIC_LIMIT)
    {
	if((channel->limit - channel->members.used) < chanserv_conf.adjust_threshold)
	{
	    /* The user count has begun "bumping" into the channel
	       limit, so set a timer to raise the limit a bit. Any
	       previous timers are removed so three incoming users
	       within the delay results in one limit change, not
	       three. */

	    timeq_del(0, chanserv_adjust_limit, cData, TIMEQ_IGNORE_WHEN);
	    timeq_add(now + chanserv_conf.adjust_delay, chanserv_adjust_limit, cData);
	}
    }

    if(validate_auto_op(user, channel))
    {
	AddChannelOp(1, &user, channel, chanserv, 1);
    }
    else if(cData->flags & CHANNEL_VOICE_ALL)
    {
        AddChannelVoice(1, &user, channel, chanserv, 1);
    }

    handle = user->handle_info;
    if(handle && (uData = GetTrueChannelAccess(cData, handle)))
    {
	/* Ops and above were handled by the above case. */
	if((uData->access == ulPeon)
           && !(uData->flags & USER_SUSPENDED)
           && !(cData->flags & CHANNEL_VOICE_ALL))
	{
	    AddChannelVoice(1, &user, channel, chanserv, 1);
	}

	if((uData->access >= ulMaster) && !(uData->flags & USER_SUSPENDED))
	{
	    cData->visited = now;
	}

	if(uData->info
           && !(uData->flags & USER_SUSPENDED)
           && (cData->flags & CHANNEL_INFO_LINES)
           && ((now - uData->seen) >= chanserv_conf.info_delay)
           && !uData->present)
	{
	    char *info;
	    info = alloca(strlen(uData->info) + strlen(user->nick) + 4);
	    sprintf(info, "[%s] %s", user->nick, uData->info);
	    irc_privmsg(chanserv->numeric, channel->name, info);
	}

	uData->seen = now;
	uData->present = 1;

        if((cData->flags & CHANNEL_GREET_USERS) && cData->greeting)
        {
            chanserv_notice(user, "(%s) %s", channel->name, cData->greeting);
        }
    } else {
        if((cData->flags & CHANNEL_GREET_ALL) && cData->greeting)
        {
            chanserv_notice(user, "(%s) %s", channel->name, cData->greeting);
        }
    }
}

static void
handle_auth(struct userNode *user, struct handle_info *old_handle)
{
    struct chanList *channel;

    /* If the user was authenticated to a previous handle, reset presence
       for the previous handle's channels. */
    if(old_handle)
    {
	for(channel = old_handle->channels; channel; channel = channel->next)
	{
	    if(!channel->channel->channel || IsSuspended(channel->channel))
	    {
		continue;
	    }
	    channel->user->seen = now;
	    scan_handle_presence(channel->channel->channel, old_handle, user);
	}
    }

    if(!user->handle_info) return;
    for(channel = user->handle_info->channels; channel; channel = channel->next)
    {
        struct chanNode *cn;
        if(IsSuspended(channel->channel) || !(cn = channel->channel->channel) || !GetUserMode(cn, user)) continue;

	if(channel->user->access >= ulMaster)
	{
	    channel->channel->visited = now;
	}

        if(validate_auto_op(user, cn))
        {
            AddChannelOp(1, &user, cn, chanserv, 1);
        }
        else if(channel->user->access == ulPeon)
        {
            AddChannelVoice(1, &user, cn, chanserv, 1);
        }

	channel->user->seen = now;
	channel->user->present = 1;
    }

    if(chanserv_conf.support_channel && IsSupportHelper(user)
       && GetUserMode(chanserv_conf.support_channel, user))
    {
        HANDLE_SET_FLAG(user->handle_info, HELPING);
    }
}

static void
handle_part(struct userNode *user, struct chanNode *channel)
{
    struct chanData *cData;
    struct userData *uData = NULL;
    struct handle_info *handle;

    if(!channel->channel_info || IsSuspended(channel->channel_info) || IsLocal(user)) return;
    cData = channel->channel_info;

    if((cData->flags & CHANNEL_DYNAMIC_LIMIT) && cData->limit)
    {
	/* Allow for a bit of padding so that the limit doesn't
	   track the user count exactly, which could get annoying. */
	if((channel->limit - channel->members.used) > chanserv_conf.adjust_threshold + 5)
	{
	    timeq_del(0, chanserv_adjust_limit, cData, TIMEQ_IGNORE_WHEN);
	    timeq_add(now + chanserv_conf.adjust_delay, chanserv_adjust_limit, cData);
	}
    }

    if((handle = user->handle_info) && (uData = GetTrueChannelAccess(cData, handle)))
    {
	uData->seen = now;
	scan_handle_presence(channel, handle, user);
    }

    if((channel == chanserv_conf.support_channel) && IsSupportHelper(user))
    {
        HANDLE_CLEAR_FLAG(user->handle_info, HELPING);
    }

    return;
}

static int
handle_kick(struct userNode *kicker, struct userNode *victim, struct chanNode *channel)
{
    char *reason = CSMSG_PROTECT_REASON;

    if(!channel->channel_info || IsService(kicker) || (kicker == victim) || IsSuspended(channel->channel_info)) return 0;

    if(protect_user(victim, kicker, channel->channel_info))
    {
	KickChannelUser(kicker, channel, chanserv, reason);
    }

    return 0;
}

static int
handle_topic(struct userNode *user, struct chanNode *channel, const char *old_topic)
{
    struct chanData *cData;

    if(!channel->channel_info || !user || IsSuspended(channel->channel_info) || IsService(user)) return 0;

    cData = channel->channel_info;
    if(bad_topic(channel, user, channel->topic))
    {
        chanserv_notice(user, CSMSG_TOPIC_LOCKED, channel->name);
        if(cData->topic_mask && match_ircglob(old_topic, cData->topic_mask))
        {
            SetChannelTopic(channel, chanserv, old_topic, 1);
        }
        else
        {
            SetChannelTopic(channel, chanserv, cData->topic, 1);
        }
        return 1;
    }
    if(cData->flags & CHANNEL_TOPIC_SNARF)
    {
	/* Grab the topic and save it as the default topic. */
	safestrncpy(cData->topic, channel->topic, sizeof(cData->topic));
    }
    return 0;
}

static void
handle_mode(struct chanNode *channel, struct userNode *user, int change_type, void *change)
{
    if(!channel->channel_info || IsLocal(user) || IsSuspended(channel->channel_info) || IsService(user)) return;

    if(change_type == MODE_CHANGE_MISC)
    {
	chan_mode_t on;
	(void)change;

	if(!check_user_level(user, channel, channel->channel_info->options[optEnfModes], 1, 0))
	{
            on = channel->channel_info->mode_on;
	    if((on && ((channel->modes & on) != on)) || (channel->modes & channel->channel_info->mode_off))
	    {
		char correct[MODELEN];
		create_mode_lock(correct, channel->channel_info);

		mod_chanmode(channel, correct, NULL, 0);
		irc_mode(chanserv, channel, correct);

		chanserv_notice(user, CSMSG_MODE_LOCKED, correct, channel->name);
	    }
	}
    }
    else if(change_type == MODE_CHANGE_OP)
    {
	struct userNode *victim = change;

	if(!validate_op(user, channel, victim) && !IsService(victim))
	{
	    DelChannelOp(1, &victim, channel, chanserv, 1);
	}
    }
    else if(change_type == MODE_CHANGE_DEOP)
    {
	struct userNode *victim = change;

	if(protect_user(victim, user, channel->channel_info))
	{
	    char change[7 + (2 * COMBO_NUMERIC_LEN)];

	    DelChannelOp(1, &user, channel, chanserv, 0);
	    AddChannelOp(1, &victim, channel, chanserv, 0);

	    sprintf(change, "+o-o %s %s", victim->numeric, user->numeric);

	    irc_mode(chanserv, channel, change);
	    chanserv_notice(user, CSMSG_USER_PROTECTED, victim->nick);
	}
    }
    else if(change_type == MODE_CHANGE_BAN)
    {
	if(bad_channel_ban(channel, user, change, NULL, NULL))
	{
	    char *remove = change;
	    DelChannelBan(1, &remove, channel, chanserv, 1);
	    chanserv_notice(user, CSMSG_MASK_PROTECTED, remove);
	    return;
	}	
    }
}

static void
handle_unreg(struct userNode *user, struct handle_info *handle)
{
    struct chanList *channel, *next;
    struct userNode *h_user;

    (void)user;
    for(h_user = handle->users; h_user; h_user = h_user->next_authed)
    {
        chanserv_notice(h_user, CSMSG_HANDLE_UNREGISTERED);
    }
    for(channel = handle->channels; channel; channel = next)
    {
	/* del_channel_user() deletes the chanList nodes, so we
	   don't worry about them. */
	next = channel->next;
	del_channel_user(channel->channel, channel->user);
    }

    return;
}

static void
handle_server_link(struct server *server)
{
    struct chanData *cData;

    (void)server;
    for(cData = channelList; cData; cData = cData->next)
    {
        if(!IsSuspended(cData))
        {
            cData->may_opchan = 1;
        }
    }
}

static void
chanserv_helpfile_read(void)
{
    chanserv_helpfile = open_helpfile(CHANSERV_HELPFILE_NAME, chanserv_expand_variable);
}

static CHANSERV_FUNC(cmd_help)
{
    const char *topic;
    (void)channel;
    if(argc < 2)
    {
	topic = NULL;
    }
    else
    {
	topic = argv[1];
	unsplit_string(argv+1, argc-1);
    }
    chanserv_help(user, topic, 1);
    return 1;
}

static CHANSERV_FUNC(cmd_readhelp)
{
    struct timeval start, stop;
    struct helpfile *old_helpfile = chanserv_helpfile;
    (void)channel; (void)argc; (void)argv;

    gettimeofday(&start, NULL);
    chanserv_helpfile_read();
    if(chanserv_helpfile)
    {
	close_helpfile(old_helpfile);
    }
    else
    {
	chanserv_helpfile = old_helpfile;
    }
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if(stop.tv_usec < 0)
    {
	stop.tv_sec -= 1;
	stop.tv_usec += 1000000;
    }
    chanserv_notice(user, CSMSG_HELP_READ_SUCCESS, stop.tv_sec, stop.tv_usec/1000);
    log(CS_LOG, LOG_INFO, "%s ordered a help file read.\n", user->nick);

    return 1;
}

static void
chanserv_conf_read(void)
{
    dict_t conf_node;
    const char *str;
    struct string_list *strlist;

    if(!(conf_node = conf_get_data(CHANSERV_CONF_NAME, RECDB_OBJECT)))
    {
	log(CS_LOG, LOG_ERROR, "Invalid config node `%s'.\n", CHANSERV_CONF_NAME);
	return;
    }
    str = database_get_data(conf_node, KEY_DEBUG_CHANNEL, RECDB_QSTRING);
    if(str)
    {
	chanserv_conf.debug_channel = AddChannel(str, now, "+tinms", NULL, 0, chanserv->numeric, NULL);
	AddChannelOp(1, &chanserv, chanserv_conf.debug_channel, chanserv, 0);
    }
    else
    {
	chanserv_conf.debug_channel = NULL;
    }
    str = database_get_data(conf_node, KEY_SUPPORT_CHANNEL, RECDB_QSTRING);
    if(str)
    {
        chanserv_conf.support_channel = AddChannel(str, now, "+nt", NULL, 0, chanserv->numeric, NULL);
        AddChannelOp(1, &chanserv, chanserv_conf.support_channel, chanserv, 0);
    }
    else
    {
        chanserv_conf.support_channel = NULL;
    }
    str = database_get_data(conf_node, KEY_DB_BACKUP_FREQ, RECDB_QSTRING);
    chanserv_conf.db_backup_frequency = str ? ParseInterval(str) : 7200;
    str = database_get_data(conf_node, KEY_INFO_DELAY, RECDB_QSTRING);
    chanserv_conf.info_delay = str ? ParseInterval(str) : 180;
    str = database_get_data(conf_node, KEY_MAX_GREETLEN, RECDB_QSTRING);
    chanserv_conf.greeting_length = str ? atoi(str) : 120;
    str = database_get_data(conf_node, KEY_ADJUST_THRESHOLD, RECDB_QSTRING);
    chanserv_conf.adjust_threshold = str ? atoi(str) : 15;
    str = database_get_data(conf_node, KEY_JOINFLOOD_THRESHOLD, RECDB_QSTRING);
    chanserv_conf.join_flood_threshold = str ? atoi(str) : 5;
    str = database_get_data(conf_node, KEY_ADJUST_DELAY, RECDB_QSTRING);
    chanserv_conf.adjust_delay = str ? ParseInterval(str) : 30;
    str = database_get_data(conf_node, KEY_CHAN_EXPIRE_FREQ, RECDB_QSTRING);
    chanserv_conf.channel_expire_frequency = str ? ParseInterval(str) : 86400;
    str = database_get_data(conf_node, KEY_CHAN_EXPIRE_DELAY, RECDB_QSTRING);
    chanserv_conf.channel_expire_delay = str ? ParseInterval(str) : 86400*30;
    str = database_get_data(conf_node, KEY_MAX_CHAN_USERS, RECDB_QSTRING);
    chanserv_conf.max_chan_users = str ? atoi(str) : 512;
    str = database_get_data(conf_node, KEY_MAX_CHAN_BANS, RECDB_QSTRING);
    chanserv_conf.max_chan_bans = str ? atoi(str) : 512;
    str = database_get_data(conf_node, KEY_NICK, RECDB_QSTRING);
    if(str) NickChange(chanserv, str, 0);
    str = database_get_data(conf_node, KEY_TRIGGER, RECDB_QSTRING);
    chanserv_conf.trigger_char = str ? str[0] : 0;
    str = database_get_data(conf_node, KEY_LAME_TRICKS, RECDB_QSTRING);
    chanserv_conf.lame_tricks = str ? atoi(str) : 1;
    str = database_get_data(conf_node, KEY_REFRESH_PERIOD, RECDB_QSTRING);
    chanserv_conf.refresh_period = str ? ParseInterval(str) : 3*60*60;
    str = database_get_data(conf_node, KEY_CTCP_SHORT_BAN_DURATION, RECDB_QSTRING);
    chanserv_conf.ctcp_short_ban_duration = str ? strdup(str) : strdup("3m");
    str = database_get_data(conf_node, KEY_CTCP_LONG_BAN_DURATION, RECDB_QSTRING);
    chanserv_conf.ctcp_long_ban_duration = str ? strdup(str) : strdup("1h");
    str = database_get_data(conf_node, KEY_MAX_OWNED, RECDB_QSTRING);
    chanserv_conf.max_owned = str ? atoi(str) : 5;
    if(chanserv_conf.lame_tricks)
    {
        strlist = database_get_data(conf_node, KEY_8BALL_RESPONSES, RECDB_STRING_LIST);
        if (strlist) {
            strlist = string_list_copy(strlist);
        } else {
            strlist = alloc_string_list(4);
            string_list_append(strlist, strdup("Yes."));
            string_list_append(strlist, strdup("No."));
            string_list_append(strlist, strdup("Maybe so."));
        }
        chanserv_conf.eightball = strlist;
    }
    if(TRANSITION)
    {
        str = database_get_data(conf_node, KEY_OLD_CHANSERV_NAME, RECDB_QSTRING);
        chanserv_conf.old_chanserv_name = str;
        str = database_get_data(conf_node, KEY_MAX_SWITCH_LOAD, RECDB_QSTRING);
        chanserv_conf.max_switch_load = str ? atoi(str) : 64;
        str = database_get_data(conf_node, KEY_SWITCH_TIMEOUT, RECDB_QSTRING);
        chanserv_conf.switch_timeout = str ? ParseInterval(str) : 30;
        str = database_get_data(conf_node, KEY_ACCESS_REGEX, RECDB_QSTRING);
        if(str)
        {
            int res = regcomp(&chanserv_conf.access_regex, str, REG_EXTENDED);
            if(res)
            {
                log(CS_LOG, LOG_ERROR, "regcomp on access_regex returned error code %d\n", res);
                chanserv_conf.old_chanserv_name = NULL;
            }
        }
        else
        {
            chanserv_conf.old_chanserv_name = NULL;
        }
        str = database_get_data(conf_node, KEY_AUTH_REGEX, RECDB_QSTRING);
        if(str)
        {
            int res = regcomp(&chanserv_conf.auth_regex, str, REG_EXTENDED);
            if(res)
            {
                log(CS_LOG, LOG_ERROR, "regcomp on auth_regex returned error code %d\n", res);
                chanserv_conf.old_chanserv_name = NULL;
            }
        } else {
            chanserv_conf.old_chanserv_name = NULL;
        }
        if(chanserv_conf.old_chanserv_name)
        {
            switch_dict = dict_new();
            dict_set_free_data(switch_dict, free);
        }
    }
}

int chanserv_command_search(const char *command)
{
    struct chanserv_cmd *cmd = dict_find(chanserv_func_dict, command, NULL);
    return cmd ? 1 : 0;
}

int chanserv_add_alias(const char *alias, const char *command)
{
    struct chanserv_cmd *cmd = dict_find(chanserv_func_dict, command, NULL);
    if(!cmd) return 0;
    cmd->refs++;
    dict_insert(chanserv_func_dict, alias, cmd);
    return 1;
}

int chanserv_del_alias(const char *alias)
{
    return dict_remove(chanserv_func_dict, alias);
}

static struct chanserv_cmd *
chanserv_define_func(const char *name, long flags, enum userLevel access, chanserv_func_t *func)
{
    struct chanserv_cmd *cmd = malloc(sizeof(struct chanserv_cmd));
    cmd->func = func;
    cmd->weight = 1.0;
    cmd->access = access;
    cmd->flags = flags;
    cmd->uses = 0;
    cmd->refs = 1;

    dict_insert(chanserv_func_dict, name, cmd);

    return cmd;
}

static void
chanserv_unref_cmd(void *data)
{
    struct chanserv_cmd *cmd = data;
    if(!--cmd->refs) free(cmd);
}

static enum userLevel
leveltoaccess(int level)
{
    enum userLevel access;

    for(access = USER_LEVEL_HIGHEST; access >= USER_LEVEL_LOWEST; access--)
    {
	if(level >= levelMap[access])
	{
	    return access;
	}
    }

    return ulNone;
}

static void
chanserv_note_type_read(const char *key, struct record_data *rd)
{
    dict_t obj;
    struct note_type *ntype;
    const char *str;

    if(!(obj = GET_RECORD_OBJECT(rd)))
    {
        log(CS_LOG, LOG_ERROR, "Invalid note type %s.\n", key);
        return;
    }
    if(!(ntype = create_note_type(key)))
    {
        log(CS_LOG, LOG_ERROR, "Memory allocation failed for note %s.\n", key);
        return;
    }

    /* Figure out set access */
    if((str = database_get_data(obj, KEY_NOTE_OPSERV_ACCESS, RECDB_QSTRING)))
    {
        ntype->set_access_type = NOTE_SET_PRIVILEGED;
        ntype->set_access.min_opserv = strtoul(str, NULL, 0);
    }
    else if((str = database_get_data(obj, KEY_NOTE_CHANNEL_ACCESS, RECDB_QSTRING)))
    {
        ntype->set_access_type = NOTE_SET_CHANNEL_ACCESS;
        ntype->set_access.min_ulevel = leveltoaccess(strtoul(str, NULL, 0));
    }
    else if((str = database_get_data(obj, KEY_NOTE_SETTER_ACCESS, RECDB_QSTRING)))
    {
        ntype->set_access_type = NOTE_SET_CHANNEL_SETTER;
    }
    else
    {
        log(CS_LOG, LOG_ERROR, "Could not find access type for note %s; defaulting to OpServ access level 0.\n");
        ntype->set_access_type = NOTE_SET_PRIVILEGED;
        ntype->set_access.min_opserv = 0;
    }

    /* Figure out visiblity */
    if((str = database_get_data(obj, KEY_NOTE_VISIBILITY, RECDB_QSTRING)))
    {
        if(!strcasecmp(str, KEY_NOTE_VIS_PRIVILEGED))
        {
            ntype->visible_type = NOTE_VIS_PRIVILEGED;
        }
        else if(!strcasecmp(str, KEY_NOTE_VIS_CHANNEL_USERS))
        {
            ntype->visible_type = NOTE_VIS_CHANNEL_USERS;
        }
        else if(!strcasecmp(str, KEY_NOTE_VIS_ALL))
        {
            ntype->visible_type = NOTE_VIS_ALL;
        }
        else
        {
            ntype->visible_type = NOTE_VIS_PRIVILEGED;
        }
    }
    else
    {
        ntype->set_access_type = NOTE_VIS_PRIVILEGED;
    }

    str = database_get_data(obj, KEY_NOTE_MAX_LENGTH, RECDB_QSTRING);
    ntype->max_length = str ? strtoul(str, NULL, 0) : 400;
}

static int 
user_read_helper(const char *key, struct record_data *rd, struct chanData *chan)
{
    enum userLevel access;
    char *seen, *inf, *flags;
    time_t last_seen;
    int level;
    struct handle_info *handle;
    struct userData *uData;

    if(rd->type != RECDB_OBJECT || !dict_size(rd->d.object))
    {
	log(CS_LOG, LOG_ERROR, "Invalid user in %s.\n", chan->channel->name);
	return 0;
    }

    level = atoi(database_get_data(rd->d.object, KEY_LEVEL, RECDB_QSTRING));
    access = leveltoaccess(level);

    if(access > USER_LEVEL_HIGHEST || access < USER_LEVEL_LOWEST)
    {
	log(CS_LOG, LOG_ERROR, "Invalid access level for %s in %s.\n", key, chan->channel->name);
	return 0;
    }

    inf = database_get_data(rd->d.object, KEY_INFO, RECDB_QSTRING);
    seen = database_get_data(rd->d.object, KEY_SEEN, RECDB_QSTRING);
    last_seen = seen ? (signed)strtoul(seen, NULL, 0) : now;
    flags = database_get_data(rd->d.object, KEY_FLAGS, RECDB_QSTRING);
    handle = get_handle_info(key);
    if(!handle)
    {
        log(CS_LOG, LOG_ERROR, "Nonexistent account %s in %s.\n", key, chan->channel->name);
        return 0;
    }

    uData = add_channel_user(chan, handle, access, last_seen, inf);
    uData->flags = flags ? strtoul(flags, NULL, 0) : 0;

    return 0;
}

static int
ban_read_helper(const char *key, struct record_data *rd, struct chanData *chan)
{
    struct banData *bData;
    char *set, *triggered, *s_duration, *s_expires, *reason, *owner;
    time_t set_time, triggered_time, expires_time;

    if(rd->type != RECDB_OBJECT || !dict_size(rd->d.object))
    {
	log(CS_LOG, LOG_ERROR, "Invalid ban in %s.\n", chan->channel->name);
	return 0;
    }

    set = database_get_data(rd->d.object, KEY_SET, RECDB_QSTRING);
    triggered = database_get_data(rd->d.object, KEY_TRIGGERED, RECDB_QSTRING);
    s_duration = database_get_data(rd->d.object, KEY_DURATION, RECDB_QSTRING);
    s_expires = database_get_data(rd->d.object, KEY_EXPIRES, RECDB_QSTRING);
    owner = database_get_data(rd->d.object, KEY_OWNER, RECDB_QSTRING);
    reason = database_get_data(rd->d.object, KEY_REASON, RECDB_QSTRING);

    set_time = set ? (time_t)strtoul(set, NULL, 0) : now;
    triggered_time = triggered ? (time_t)strtoul(triggered, NULL, 0) : 0;
    if(s_expires)
    {
        expires_time = strtoul(s_expires, NULL, 0);
    }
    else if(s_duration)
    {
        expires_time = set_time + strtoul(s_duration, NULL, 0);
    }
    else
    {
        expires_time = 0;
    }

    if(expires_time && (expires_time < now))
    {
	/* The ban has expired. */
	return 0;
    }

    bData = add_channel_ban(chan, key, owner, set_time, triggered_time, expires_time, reason);
    return 0;
}

static int
chanserv_channel_read(const char *key, struct record_data *hir)
{
    struct chanNode *cNode = NULL;
    struct chanData *cData = NULL;
    dict_t users, bans, notes;
    dict_iterator_t it;

    char *flags, *topic, *topic_mask, *modes, *data, *greeting, *registrar, *suspender;
    char *options;
    unsigned long num_flags;
    unsigned int max;
    time_t registered, visited;

    users = database_get_data(hir->d.object, KEY_USERS, RECDB_OBJECT);

    if(!users)
    {
	log(CS_LOG, LOG_ERROR, "%s has no users.\n", key);
	return 0;
    }

    bans = database_get_data(hir->d.object, KEY_BANS, RECDB_OBJECT);
    notes = database_get_data(hir->d.object, KEY_NOTES, RECDB_OBJECT);
    flags = database_get_data(hir->d.object, KEY_FLAGS, RECDB_QSTRING);
    topic = database_get_data(hir->d.object, KEY_TOPIC, RECDB_QSTRING);
    topic_mask = database_get_data(hir->d.object, KEY_TOPIC_MASK, RECDB_QSTRING);
    registrar = database_get_data(hir->d.object, KEY_REGISTRAR, RECDB_QSTRING);
    greeting = database_get_data(hir->d.object, KEY_GREETING, RECDB_QSTRING);
    modes = database_get_data(hir->d.object, KEY_MODES, RECDB_QSTRING);

    data = database_get_data(hir->d.object, KEY_REGISTERED, RECDB_QSTRING);
    registered = data ? (signed)strtoul(data, NULL, 0) : now;

    data = database_get_data(hir->d.object, KEY_VISITED, RECDB_QSTRING);
    visited = data ? (signed)strtoul(data, NULL, 0) : now;

    data = database_get_data(hir->d.object, KEY_MAX, RECDB_QSTRING);
    max = data ? strtoul(data, NULL, 0) : 0;

    if(flags)
    {
	num_flags = base64toint(flags, FLAGS_STRING_LENGTH);
	options = flags + FLAGS_STRING_LENGTH;
    }
    else
    {
	num_flags = CHANNEL_DEFAULT_FLAGS;
	options = CHANNEL_DEFAULT_OPTIONS;
    }
    
    if(!(num_flags & CHANNEL_SUSPENDED))
    {
	cNode = AddChannel(key, now, NULL, NULL, 0, chanserv->numeric, NULL);
        AddChannelOp(1, &chanserv, cNode, chanserv, 0);
    }

    cData = register_channel(registrar, topic, greeting, modes, options, num_flags, cNode);

    if(!cData)
    {
        log(CS_LOG, LOG_ERROR, "Unable to register channel %s from database.\n", key);
	return 0;
    }

    if(num_flags & CHANNEL_SUSPENDED)
    {
	struct suspended *suspended;
        char *reason, *exp_text;
        time_t expires = 0;

        if((exp_text = database_get_data(hir->d.object, KEY_SUSPEND_EXPIRES, RECDB_QSTRING)))
        {
            expires = strtoul(exp_text, NULL, 0);
            if(expires < now)
            {
                cData->flags &= ~CHANNEL_SUSPENDED;
                goto suspended_done;
            }
        }
	suspended = malloc(sizeof(struct suspended));
        if((suspended->expires = expires))
        {
            timeq_add(suspended->expires, chanserv_expire_suspension, suspended);
        }
	suspended->name = strdup(key);
        suspender = database_get_data(hir->d.object, KEY_SUSPENDER, RECDB_QSTRING);
	suspended->suspender = strdup(suspender);
        reason = database_get_data(hir->d.object, KEY_SUSPEND_REASON, RECDB_QSTRING);
	suspended->reason = reason ? strdup(reason) : NULL;

	suspended->cData = cData;
	cData->suspended = suspended;

	dict_insert(sChannels, suspended->name, suspended);
      suspended_done: ;
    }

    cData->registered = registered;
    cData->visited = visited;
    cData->max = max;
    if(topic_mask) cData->topic_mask = strdup(topic_mask);

    for(it = dict_first(users); it; it = iter_next(it))
    {
	user_read_helper(iter_key(it), iter_data(it), cData);
    }

    if(!cData->users)
    {
        log(CS_LOG, LOG_ERROR, "Channel %s had no users in database, unregistering it.\n", key);
	unregister_channel(cData);
	if(cNode)
	{
	    DelChannelUser(chanserv, cNode, "Channel has empty user list.", 0);
	}
        return 0;
    }

    for(it = dict_first(bans); it; it = iter_next(it))
    {
        ban_read_helper(iter_key(it), iter_data(it), cData);
    }

    for(it = dict_first(notes); it; it = iter_next(it))
    {
        struct note_type *ntype = dict_find(note_types, iter_key(it), NULL);
        struct record_data *rd = iter_data(it);
        const char *note, *setter;

        if(rd->type != RECDB_OBJECT)
        {
            log(CS_LOG, LOG_ERROR, "Bad record type for note %s in channel %s.\n", iter_key(it), key);
        }
        else if(!ntype)
        {
            log(CS_LOG, LOG_ERROR, "Bad note type name %s in channel %s.\n", iter_key(it), key);
        }
        else if(!(note = database_get_data(rd->d.object, KEY_NOTE_NOTE, RECDB_QSTRING)))
        {
            log(CS_LOG, LOG_ERROR, "Missing note text for note %s in channel %s.\n", iter_key(it), key);
        }
        else
        {
            setter = database_get_data(rd->d.object, KEY_NOTE_SETTER, RECDB_QSTRING);
            if(!setter) setter = "<unknown>";
            chanserv_add_channel_note(cData, ntype, setter, note);
        }
    }

    return 0;
}

static int
chanserv_command_read(const char *key, struct record_data *hir)
{
    struct chanserv_cmd *cmd = dict_find(chanserv_func_dict, key, NULL);
    char *data;

    if(!cmd)
    {
	return 0;
    }

    data = database_get_data(hir->d.object, KEY_ACCESS, RECDB_QSTRING);
    if(data)
    {
	cmd->access = (enum userLevel)atoi(data);
    }

    data = database_get_data(hir->d.object, KEY_FLAGS, RECDB_QSTRING);
    if(data)
    {
	/* Make sure locked flags don't get turned off. Allow people
	   to restrict commands more, though. */
	cmd->flags = strtoul(data, NULL, 0) | (cmd->flags & CMD_REQUIRE_LOCK);
    }

    data = database_get_data(hir->d.object, KEY_WEIGHT, RECDB_QSTRING);
    if(data)
    {
	cmd->weight = atof(data);
    }

    return 0;
}

static void
chanserv_dnr_read(const char *key, struct record_data *hir)
{
    const char *setter, *reason;
    setter = database_get_data(hir->d.object, KEY_DNR_SETTER, RECDB_QSTRING);
    if(!setter)
    {
        log(CS_LOG, LOG_ERROR, "Missing setter for DNR %s.\n", key);
        return;
    }
    reason = database_get_data(hir->d.object, KEY_DNR_REASON, RECDB_QSTRING);
    if(!reason)
    {
        log(CS_LOG, LOG_ERROR, "Missing reason for DNR %s.\n", key);
        return;
    }
    chanserv_add_dnr(key, setter, reason);
}

static void
chanserv_db_read(const char *fname)
{
    dict_t database, section;
    dict_iterator_t it;

    database = parse_database(fname);

    if(!database) return;

    if((section = database_get_data(database, KEY_NOTE_TYPES, RECDB_OBJECT)))
    {
        for(it = dict_first(section); it; it = iter_next(it))
        {
            chanserv_note_type_read(iter_key(it), iter_data(it));
        }
    }

    if((section = database_get_data(database, KEY_CHANNELS, RECDB_OBJECT)))
    {
	for(it = dict_first(section); it; it = iter_next(it))
	{
	    chanserv_channel_read(iter_key(it), iter_data(it));
	}
    }

    if((section = database_get_data(database, KEY_COMMANDS, RECDB_OBJECT)))
    {
	for(it = dict_first(section); it; it = iter_next(it))
	{
	    chanserv_command_read(iter_key(it), iter_data(it));
	}
    }

    if((section = database_get_data(database, KEY_DNR, RECDB_OBJECT)))
    {
        for(it = dict_first(section); it; it = iter_next(it))
        {
            chanserv_dnr_read(iter_key(it), iter_data(it));
        }
    }

    free_database(database);
}

static void
chanserv_write_users(dict_t channel, struct userData *uData)
{
    dict_t users = alloc_object();
    char *str;

    while(uData)
    {
	dict_t user = alloc_object();

	str = alloca(12);
	snprintf(str, 12, "%d", levelMap[uData->access]);
	dict_insert(user, KEY_LEVEL, alloc_record_data_qstring(str));

	snprintf(str, 12, FMT_TIME_T, uData->seen);
	dict_insert(user, KEY_SEEN, alloc_record_data_qstring(str));

        if(uData->flags)
        {
            snprintf(str, 12, "%u", uData->flags);
            dict_insert(user, KEY_FLAGS, alloc_record_data_qstring(str));
        }

	if(uData->info) dict_insert(user, KEY_INFO, alloc_record_data_qstring(uData->info));

	dict_insert(users, uData->handle->handle, alloc_record_data_object(user));
	uData = uData->next;
    }

    dict_insert(channel, KEY_USERS, alloc_record_data_object(users));
    return;
}

static void
chanserv_write_bans(dict_t channel, struct banData *bData)
{
    dict_t bans = alloc_object();
    char *str;

    while(bData)
    {
	dict_t ban = alloc_object();

	str = alloca(12);
	snprintf(str, 12, FMT_TIME_T, bData->set);
	dict_insert(ban, KEY_SET, alloc_record_data_qstring(str));

        if(bData->triggered)
        {
            snprintf(str, 12, FMT_TIME_T, bData->triggered);
            dict_insert(ban, KEY_TRIGGERED, alloc_record_data_qstring(str));
        }

	if(bData->expires)
	{
	    snprintf(str, 12, "%li", bData->expires);
	    dict_insert(ban, KEY_EXPIRES, alloc_record_data_qstring(str));
	}

	if(bData->owner[0]) dict_insert(ban, KEY_OWNER, alloc_record_data_qstring(bData->owner));
	if(bData->reason) dict_insert(ban, KEY_REASON, alloc_record_data_qstring(bData->reason));

	dict_insert(bans, bData->mask, alloc_record_data_object(ban));
	bData = bData->next;
    }

    dict_insert(channel, KEY_BANS, alloc_record_data_object(bans));
    return;
}

static void
chanserv_write_channel(dict_t db, struct chanData *channel)
{
    dict_t obj = alloc_object();
    char *str, flags[FLAGS_STRING_LENGTH + OPTION_STRING_SIZE], *channame;

    str = alloca(12);
    snprintf(str, 12, FMT_TIME_T, channel->registered);
    dict_insert(obj, KEY_REGISTERED, alloc_record_data_qstring(str));

    snprintf(str, 12, FMT_TIME_T, channel->visited);
    dict_insert(obj, KEY_VISITED, alloc_record_data_qstring(str));

    snprintf(str, 12, "%d", channel->max);
    dict_insert(obj, KEY_MAX, alloc_record_data_qstring(str));

    if(channel->topic[0])
    {
	dict_insert(obj, KEY_TOPIC, alloc_record_data_qstring(channel->topic));
    }

    if(channel->registrar)
    {
	dict_insert(obj, KEY_REGISTRAR, alloc_record_data_qstring(channel->registrar));
    }

    if(channel->greeting)
    {
	dict_insert(obj, KEY_GREETING, alloc_record_data_qstring(channel->greeting));
    }

    if(channel->topic_mask)
    {
        dict_insert(obj, KEY_TOPIC_MASK, alloc_record_data_qstring(channel->topic_mask));
    }

    if(IsSuspended(channel))
    {
	dict_insert(obj, KEY_SUSPENDER, alloc_record_data_qstring(channel->suspended->suspender));
        if(channel->suspended->expires)
        {
            snprintf(str, 12, FMT_TIME_T, channel->suspended->expires);
            dict_insert(obj, KEY_SUSPEND_EXPIRES, alloc_record_data_qstring(str));
        }
        if(channel->suspended->reason)
        {
            dict_insert(obj, KEY_SUSPEND_REASON, alloc_record_data_qstring(channel->suspended->reason));
        }
    }

    inttobase64(flags, channel->flags, FLAGS_STRING_LENGTH);
    safestrncpy(flags + FLAGS_STRING_LENGTH, channel->options, sizeof(flags) - (FLAGS_STRING_LENGTH + 1));
    dict_insert(obj, KEY_FLAGS, alloc_record_data_qstring(flags));

    if(channel->mode_on || channel->mode_off)
    {
	char modes[MODELEN];

	create_mode_lock(modes, channel);
	dict_insert(obj, KEY_MODES, alloc_record_data_qstring(modes));
    }

    chanserv_write_users(obj, channel->users);

    if(channel->banCount)
    {
	chanserv_write_bans(obj, channel->bans);
    }

    if(dict_size(channel->notes))
    {
        dict_t subobj = alloc_object();
        dict_iterator_t it;

        for(it = dict_first(channel->notes); it; it = iter_next(it))
        {
            dict_insert(obj, iter_key(it), alloc_record_data_qstring(iter_data(it)));
        }
        dict_insert(obj, KEY_NOTES, alloc_record_data_object(subobj));
    }

    channame = IsSuspended(channel) ? channel->suspended->name : channel->channel->name;
    dict_insert(db, channame, alloc_record_data_object(obj));
    return;
}

static void
chanserv_write_channels(dict_t db)
{
    struct chanData *channel;
    dict_t channels = alloc_object();

    dict_insert(db, KEY_CHANNELS, alloc_record_data_object(channels));

    for(channel = channelList; channel; channel = channel->next)
    {
	chanserv_write_channel(channels, channel);
    }
}

static int
chanserv_write_command(const char *key, void *data, void *extra)
{
    struct chanserv_cmd *command = data;
    dict_t obj = alloc_object();
    char *str;

    str = alloca(12);
    snprintf(str, 12, "%d", command->access);
    dict_insert(obj, KEY_ACCESS, alloc_record_data_qstring(str));

    snprintf(str, 12, "%li", command->flags);
    dict_insert(obj, KEY_FLAGS, alloc_record_data_qstring(str));

    snprintf(str, 12, "%.2f", command->weight);
    dict_insert(obj, KEY_WEIGHT, alloc_record_data_qstring(str));

    dict_insert(extra, key, alloc_record_data_object(obj));
    return 0;
}

static void
chanserv_write_commands(dict_t db)
{
    dict_t commands = alloc_object();

    dict_insert(db, KEY_COMMANDS, alloc_record_data_object(commands));
    dict_foreach(chanserv_func_dict, chanserv_write_command, commands);
}

static void
chanserv_write_note_types(dict_t db)
{
    dict_t types, type_obj;
    dict_iterator_t it;
    struct note_type *ntype;
    char buffer[16];
    const char *str;

    types = alloc_object();
    for(it = dict_first(note_types); it; it = iter_next(it))
    {
        ntype = iter_data(it);
        type_obj = alloc_object();
        switch(ntype->set_access_type)
        {
        case NOTE_SET_CHANNEL_ACCESS:
            snprintf(buffer, sizeof(buffer), "%d", levelMap[ntype->set_access.min_ulevel]);
            dict_insert(type_obj, KEY_NOTE_CHANNEL_ACCESS, alloc_record_data_qstring(buffer));
            break;
        case NOTE_SET_CHANNEL_SETTER:
            dict_insert(type_obj, KEY_NOTE_SETTER_ACCESS, alloc_record_data_qstring("1"));
            break;
        case NOTE_SET_PRIVILEGED: default:
            snprintf(buffer, sizeof(buffer), "%d", ntype->set_access.min_opserv);
            dict_insert(type_obj, KEY_NOTE_OPSERV_ACCESS, alloc_record_data_qstring(buffer));
            break;
        }
        switch(ntype->visible_type)
        {
        case NOTE_VIS_ALL: str = KEY_NOTE_VIS_ALL; break;
        case NOTE_VIS_CHANNEL_USERS: str = KEY_NOTE_VIS_CHANNEL_USERS; break;
        case NOTE_VIS_PRIVILEGED: default: str = KEY_NOTE_VIS_PRIVILEGED; break;
        }
        dict_insert(type_obj, KEY_NOTE_VISIBILITY, alloc_record_data_qstring(str));
        snprintf(buffer, sizeof(buffer), "%u", ntype->max_length);
        dict_insert(type_obj, KEY_NOTE_MAX_LENGTH, alloc_record_data_qstring(buffer));
        dict_insert(types, iter_key(it), alloc_record_data_object(type_obj));
    }
    dict_insert(db, KEY_NOTE_TYPES, alloc_record_data_object(types));
}

static void
write_dnrs_helper(dict_t db, dict_t dnrs)
{
    dict_t dnr_obj;
    struct do_not_register *dnr;
    dict_iterator_t it;

    for(it = dict_first(dnrs); it; it = iter_next(it))
    {
        dnr = iter_data(it);
        dnr_obj = alloc_object();
        dict_insert(dnr_obj, KEY_DNR_SETTER, alloc_record_data_qstring(dnr->setter));
        dict_insert(dnr_obj, KEY_DNR_REASON, alloc_record_data_qstring(dnr->reason));
        dict_insert(db, dnr->chan_name, alloc_record_data_object(dnr_obj));
    }
}

static void
chanserv_write_dnrs(dict_t db)
{
    dict_t dnrs;

    dnrs = alloc_object();
    write_dnrs_helper(dnrs, handle_dnrs);
    write_dnrs_helper(dnrs, plain_dnrs);
    write_dnrs_helper(dnrs, mask_dnrs);
    dict_insert(db, KEY_DNR, alloc_record_data_object(dnrs));
}

static void
chanserv_db_write(void)
{
    FILE *f = fopen(TEMP_DB_NAME, "w");
    dict_t db = alloc_database();

    chanserv_write_note_types(db);
    chanserv_write_commands(db);
    chanserv_write_dnrs(db);
    chanserv_write_channels(db);

    write_database(f, db);
    fclose(f);

    free_database(db);

    if(rename(TEMP_DB_NAME, CHANSERV_DB_NAME) == -1)
    {
	log(CS_LOG, LOG_ERROR, "Error renaming database: %s\n", strerror(errno));
    }
    lastwrite = now;
}

void
chanserv_timed_write(void *data)
{
    (void)data;

    chanserv_db_write();
    log(CS_LOG, LOG_INFO, "Made automatic backup of ChanServ database.\n");
    timeq_add(now + chanserv_conf.db_backup_frequency, chanserv_timed_write, NULL);
}

static CHANSERV_FUNC(cmd_write)
{
    struct timeval start, stop;
    (void)channel;(void)argc;(void)argv;

    gettimeofday(&start, NULL);
    chanserv_db_write();
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
	stop.tv_sec -= 1;
	stop.tv_usec += 1000000;
    }
    chanserv_notice(user, CSMSG_WRITE_SUCCESS, stop.tv_sec, stop.tv_usec/1000);
    log(CS_LOG, LOG_INFO, "%s ordered a ChanServ database write.\n", user->nick);
    return 1;
}

struct channel_trans {
    struct userNode *owner;
    struct chanNode *channel;
    int got_access, access_ok : 1;
    int got_auth, auth_ok : 1;
};

static void channel_switch_timeout(void *data);

static void
abort_channel_switch(struct chanNode *channel)
{
    timeq_del(0, channel_switch_timeout, channel, TIMEQ_IGNORE_WHEN);
    dict_remove(switch_dict, channel->name);
}

struct switch_annoy {
    struct userNode *user;
    struct channelList clist;
};

static int
find_channel_switches_for(const char *channel, void *data, void *extra)
{
    struct channel_trans *sw = data;
    struct switch_annoy *sa = extra;
    (void)channel;
    if (sw->owner == sa->user) {
        channelList_append(&sa->clist, sw->channel);
    }
    return 0;
}

static void
channel_switch_timeout(void *data)
{
    struct chanNode *channel = data;
    abort_channel_switch(channel);
}

static void
abort_channel_switches_for(struct userNode *user)
{
    struct switch_annoy sa;
    unsigned int n;
    sa.user = user;
    channelList_init(&sa.clist);
    if(!switch_dict) return;
    dict_foreach(switch_dict, find_channel_switches_for, &sa);
    for (n=0; n<sa.clist.used; n++) {
        abort_channel_switch(sa.clist.list[n]);
    }
    channelList_clean(&sa.clist);
}

static void
privmsg_old_chanserv(const char *text)
{
    struct userNode *oldcs;
    oldcs = GetUserH(chanserv_conf.old_chanserv_name);
    if (oldcs) {
        irc_privmsg(chanserv->numeric, oldcs->numeric, text);
    } else {
        log(CS_LOG, LOG_ERROR, "Could not find user %s to privmsg!\n", chanserv_conf.old_chanserv_name);
    }
}

static CHANSERV_FUNC(cmd_switch)
{
    struct userNode *owner;
    struct channel_trans *trans;
    static const char *cs_msg_fmt = "access %s %s";
    char *cs_msg;
    int cs_msg_len;

    if (!chanserv_conf.old_chanserv_name) {
        chanserv_notice(user, "%s is not supported on this network.", argv[0]);
        return 0;
    }

    /* get owner */
    if (argc > 1) {
        if (IsPrivileged(user)) {
            if (!(owner = GetUserH(argv[1]))) {
                chanserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
                return 0;
            }
        } else {
            chanserv_notice(user, "Transitioning channels you don't own is a privileged operation.");
            return 0;
        }
    } else {
        owner = user;
    }

    /* check that owner is authed with nickserv */
    if (!owner->handle_info) {
        if (owner == user) {
            chanserv_notice(user, MSG_AUTHENTICATE);
        } else {
            chanserv_notice(user, MSG_USER_AUTHENTICATE, owner->nick);
        }
        return 0;
    }

    /* check that channel isn't registered with us now */
    if (channel->channel_info) {
        chanserv_notice(user, "You may not transition a channel already registered with me.");
        return 0;
    }

    /* check how busy we are now */
    if (dict_size(switch_dict) >= chanserv_conf.max_switch_load) {
        chanserv_notice(user, "I'm sorry, I'm too busy switching other channels over.  Please wait a while and try again.");
        return 0;
    }
    /* everything looks good.  build a transition struct.. */
    trans = malloc(sizeof(*trans));
    trans->owner = owner;
    trans->channel = channel;
    trans->got_access = 0;
    trans->got_auth = 0;
    dict_insert(switch_dict, channel->name, trans);
    /* fire off a message to old chanserv.. */
    cs_msg_len = strlen(cs_msg_fmt) + strlen(channel->name) + strlen(owner->nick) + 1;
    cs_msg = alloca(cs_msg_len);
    snprintf(cs_msg, cs_msg_len, cs_msg_fmt, channel->name, owner->nick);
    privmsg_old_chanserv(cs_msg);
    /* set up a timeout for the switchover.. */
    timeq_add(now + chanserv_conf.switch_timeout, channel_switch_timeout, channel);
    /* and tell the user we're starting. */
    chanserv_notice(user, "Starting to transition channel %s..", channel->name);
    return 1;
}

#define EXTRACT(N, TARGET) \
  if (pmatch[N].rm_so == -1) { \
    log(CS_LOG, LOG_ERROR, "access_regex failure: matched but no $%d\n", N); \
    return NULL; \
  } \
  partlen = pmatch[N].rm_eo - pmatch[N].rm_so; \
  TARGET = alloca(partlen + 1); \
  strncpy(TARGET, text + pmatch[N].rm_so, partlen); \
  TARGET[partlen] = 0;

static struct channel_trans *
trans_check_access_reply(const char *text, regmatch_t pmatch[])
{
    struct channel_trans *trans;
    int partlen;
    char *nick, *access, *channel;

    /* pmatch[1] should have nick */
    EXTRACT(1, nick);
    /* pmatch[2] should have access level */
    EXTRACT(2, access);
    /* pmatch[3] should have channel name */
    EXTRACT(3, channel);

    /* now look things up.. */
    trans = dict_find(switch_dict, channel, NULL);
    if (!trans || trans->got_access) {
        log(CS_LOG, LOG_INFO, "access_regex: matched, but didn't expect info for %s\n", channel);
        return trans;
    }
    if (strcasecmp(nick, trans->owner->nick)) {
        log(CS_LOG, LOG_INFO, "access_regex: matched, but didn't expect info for %s in %s\n", nick, channel);
        return NULL;
    }
    trans->got_access = 1;
    if (strcasecmp(access, "an owner")) {
        log(CS_LOG, LOG_INFO, "access_regex: matched, but %s is not an owner for %s\n", nick, channel);
        trans->access_ok = 0;
    } else {
        trans->access_ok = 1;
    }
    return trans;
}

static struct channel_trans *
trans_check_auth_reply(const char *text, regmatch_t pmatch[])
{
    struct channel_trans *trans = NULL;
    int partlen;
    char *nick, *channel;

    /* pmatch[1] should be nick */
    EXTRACT(1, nick);
    /* (work on pmatch[2] later in this function) */
    /* pmatch[3] should be channel */
    EXTRACT(3, channel);
    /* look things up.. */
    trans = dict_find(switch_dict, channel, NULL);
    if (!trans || trans->got_auth) {
        log(CS_LOG, LOG_INFO, "auth_regex: matched, but didn't expect info for %s\n", channel);
        return trans;
    }
    if (strcasecmp(nick, trans->owner->nick)) {
        log(CS_LOG, LOG_INFO, "auth_regex: matched, but didn't expect info for %s in %s\n", nick, channel);
        return NULL;
    }
    trans->got_auth = 1;
    /* if pmatch[2] is non-empty, they are not authenticated */
    if (pmatch[2].rm_so != -1) {
        log(CS_LOG, LOG_INFO, "auth_regex: matched, but %s is not authenticated in %s\n", nick, channel);
        trans->auth_ok = 0;
    } else {
        trans->auth_ok = 1;
    }
    return trans;
}

static void
finalize_transition(struct channel_trans *trans)
{
    static const char *cs_msg_fmt = "unreg %s";
    char *cs_msg;
    int cs_msg_len;
    struct chanData *new_chan;

    /* try to register it internally */
    new_chan = register_channel(trans->owner->handle_info->handle, NULL, NULL, NULL, NULL, 0, trans->channel);

    add_channel_user(new_chan, trans->owner->handle_info, USER_LEVEL_HIGHEST, 0, NULL);
    scan_handle_presence(trans->channel, trans->owner->handle_info, NULL);

    /* unregister with old chanserv */
    cs_msg_len = strlen(cs_msg_fmt) + strlen(trans->channel->name) + 1;
    cs_msg = alloca(cs_msg_len);
    snprintf(cs_msg, cs_msg_len, cs_msg_fmt, trans->channel->name);
    privmsg_old_chanserv(cs_msg);

    /* put us in there */
    AddChannelUser(chanserv, trans->channel);
    AddChannelOp(1, &chanserv, trans->channel, chanserv, 1);

    parse_mode_lock(&CHANSERV_DEFAULT_MODES, 1, new_chan);
    mod_chanmode(trans->channel, CHANSERV_DEFAULT_MODES, NULL, 0);
    irc_mode(chanserv, trans->channel, CHANSERV_DEFAULT_MODES);

    /* .. and let the user know what they won */
    chanserv_notice(trans->owner, CSMSG_REG_SUCCESS, trans->channel->name);
    log(CS_LOG, LOG_INFO, "(%s) Channel transitioned; %s is owner.\n", trans->channel->name, trans->owner->nick);
}

#undef EXTRACT

static void
chanserv_parse_notice(struct userNode *user, struct userNode *target, char *text, int server_qualified)
{
    regmatch_t pmatch[4];
    struct channel_trans *trans = NULL;
    (void)target; (void)server_qualified;

    if (strcasecmp(user->nick, chanserv_conf.old_chanserv_name)) return;
    if (!regexec(&chanserv_conf.access_regex, text, ArrayLength(pmatch), pmatch, 0)) {
        trans = trans_check_access_reply(text, pmatch);
    } else if (!regexec(&chanserv_conf.auth_regex, text, ArrayLength(pmatch), pmatch, 0)) {
        trans = trans_check_auth_reply(text, pmatch);
    }

    if (!trans || !trans->got_access || !trans->got_auth) return;
    if (!trans->access_ok) {
        chanserv_notice(trans->owner, "Cannot transfer %s: You are not listed as owner with %s\n", trans->channel->name, chanserv_conf.old_chanserv_name);
    } else if (!trans->auth_ok) {
        chanserv_notice(trans->owner, "Cannot transfer %s: You are not authenticated with %s there\n", trans->channel->name, chanserv_conf.old_chanserv_name);
    } else {
        finalize_transition(trans);
    }
    abort_channel_switch(trans->channel);
}

static void
chanserv_db_cleanup(void) {
    unreg_part_func(handle_part);
    if(TRANSITION)
    {
        unreg_del_user_func(abort_channel_switches_for);
    }
    close_helpfile(chanserv_helpfile);
    dict_delete(sChannels);
    while (channelList) unregister_channel(channelList);
    dict_delete(chanserv_func_dict);
    dict_delete(set_func_dict);
    dict_delete(handle_dnrs);
    dict_delete(plain_dnrs);
    dict_delete(mask_dnrs);
    dict_delete(note_types);
    free(chanserv_conf.ctcp_short_ban_duration);
    free(chanserv_conf.ctcp_long_ban_duration);
    if (chanserv_conf.eightball) free_string_list(chanserv_conf.eightball);
}

#define DEFINE_COMMAND(NAME, FLAGS, LEVEL) chanserv_define_func(#NAME, FLAGS, LEVEL, cmd_##NAME)
#define DEFINE_OPTION(NAME) dict_insert(set_func_dict, #NAME, opt_##NAME)

void
init_chanserv(const char *nick)
{
    chanserv = AddService(nick, "Channel Services");
    chanserv_conf_read();
    conf_register_reload(chanserv_conf_read);

    reg_privmsg_func(chanserv->num_local, chanserv_privmsg);
    if (chanserv_conf.trigger_char) {
        reg_chanmsg_func(chanserv_conf.trigger_char, chanserv, chanserv_chanmsg);
    }
    reg_chanmsg_func('\001', chanserv, chanserv_ctcp_check);

    reg_auth_func(handle_auth);
    reg_server_link_func(handle_server_link);

    reg_new_channel_func(handle_new_channel);
    reg_join_func(handle_join);
    reg_part_func(handle_part);
    reg_kick_func(handle_kick);
    reg_topic_func(handle_topic);
    reg_mode_change_func(handle_mode);

    reg_unreg_func(handle_unreg);

    if(TRANSITION)
    {
        reg_del_channel_func(abort_channel_switch);
        reg_del_user_func(abort_channel_switches_for);
        reg_notice_func(chanserv->num_local, chanserv_parse_notice);
    }

    sChannels = dict_new();
    handle_dnrs = dict_new();
    dict_set_free_data(handle_dnrs, free);
    plain_dnrs = dict_new();
    dict_set_free_data(plain_dnrs, free);
    mask_dnrs = dict_new();
    dict_set_free_data(mask_dnrs, free);

    chanserv_func_dict = dict_new();
    dict_set_free_data(chanserv_func_dict, chanserv_unref_cmd);

    DEFINE_COMMAND(help, CMD_IGNORE_EVENT, ulNone);
    DEFINE_COMMAND(readhelp, CMD_REQUIRE_OPER, ulNone);

    DEFINE_COMMAND(god, CMD_REQUIRE_GODDABLE, ulNone);
    DEFINE_COMMAND(register, CMD_REQUIRE_PRIVILEGE, ulNone);
    DEFINE_COMMAND(noregister, CMD_REQUIRE_PRIVILEGE, ulNone);
    DEFINE_COMMAND(allowregister, CMD_REQUIRE_PRIVILEGE, ulNone);
    DEFINE_COMMAND(move, CMD_REQUIRE_PRIVILEGE | CMD_REQUIRE_REGISTERED | CMD_SUSPEND_OVERRIDE, ulNone);
    DEFINE_COMMAND(opchan, CMD_REQUIRE_REGISTERED, ulCoowner);
    DEFINE_COMMAND(csuspend, CMD_REQUIRE_REGISTERED | CMD_REQUIRE_PRIVILEGE, ulNone);
    DEFINE_COMMAND(cunsuspend, CMD_REQUIRE_PRIVILEGE, ulNone);

    DEFINE_COMMAND(unregister, CMD_REQUIRE_REGISTERED | CMD_SUSPEND_OVERRIDE, ulOwner);
    DEFINE_COMMAND(merge, CMD_REQUIRE_DEFAULT, ulOwner);
    if(TRANSITION)
    {
        DEFINE_COMMAND(switch, CMD_REQUIRE_HANDLE | CMD_REQUIRE_CHANNEL, ulNone);
    }

    DEFINE_COMMAND(addowner, CMD_REQUIRE_DEFAULT, ulHelper);
    DEFINE_COMMAND(addcoowner, CMD_REQUIRE_DEFAULT, ulOwner);
    DEFINE_COMMAND(addmaster, CMD_REQUIRE_DEFAULT, ulCoowner);
    DEFINE_COMMAND(addop, CMD_REQUIRE_DEFAULT, ulMaster);
    DEFINE_COMMAND(addpeon, CMD_REQUIRE_DEFAULT, ulMaster);
    DEFINE_COMMAND(adduser, CMD_REQUIRE_DEFAULT, ulMaster);

    DEFINE_COMMAND(delowner, CMD_REQUIRE_DEFAULT, ulHelper);
    DEFINE_COMMAND(delcoowner, CMD_REQUIRE_DEFAULT, ulOwner);
    DEFINE_COMMAND(delmaster, CMD_REQUIRE_DEFAULT, ulCoowner);
    DEFINE_COMMAND(delop, CMD_REQUIRE_DEFAULT, ulMaster);
    DEFINE_COMMAND(delpeon, CMD_REQUIRE_DEFAULT, ulMaster);
    DEFINE_COMMAND(deluser, CMD_REQUIRE_DEFAULT, ulMaster);

    DEFINE_COMMAND(mdelowner, CMD_REQUIRE_DEFAULT, ulHelper);
    DEFINE_COMMAND(mdelcoowner, CMD_REQUIRE_DEFAULT, ulOwner);
    DEFINE_COMMAND(mdelmaster, CMD_REQUIRE_DEFAULT, ulCoowner);
    DEFINE_COMMAND(mdelop, CMD_REQUIRE_DEFAULT, ulMaster);
    DEFINE_COMMAND(mdelpeon, CMD_REQUIRE_DEFAULT, ulMaster);

    DEFINE_COMMAND(trim, CMD_REQUIRE_DEFAULT, ulMaster);

    DEFINE_COMMAND(clvl, CMD_REQUIRE_DEFAULT, ulMaster);
    DEFINE_COMMAND(giveownership, CMD_REQUIRE_DEFAULT, ulOwner);

    DEFINE_COMMAND(up, CMD_REQUIRE_DEFAULT, ulPeon);
    DEFINE_COMMAND(down, CMD_REQUIRE_REGISTERED, ulNone);
    DEFINE_COMMAND(upall, CMD_REQUIRE_HANDLE, ulNone);
    DEFINE_COMMAND(downall, CMD_REQUIRE_HANDLE, ulNone);
    DEFINE_COMMAND(op, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(deop, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(voice, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(devoice, CMD_REQUIRE_DEFAULT, ulOp);

    DEFINE_COMMAND(kickban, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(kick, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(ban, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(unban, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(unbanall, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(unbanme, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(open, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(topic, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(mode, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(invite, CMD_REQUIRE_DEFAULT, ulPeon);
    DEFINE_COMMAND(set, CMD_REQUIRE_DEFAULT, ulOp);
    DEFINE_COMMAND(setinfo, CMD_REQUIRE_DEFAULT | CMD_IGNORE_EVENT, USER_LEVEL_LOWEST);

    DEFINE_COMMAND(events, CMD_REQUIRE_DEFAULT | CMD_IGNORE_EVENT | CMD_SUSPEND_OVERRIDE, ulCoowner);
    DEFINE_COMMAND(addban, CMD_REQUIRE_DEFAULT, ulMaster);
    DEFINE_COMMAND(addtimedban, CMD_REQUIRE_DEFAULT, ulMaster);
    DEFINE_COMMAND(delban, CMD_REQUIRE_DEFAULT, ulMaster);
    DEFINE_COMMAND(togop, CMD_REQUIRE_DEFAULT, ulPeon);
    DEFINE_COMMAND(suspend, CMD_REQUIRE_DEFAULT, ulCoowner);
    DEFINE_COMMAND(unsuspend, CMD_REQUIRE_DEFAULT, ulCoowner);

    DEFINE_COMMAND(bans, CMD_REQUIRE_DEFAULT | CMD_IGNORE_EVENT | CMD_SUSPEND_OVERRIDE, ulPeon);
    DEFINE_COMMAND(peek, CMD_REQUIRE_DEFAULT | CMD_IGNORE_EVENT, ulOp);

    DEFINE_COMMAND(access, CMD_REQUIRE_REGISTERED | CMD_IGNORE_EVENT | CMD_SUSPEND_OVERRIDE, ulNone);
    DEFINE_COMMAND(users, CMD_REQUIRE_REGISTERED | CMD_IGNORE_EVENT | CMD_SUSPEND_OVERRIDE, ulNone);
    DEFINE_COMMAND(wlist, CMD_REQUIRE_REGISTERED | CMD_IGNORE_EVENT | CMD_SUSPEND_OVERRIDE, ulNone);
    DEFINE_COMMAND(clist, CMD_REQUIRE_REGISTERED | CMD_IGNORE_EVENT | CMD_SUSPEND_OVERRIDE, ulNone);
    DEFINE_COMMAND(mlist, CMD_REQUIRE_REGISTERED | CMD_IGNORE_EVENT | CMD_SUSPEND_OVERRIDE, ulNone);
    DEFINE_COMMAND(olist, CMD_REQUIRE_REGISTERED | CMD_IGNORE_EVENT | CMD_SUSPEND_OVERRIDE, ulNone);
    DEFINE_COMMAND(plist, CMD_REQUIRE_REGISTERED | CMD_IGNORE_EVENT | CMD_SUSPEND_OVERRIDE, ulNone);
    DEFINE_COMMAND(info, CMD_IGNORE_EVENT, ulNone);
    DEFINE_COMMAND(seen, CMD_REQUIRE_REGISTERED | CMD_IGNORE_EVENT | CMD_SUSPEND_OVERRIDE, ulNone);

    DEFINE_COMMAND(note, CMD_REQUIRE_REGISTERED, ulNone);
    DEFINE_COMMAND(delnote, CMD_REQUIRE_ACCESS, ulNone);

    DEFINE_COMMAND(command, CMD_IGNORE_EVENT, ulNone);
    DEFINE_COMMAND(netinfo, CMD_IGNORE_EVENT, ulNone);
    DEFINE_COMMAND(ircops, CMD_IGNORE_EVENT, ulNone);
    DEFINE_COMMAND(helpers, CMD_IGNORE_EVENT, ulNone);
    DEFINE_COMMAND(staff, CMD_IGNORE_EVENT, ulNone);
    DEFINE_COMMAND(version, CMD_IGNORE_EVENT, ulNone);

    DEFINE_COMMAND(say, CMD_REQUIRE_CHANNEL | CMD_REQUIRE_OPER, ulNone);
    DEFINE_COMMAND(emote, CMD_REQUIRE_CHANNEL | CMD_REQUIRE_OPER, ulNone);
    DEFINE_COMMAND(last, CMD_REQUIRE_PRIVILEGE | CMD_IGNORE_EVENT, ulNone);
    DEFINE_COMMAND(expire, CMD_REQUIRE_PRIVILEGE, ulNone);
    DEFINE_COMMAND(search, CMD_REQUIRE_PRIVILEGE | CMD_IGNORE_EVENT, ulNone);
    DEFINE_COMMAND(unvisited, CMD_REQUIRE_PRIVILEGE | CMD_IGNORE_EVENT, ulNone);
    DEFINE_COMMAND(write, CMD_REQUIRE_HANDLE | CMD_REQUIRE_OPER, ulNone);

#ifndef NO_LAME_TRICKS
    if (chanserv_conf.lame_tricks) {
        DEFINE_COMMAND(unf, CMD_IGNORE_EVENT | CMD_TOY, ulNone);
        DEFINE_COMMAND(ping, CMD_IGNORE_EVENT | CMD_TOY, ulNone);
        DEFINE_COMMAND(wut, CMD_IGNORE_EVENT | CMD_TOY, ulNone);
        DEFINE_COMMAND(8ball, CMD_IGNORE_EVENT | CMD_TOY, ulNone);
        DEFINE_COMMAND(d, CMD_IGNORE_EVENT | CMD_TOY, ulNone);
        DEFINE_COMMAND(huggle, CMD_IGNORE_EVENT | CMD_TOY, ulNone);
    }
#endif

    set_func_dict = dict_new();
    DEFINE_OPTION(defaulttopic);
    DEFINE_OPTION(topicmask);
    DEFINE_OPTION(greeting);
    DEFINE_OPTION(modes);
    DEFINE_OPTION(strictop);
    DEFINE_OPTION(autoop);
    DEFINE_OPTION(protect);
    DEFINE_OPTION(enfmodes);
    DEFINE_OPTION(enftopic);
    DEFINE_OPTION(pubcmd);
    DEFINE_OPTION(voice);
    DEFINE_OPTION(greet);
    DEFINE_OPTION(greetusers);
    DEFINE_OPTION(userinfo);
    DEFINE_OPTION(dynlimit);
    DEFINE_OPTION(topicsnarf);
    DEFINE_OPTION(nodelete);
    DEFINE_OPTION(toys);
    DEFINE_OPTION(setters);
    DEFINE_OPTION(topicrefresh);
    DEFINE_OPTION(ctcpusers);
    DEFINE_OPTION(ctcpreaction);
    /* compatability alias.. */
    dict_insert(set_func_dict, "topic", opt_defaulttopic);

    note_types = dict_new();
    dict_set_free_data(note_types, chanserv_deref_note_type);

    alias_register_service("ChanServ", chanserv_command_search, chanserv_add_alias, chanserv_del_alias);

    chanserv_db_read(CHANSERV_DB_NAME);
    chanserv_helpfile_read();

    if(chanserv_conf.db_backup_frequency)
    {
	timeq_add(now + chanserv_conf.db_backup_frequency, chanserv_timed_write, NULL);
    }

    if(chanserv_conf.channel_expire_frequency)
    {
	timeq_add(now + chanserv_conf.channel_expire_frequency, expire_channels, NULL);
    }

    if(chanserv_conf.refresh_period)
    {
        time_t next_refresh;
        next_refresh = (now + chanserv_conf.refresh_period - 1) / chanserv_conf.refresh_period * chanserv_conf.refresh_period;
        timeq_add(next_refresh, chanserv_refresh_topics, NULL);
    }

    reg_exit_func(chanserv_db_cleanup);
    reg_db_write_func(chanserv_db_write);
}
